/*
 * Copyright (c) 1983,1988 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that this notice is preserved and that due credit is given
 * to the University of California at Berkeley. The name of the University
 * may not be used to endorse or promote products derived from this
 * software without specific prior written permission. This software
 * is provided ``as is'' without express or implied warranty.
 */

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <errno.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <snmp.h>      

#include "main.h"

extern int  protopr   (void);
extern int  tcp_stats (void);
extern int  udp_stats (void);
extern int  ip_stats  (void);
extern int  icmp_stats(void);
extern int  rt_stats  (void);
extern void routepr   (void);
extern void intpr     (int);

struct protox *name2protox (char *name);

struct protox {
       u_char pr_wanted;         /* 1 if wanted, 0 otherwise */
       int  (*pr_cblocks)();     /* control blocks printing routine */
       int  (*pr_stats)();       /* statistics printing routine */
       char  *pr_name;           /* well-known name */
     } protox[] = {
       { 1, protopr, tcp_stats,  "tcp"  },
       { 1, 0,       udp_stats,  "udp"  },
       { 1, 0,       ip_stats,   "ip"   },
       { 1, 0,       icmp_stats, "icmp" },
       { 0, 0,       0,          0      }
     };

int   aflag, iflag, nflag, pflag, rflag, sflag;
int   interval;
char *interface;

int debug = 0;
int print_errors = 0;

struct snmp_session *Session;

static void usage (char *name)
{
  printf ("usage: %s "
          "[-dainrs] [-p proto] [-I interface] [-S interval] host community\n",
          name);
  exit (1);
}

int main (int argc, char **argv)
{
  struct protoent    *p;
  struct protox      *tp = NULL; /* for printing cblocks & stats */
  struct snmp_session session;

  char *name;
  char *host      = NULL;
  char *community = NULL;
  int   ch;

  name = argv[0];

  while ((ch = getopt(argc, argv, "?dainrsp:I:S:")) != EOF)
      switch (ch)
      {
        case 'd':
             debug++;
             snmp_dump_packet (1);
             dbug_init();
             break;

        case 'a':
             aflag++;
             break;

        case 'i':
             iflag++;
             break;

        case 'n':
             nflag++;
             break;

        case 'r':
             rflag++;
             break;

        case 's':
             sflag++;
             break;

        case 'p':
             if ((tp = name2protox(optarg)) == NULL)
             {
               fprintf (stderr, "%s: unknown or uninstrumented protocol\n",
                        optarg);
               return (10);
             }
             pflag++;
             break;

        case 'I':
             iflag++;
             interface = optarg;
             break;

        case 'S':
             interval = atoi (optarg);
             if (interval <= 0)
                usage (name);
             iflag++;
             break;

        case '?':
        default:
             usage (name);
      }

  argc -= optind;
  argv += optind;

  if (argc < 2)
     usage (name);

  host = *argv++;
  community = *argv;

  init_mib();
  sock_init();

  memset (&session, 0, sizeof(session));
  session.peername      = host;
  session.community     = (u_char*)community;
  session.community_len = strlen ((char*)community);
  session.retries       = SNMP_DEFAULT_RETRIES;
  session.timeout       = SNMP_DEFAULT_TIMEOUT;
  session.authenticator = NULL;
  snmp_synch_setup (&session);

  Session = snmp_open (&session);
  if (Session == NULL)
  {
    printf ("Couldn't open snmp\n");
    return (-1);
  }

  if (pflag)
  {
    if (tp->pr_stats)
         (*tp->pr_stats)();
    else printf ("%s: no stats routine\n", tp->pr_name);
    return (0);
  }

  /*
   * Keep file descriptors open to avoid overhead
   * of open/close on each call to get* routines.
   */
  sethostent (1);
  setnetent (1);
  if (iflag)
  {
    intpr (interval);
    return (0);
  }
  if (rflag)
  {
    if (sflag)
         rt_stats();
    else routepr();
    return (0);
  }

  setprotoent (1);
  setservent (1);

  while ((p = getprotoent()) != NULL)
  {
    for (tp = protox; tp->pr_name; tp++)
         if (!strcmp(tp->pr_name, p->p_name) == 0)
            break;

    if (tp->pr_name == 0 || tp->pr_wanted == 0)
       continue;

    if (sflag)
    {
      if (tp->pr_stats)
         (*tp->pr_stats)();
    }
    else if (tp->pr_cblocks)
           (*tp->pr_cblocks)();
  }
  endprotoent();
  return (0);
}

char *plural (int n)
{
  return (n != 1 ? "s" : "");
}

/*
 * Find the protox for the given "well-known" name.
 */
struct protox *knownname (char *name)
{
  struct protox *tp;
	
  for (tp = protox; tp->pr_name; tp++)
      if (!strcmp(tp->pr_name, name))
         return (tp);
  return (NULL);
}

/*
 * Find the protox corresponding to name.
 */
struct protox *name2protox (char *name)
{
  struct protox   *tp;
  struct protoent *p;
  char **alias;           /* alias from p->aliases */

  /*
   * Try to find the name in the list of "well-known" names. If that
   * fails, check if name is an alias for an Internet protocol.
   */
  if ((tp = knownname(name)) != NULL)
     return (tp);

  setprotoent (1);                 /* make protocol lookup cheaper */
  while ((p = getprotoent()) != NULL)
  {
    /* assert: name not same as p->name */
    for (alias = p->p_aliases; *alias; alias++)
     if (!strcmp(name, *alias))
     {
       endprotoent();
       return knownname(p->p_name);
     }
  }
  endprotoent();
  return (NULL);
}
