/***********************************************************
	Copyright 1988, 1989 by Carnegie Mellon University

                      All Rights Reserved

Permission to use, copy, modify, and distribute this software and its 
documentation for any purpose and without fee is hereby granted, 
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in 
supporting documentation, and that the name of CMU not be
used in advertising or publicity pertaining to distribution of the
software without specific, written prior permission.  

CMU DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING
ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL
CMU BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR
ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
SOFTWARE.
******************************************************************/

/*
 * snmpbulk.c - send snmp GETBULK requests to a network entity.
 *
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <snmp.h>


int non_repeaters = 0;
int max_reps = 50;

void usage (void)
{
  printf ("usage: snmpbulk [-n non-repeaters] [-r repeaters] gateway-name\n"
          "  community-name object-identifier [object-identifier ...]\n");
  exit (1);
}


int main (int argc, char **argv)
{
  struct snmp_session   session, *ss;
  struct snmp_pdu      *pdu, *response;
  struct variable_list *vars;
  char *gateway   = NULL;
  char *community = NULL;
  int   count, current_name = 0;
  char *names[128];
  oid   name[MAX_NAME_LEN];
  int   name_length;
  int   ch, status;


  /*
   * usage: snmpbulk gateway-name community-name
   */
  while ((ch = getopt(argc,argv, "?dn:r:")) != EOF)
        switch (ch)
        {
          case 'd':
               snmp_dump_packet (1);
               dbug_init();
               break;
          case 'n':
               non_repeaters = atoi (optarg);
               break;
          case 'r':
               max_reps = atoi (optarg);
               break;
          default:
               usage();
        }

  argc -= optind;
  argv += optind;

  gateway   = *argv++;
  community = *argv++;
  argc -= 2;
  while (argc > 0 && *argv)
      names[current_name++] = *argv++;

  if (!gateway || !community || !current_name)
     usage();

  init_mib();
  sock_init();

  memset (&session,0,sizeof(session));
  session.peername      = gateway;
  session.community     = (u_char*) community;
  session.community_len = strlen ((char*)community);
  session.retries       = SNMP_DEFAULT_RETRIES;
  session.timeout       = SNMP_DEFAULT_TIMEOUT;
  session.authenticator = NULL;
  session.Version       = SNMP_VERSION_2;
  snmp_synch_setup (&session);

  ss = snmp_open (&session);
  if (ss == NULL)
  {
    printf ("Couldn't open snmp\n");
    return (-1);
  }

  pdu = snmp_pdu_create (SNMP_PDU_GETBULK);
  pdu->errstat  = non_repeaters;
  pdu->errindex = max_reps;

  for (count = 0; count < current_name; count++)
  {
    name_length = MAX_NAME_LEN;
    if (!read_objid (names[count], name, &name_length))
       printf ("Invalid object identifier: %s\n", names[count]);

    snmp_add_null_var (pdu, name, name_length);
  }

retry:
  status = snmp_synch_response (ss, pdu, &response);
  if (status == STAT_SUCCESS)
  {
    if (response->errstat == SNMP_ERR_NOERROR)
    {
      if (response->command == SNMP_PDU_REPORT)
         printf ("This is an error report.\n");
      for (vars = response->variables; vars; vars = vars->next_variable)
         print_variable (vars->name, vars->name_length, vars);
    }
    else
    {
      printf ("Error in packet.\nReason: %s\n", snmp_errstring (response->errstat));
      if (response->errstat == SNMP_ERR_NOSUCHNAME)
      {
        printf ("This name doesn't exist: ");
        for (count = 1, vars = response->variables;
             vars && count != response->errindex;
             vars = vars->next_variable, count++)
            ;
        if (vars)
           print_objid (vars->name, vars->name_length);
        printf ("\n");
      }
      if ((pdu = snmp_fix_pdu(response, GET_REQ_MSG)) != NULL)
        goto retry;
    }
  }
  else if (status == STAT_TIMEOUT)
       printf ("No Response from %s\n", gateway);

  else      /* status == STAT_ERROR */
    printf ("An error occurred, Quitting\n");

  if (response)
    snmp_free_pdu (response);
  snmp_close (ss);

  return (0);
}
