/**********************************************************************
 *
 *           Copyright 1998 by Carnegie Mellon University
 * 
 *                       All Rights Reserved
 * 
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose and without fee is hereby granted,
 * provided that the above copyright notice appear in all copies and that
 * both that copyright notice and this permission notice appear in
 * supporting documentation, and that the name of CMU not be
 * used in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.
 * 
 * CMU DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING
 * ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL
 * CMU BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR
 * ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
 * ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
 * SOFTWARE.
 * 
 **********************************************************************/

/*
 * snmpdelta.c - Monitor deltas of integer valued SNMP variables
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <ctype.h>
#include <time.h>
#include <snmp.h>

#ifdef __DJGPP__
#include <unistd.h>
#endif


#define MAX_ARGS 256


char *SumFile = "Sum";
char *Argv[MAX_ARGS];
int   Argc;

/*
 * Information about the handled variables
 */
struct varInfo {
       char *name;
       oid  *oid;
       int   oidlen;
       char  descriptor[64];
       u_int value;
       int   spoiled;

       /* Peak information */
       int   PeakCount;
       float Peak;
       float PeakAverage;

       /* Max value */
       float Max;

       /* System Time */
       time_t Time;
     };

static int Verbose = 0;
static int Period  = 1;

void period_SetPeriod (int P)
{
  if (Verbose)
    printf (">> Period set to %d seconds.\n", P);
  Period = P;
}

/*
 * Wait for the next packet, or sleep.
 */
void period_WaitForPeriod (void)
{
  struct timeval time, *tv = &time;
  struct tm tm;
  time_t nexthour;
  static int target = 0;

  gettimeofday (tv,NULL);
  if (target)
  {
    /* Already calculated the target time.  Just increase it's timer,
     * and wait for it.
     */
    target += Period;
  }
  else
  {
    /* Based on the current time */
    memcpy (&tm, localtime (&tv->tv_sec), sizeof (tm));

    /* Calculate the next hour */
    tm.tm_sec = 0;
    tm.tm_min = 0;
    tm.tm_hour++;

    /* And figure out the number of seconds until then */
    nexthour = mktime (&tm);

    /* Given the number of seconds remaining until the next hour,
     * calculate how many we must skew base time, so that interval
     * always ends on an hourly basis.
     */
    target = (nexthour - tv->tv_sec) % Period;
    if (target == 0)
        target = Period;
    target += tv->tv_sec;
  }

  /* ASSERT:  Target is the time_t value of the next time we should
   * wake up.
   */

  /* How much longer is it really?
   */
  tv->tv_sec = target - tv->tv_sec;
  if (tv->tv_usec != 0)
  {
    tv->tv_sec--;
    tv->tv_usec = 1000000 - tv->tv_usec;
  }
  if (tv->tv_sec < 0)
  {
    /* ran out of time, schedule immediately */
    return;
  }

  if (Verbose)
    printf (">> Sleeping for %d seconds\n", tv->tv_sec);

  sleep (tv->tv_sec);
}

/**********************************************************************/

static int PeakInterval = 0;

void peak_SetInterval (int P)
{
  if (Verbose)
     printf (">> Peak period set to %d seconds.\n", P);
  PeakInterval = P;
}


int peak_WaitForStart (void)
{
  struct timeval time, *tv = &time;
  struct tm tm;
  time_t SecondsAtNextHour;
  int    seconds;

  seconds = Period * PeakInterval;

  /* Find the current time */
  gettimeofday (tv,NULL);

  /* Create a tm struct from it */
  memcpy (&tm, localtime(&tv->tv_sec), sizeof(tm));

  /* Calculate the next hour */
  tm.tm_sec = 0;
  tm.tm_min = 0;
  tm.tm_hour++;
  SecondsAtNextHour = mktime (&tm);

  /* Now figure out the amount of time to sleep */
  return ((SecondsAtNextHour - tv->tv_sec) % seconds);
}

/**********************************************************************/

void log (char *prefix, char *message)
{
  static char *filename = "snmpdelta.log";
  FILE        *fp       = fopen (filename,"at");

  if (fp == NULL)
  {
    fprintf (stderr, "Couldn't open %s\n", filename);
    return;
  }
  fprintf (fp, "%-30s: %s\n", prefix, message);
  fclose (fp);
}

void sprint_descriptor (char *buffer, struct varInfo *vip)
{
  char buf[256], *cp;

  sprint_objid (buf, vip->oid, vip->oidlen);
  for (cp = buf; *cp; cp++)
      ;
  while (cp >= buf)
  {
    if (isalpha (*cp))
       break;
    cp--;
  }
  while (cp >= buf)
  {
    if (*cp == '.')
       break;
    cp--;
  }
  cp++;
  if (cp < buf)
     cp = buf;
  strcpy (buffer, cp);
}

void processFileArgs (char *fileName)
{ 
  char  buf[257], *cp;
  int   blank, linenumber = 0;
  FILE *fp = fopen (fileName, "r");

  if (fp == NULL)
     return;

  while (fgets(buf,256,fp))
  {
    linenumber++;
    if (strlen (buf) > 256)
    {
      fprintf (stderr, "Line too long on line %d of %s\n",
	       linenumber, fileName);
      exit (1);
    }
    if (buf[0] == '#')
       continue;
    blank = TRUE;
    for (cp = buf; *cp; cp++)
      if (!isspace (*cp))
      {
	blank = FALSE;
	break;
      }
    if (blank)
       continue;
    Argv[Argc++] = strdup (buf);   /* ignore newline */
  }
  fclose (fp);
}

void Usage (void)
{
  fprintf (stderr,
           "usage: snmpdelta [-f commandFile] [-l] [-L SumFileName] [-s] [-k]\n"
           "                 [-d] [-t] [-S] [-v vars/pkt] [-p period] [-P peaks]\n"
           "                 [-Z port] [-V] host community oid [oid ...]\n");
  exit (1);
}

oid sysUpTimeOid[9] = { 1, 3, 6, 1, 2, 1, 1, 3, 0 };
int sysUpTimeLen    = 9;

int main (int argc, char **argv)
{
  struct snmp_session   session, *ss;
  struct snmp_pdu      *pdu, *response;
  struct variable_list *vars;
  struct varInfo varinfo[128], *vip;
  struct timeval tv;
  struct tm      tm;

  char  *gateway   = NULL;
  char  *community = NULL;
  char   buf[64], timestring[64], valueStr[64], peakStr[64], outstr[256];
  u_int  value, status;
  float  printvalue;
  int    count, current_name = 0;
  int    fileout = 0, dosum = 0;
  int    keepSeconds = 0;
  int    arg, ch, sum;
  int    begin, end, last_end;
  int    varbindsPerPacket = 60;
  int    print = 1;

  int    Port    = SNMP_DEFAULT_REMPORT;
  int    Retries = SNMP_DEFAULT_RETRIES;
  int    Timeout = SNMP_DEFAULT_TIMEOUT;

  int    PrintTimeStamp = 0;
  int    PrintMax = 0;

  int    DeltaTime = 0;
  time_t Time_Last = 0;		/* Last time we checked */
  time_t Time_Now;		/* Current time */
  time_t Time_Delta;		/* Time since last check */

  init_mib();

  /*
   * usage: snmpdelta gateway-name community-name [-f commandFile] [-l]
   * [-v vars/pkt] [-s] [-t] [-S] [-p period] [-P peaks] [-k] [ -Z port ]
   * -V [-L SumFileName] variable list ..
   */

  for (arg = 1; arg < argc && arg < MAX_ARGS; arg++)
      Argv[arg] = strdup (argv[arg]);
  Argc = argc;

  while ((ch = getopt(Argc,Argv, "?Z:R:T:Vdp:P:tsmv:Sf:lL:k")) != EOF)
      switch (ch)
      {
        case 'Z':
             Port = atoi (optarg);
	     break;
        case 'R':
             Retries = atoi (optarg);
	     break;
        case 'T':
             Timeout = atoi (optarg);
	     break;
        case 'V':
	     Verbose = 1;
	     break;
        case 'd':
	     if (Verbose)
	       printf (">> Will dump packets\n");
	     snmp_dump_packet (1);
             dbug_init();
	     break;
        case 'p':
             period_SetPeriod (atoi(optarg));
	     break;
        case 'P':
             peak_SetInterval (atoi(optarg));
	     break;
        case 't':
	     DeltaTime = 1;
	     break;
        case 's':
	     PrintTimeStamp = 1;
	     break;
        case 'm':
	     PrintMax = 1;
	     break;
        case 'v':
             varbindsPerPacket = atoi (optarg);
	     break;
        case 'S':
	     dosum = 1;
	     break;
        case 'f':
             processFileArgs (optarg);
	     break;
        case 'l':
	     fileout = 1;
	     break;
        case 'L':
             SumFile = optarg;
	     break;
        case 'k':
	     keepSeconds = 1;
	     break;
        default:
             Usage();
      }

  argc  = Argc;
  argv  = Argv;
  argc -= optind;
  argv += optind;

  gateway   = *argv++;
  community = *argv++;
  argc -= 2;
  while (argc > 0 && *argv)
      varinfo[current_name++].name = *argv++;

  if (dosum)
     varinfo[current_name++].name = NULL;

  if (!gateway || !community || !current_name)
     Usage();

  init_mib();
  sock_init();

  memset (&session,0,sizeof(session));

  session.peername      = gateway;
  session.community     = (u_char *) community;
  session.community_len = strlen (community);
  session.retries       = Retries;
  session.timeout       = Timeout;
  session.remote_port   = Port;

  snmp_synch_setup (&session);
  ss = snmp_open (&session);
  if (ss == NULL)
  {
    fprintf (stderr, "Couldn't open snmp\n");
    return (-1);
  }

  for (count = 0; count < current_name; count++)
  {
    vip = varinfo + count;
    if (vip->name)
    {
      vip->oidlen = MAX_NAME_LEN;
      vip->oid    = malloc (sizeof(oid)*vip->oidlen);
      if (!read_objid(vip->name, vip->oid, &vip->oidlen))
         fprintf (stderr, "Invalid object identifier: %s\n", vip->name);
      sprint_descriptor (vip->descriptor, vip);
    }
    else
    {
      vip->oidlen = 0;
      strcpy (vip->descriptor, SumFile);
    }
    vip->value = 0;
    vip->Time  = 0;
    vip->Max   = 0;
    if (PeakInterval)
    {
      vip->PeakCount   = -1;
      vip->Peak        = 0;
      vip->PeakAverage = 0;
    }
  }

  /* Wait until we should start
   */
  period_WaitForPeriod();

  end = current_name;
  sum = 0;
  while (1)
  {
    pdu = snmp_pdu_create (SNMP_PDU_GET);

    /* Ask about time only if we care
     */
    if (DeltaTime)
       snmp_add_null_var (pdu, sysUpTimeOid, sysUpTimeLen);

    if (end == current_name)
         count = 0;
    else count = end;
    begin = count;
    for (; count < current_name && count < (begin+varbindsPerPacket-DeltaTime);
         count++)
    {
      if (varinfo[count].oidlen)
        snmp_add_null_var (pdu, varinfo[count].oid, varinfo[count].oidlen);
    }
    last_end = end;
    end = count;

  retry:
    status = snmp_synch_response (ss, pdu, &response);
    if (status == STAT_SUCCESS)
    {
      if (response->errstat == SNMP_ERR_NOERROR)
      {
        vars = response->variables;

	/* Print a timestamp
	 */
	if (PrintTimeStamp)
	{
          gettimeofday (&tv,NULL);
          memcpy (&tm, localtime(&tv.tv_sec), sizeof(tm));

	  /* Print seconds only if we're NOT keeping track of peaks
	   * or told to.
	   */
          if (((Period % 60) && (!PeakInterval ||
              ((Period*PeakInterval) % 60))) || keepSeconds)
               sprintf (timestring, " [%02d:%02d:%02d %d/%d]",
                        tm.tm_hour, tm.tm_min, tm.tm_sec,
                        tm.tm_mon+1,tm.tm_mday);
          else sprintf (timestring, " [%02d:%02d %d/%d]",
                        tm.tm_hour, tm.tm_min,
                        tm.tm_mon+1,tm.tm_mday);
	}

	/* Time
	 */
	if (DeltaTime)
	{
	  if (!vars)
	  {
	    printf ("Missing variable in reply\n");
	    continue;
	  }
          Time_Now = *(vars->val.integer);
	  vars = vars->next_variable;
	}
	else
          Time_Now = 1;

	/* Now, parse the variable information.
	 */
	for (count = begin; count < end; count++)
	{
	  vip = varinfo + count;
	  if (vip->oidlen)
	  {
	    if (!vars)
	    {
	      printf ("Missing variable in reply\n");
	      break;
	    }
	    value = *(vars->val.integer) - vip->value;
	    vip->value = *(vars->val.integer);
	    vars = vars->next_variable;
	  }
	  else
	  {
	    value = sum;
	    sum = 0;
	  }

	  /* Calculate difference in time since last poll
	   */
	  Time_Delta = (Time_Now - vip->Time);
	  if (Time_Delta <= 0)
              Time_Delta = 100;   /* Default */

	  /* Update time stat
	   */
	  Time_Last = vip->Time;
	  vip->Time = Time_Now;

	  /* Only print stats if we have checked at least once before.
	   */
	  if (Time_Last == 0)
             continue;

	  if (vip->oidlen)
             sum += value;

	  sprintf (outstr, "%s %s", timestring, vip->descriptor);

	  /* Print time interval
	   */
	  if (DeltaTime)
	  {
	    /* per second average */
	    printvalue = ((float) value * 100) / Time_Delta;
	    sprintf (valueStr, " /sec: %.2f", printvalue);
	  }
	  else
	  {
	    /* per seconds passed */
	    printvalue = value;
            sprintf (valueStr, " /%d sec: %.0f", Period, printvalue);
	  }

	  if (!PeakInterval)
             strcat (outstr, valueStr);
          else
	  {
	    print = 0;
	    if (vip->PeakCount == -1)
	    {
              if (peak_WaitForStart() == 0)
                 vip->PeakCount = 0;
	    }
	    else
	    {
	      vip->PeakAverage += printvalue;
	      if (vip->Peak < printvalue)
                  vip->Peak = printvalue;
	      if (++vip->PeakCount == PeakInterval)
	      {
                if (DeltaTime)
                {
                  /* Print the per-second peak */
                  sprintf (peakStr,
			   " /sec: %.2f	(%d sec Peak: %.2f)",
			   vip->PeakAverage / vip->PeakCount,
			   Period, vip->Peak);
                }
		else
                {
		  /* Print the total peak */
		  sprintf (peakStr,
			   " /%d sec: %.0f	(%d sec Peak: %.0f)",
			   Period,
			   vip->PeakAverage / vip->PeakCount,
			   Period, vip->Peak);
                }
		vip->PeakAverage = 0;
                vip->Peak        = 0;
                vip->PeakCount   = 0;
		print = 1;
		strcat (outstr, peakStr);
	      }
	    }
	  }

	  /* Print the max value
	   */
	  if (PrintMax)
	  {
	    if (printvalue > vip->Max)
               vip->Max = printvalue;
            if (DeltaTime)
                 sprintf (buf, "\t(Max: %.2f)", vip->Max);
            else sprintf (buf, "\t(Max: %.0f)", vip->Max);
	    strcat (outstr, buf);
	  }

	  if (print)
	  {
	    if (fileout)
	    {
              char prefix[128];
              sprintf (prefix, "%s-%s", gateway, vip->descriptor);
              log (prefix, outstr+1);
	    }
	    else
	    {
              printf ("%s\n", outstr+1);
	      fflush (stdout);
	    }
	  }
	}
      }
      else
      {
	if (response->errstat == SNMP_ERR_TOOBIG)
	{
          if (response->errindex <= varbindsPerPacket &&
              response->errindex > 0)
            varbindsPerPacket = response->errindex - 1;
          else
	  {
	    if (varbindsPerPacket > 30)
                 varbindsPerPacket -= 5;
            else varbindsPerPacket--;
	  }
	  if (varbindsPerPacket <= 0)
             return (5);
	  end = last_end;
	  continue;
	}
	else if (response->errstat == SNMP_ERR_NOSUCHNAME)
	{
	  printf ("This object doesn't exist on %s: ", gateway);
	  for (count = 1, vars = response->variables;
	       vars && count != response->errindex;
	       vars = vars->next_variable, count++)
              ;
	  if (vars)
             print_oid_nums (vars->name, vars->name_length);
	  printf ("\n");
	}
	else
	{
	  printf ("Error in packet.\nReason: %s\n",
		  snmp_errstring (response->errstat));
	}
        return (1);
	if ((pdu = snmp_fix_pdu (response, SNMP_PDU_GET)) != NULL)
           goto retry;
      }
    }
    else if (status == STAT_TIMEOUT)
    {
      printf ("No Response from %s\n", gateway);
      response = 0;
    }
    else
    {				/* status == STAT_ERROR */
      printf ("An error occurred, Quitting\n");
      response = 0;
      break;
    }

    if (response)
       snmp_free_pdu (response);
    if (end == current_name)
       period_WaitForPeriod();
  }
  snmp_close (ss);
  return (0);
}
