/**********************************************************************
 *
 *           Copyright 1998 by Carnegie Mellon University
 * 
 *                       All Rights Reserved
 * 
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose and without fee is hereby granted,
 * provided that the above copyright notice appear in all copies and that
 * both that copyright notice and this permission notice appear in
 * supporting documentation, and that the name of CMU not be
 * used in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.
 * 
 * CMU DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING
 * ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL
 * CMU BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR
 * ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
 * ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
 * SOFTWARE.
 * 
 * $Id: snmpgetnext.c,v 1.11 1998/05/07 21:54:23 ryan Exp $
 * 
 **********************************************************************/

/*
 * snmpgetnext.c - send an snmp GETNEXT request to a network entity.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <snmp.h>

#include "version.h"

void Version (void)
{
  fprintf (stderr,
           "snmpgetnext, version %s\n"
           "using CMU SNMP Library, version %s\n",
           snmpapps_Version(), snmp_Version());
  exit (1);
}

void Usage (void)
{
  fprintf (stderr, "usage: snmpgetnext hostname community oid [oid ...]\n");
  fprintf (stderr, "\n");
  fprintf (stderr, "Args are:\n");
  fprintf (stderr, "  -d     Dump packets being sent / received\n");
  fprintf (stderr, "  -V     Print version information\n");
  fprintf (stderr, "  -v     Verbose output\n");
  fprintf (stderr, "  -p P   Set port to P\n");
  fprintf (stderr, "  -r R   Set number of retries to R (Default: 4)\n");
  fprintf (stderr, "  -t T   Set timeout to T microseconds (Default: 1000000)\n");
  fprintf (stderr, "  -1     Send SNMPv1 packet (Default)\n");
  fprintf (stderr, "  -2     Send SNMPv2 packet\n");

  exit (1);
}

int main (int argc, char **argv)
{
  struct snmp_session session, *ss;
  struct snmp_pdu *pdu, *response;
  struct variable_list *vars;
  char  *gateway = NULL;
  char  *community = NULL;
  int    count, current_name = 0;
  char  *names[128];
  oid    name[MAX_NAME_LEN];
  int    name_length;
  int    ch, status;
  int    verbose = 0;
  int    quiet   = 0;
  int    Port    = SNMP_DEFAULT_REMPORT;
  int    Retries = SNMP_DEFAULT_RETRIES;
  int    Timeout = SNMP_DEFAULT_TIMEOUT;
  int    SNMPVersion = SNMP_VERSION_1;

  /*
   * usage: snmpgetnext gateway-name community-name object-id-list
   */

  while ((ch = getopt(argc,argv, "?dp:vqV12")) != EOF)
      switch (ch)
      {
        case 'd':
	     snmp_dump_packet (1);
             dbug_init();
	     break;
        case 'p':
             Port = atoi (optarg);
	     break;
        case 'v':
	     verbose = 1;
	     break;
        case 'q':
	     quiet = 1;
	     break;
        case 'V':
             Version();
	     break;
        case '1':
	     SNMPVersion = SNMP_VERSION_1;
	     break;
        case '2':
	     SNMPVersion = SNMP_VERSION_2;
	     break;
        default:
             Usage();
      }

  argc -= optind;
  argv += optind;

  gateway   = *argv++;
  community = *argv++;
  argc -= 2;

  while (argc > 0 && *argv)
      names[current_name++] = *argv++;

  if (!gateway || !community || !current_name)
     Usage();

  init_mib();
  sock_init();

  memset (&session,0,sizeof(session));

  session.peername      = gateway;
  session.Version       = SNMPVersion;
  session.community     = (u_char*) community;
  session.community_len = strlen (community);
  session.retries       = Retries;
  session.timeout       = Timeout;
  session.remote_port   = Port;

  snmp_synch_setup (&session);
  ss = snmp_open (&session);
  if (ss == NULL)
  {
    printf ("Couldn't open snmp\n");
    return (-1);
  }

  pdu = snmp_pdu_create (SNMP_PDU_GETNEXT);

  for (count = 0; count < current_name; count++)
  {
    name_length = MAX_NAME_LEN;
    if (!read_objid (names[count], name, &name_length))
       printf ("Invalid object identifier: %s\n", names[count]);
    snmp_add_null_var (pdu, name, name_length);
  }

  /* Send the packet, and read the response */

retry:
  status = snmp_synch_response (ss, pdu, &response);

  if (status == STAT_TIMEOUT)
  {
    printf ("No Response from %s\n", gateway);
    return (1);
  }
  else if (status != STAT_SUCCESS)
  {
    printf ("An error occurred, Quitting\n");
    return (2);
  }

  /* Received a response.  Let's look at it. */
  if (response->errstat != SNMP_ERR_NOERROR)
  {
    /* Error with the packet. */
    fprintf (stderr, "Error in packet.\n");
    fprintf (stderr, "Reason: %s\n", snmp_errstring (response->errstat));

    /* SNMPv1 Compatibility: interpret the basics */
    if (response->errstat == SNMP_ERR_NOSUCHNAME)
    {
      /* Find the variable specified by errindex */
      for (count = 1, vars = response->variables;
	   vars && count != response->errindex;
	   vars = vars->next_variable, count++)
          ;

      if (vars)
      {
	fprintf (stderr, "- ");
	for (count = 0; count < vars->name_length; count++)
	  fprintf (stderr, ".%u", (unsigned char) vars->name[count]);
	fprintf (stderr, "\n");
      }

      /* Remove the bad entry, and try again. */
      pdu = snmp_fix_pdu (response, SNMP_PDU_GET);
      snmp_free_pdu (response);

      if (pdu)
         goto retry;
    }
  }
  else
  {
    /* Successfully read a response with good data in it. */

    for (vars = response->variables; vars; vars = vars->next_variable)
    {
      if (vars->type == SMI_NOSUCHOBJECT)
      {
	fprintf (stderr, "No such object: ");
	for (count = 0; count < vars->name_length; count++)
            fprintf (stderr, ".%u", (unsigned char) vars->name[count]);
      }
      else if (vars->type == SMI_NOSUCHINSTANCE)
      {
	fprintf (stderr, "No such instance: ");
	for (count = 0; count < vars->name_length; count++)
            fprintf (stderr, ".%u", (unsigned char) vars->name[count]);
      }
      else if (vars->type == SMI_ENDOFMIBVIEW)
      {
	fprintf (stderr, "End of MIB view.  Requested OID: ");
	for (count = 0; count < vars->name_length; count++)
            fprintf (stderr, ".%u", (unsigned char) vars->name[count]);
      }
      else    /* Good response, and it exists.  Print it. */
      {
	if (verbose)
	{
	  printf ("Name:          ");
	  print_objid (vars->name, vars->name_length);
	  printf ("Returned Type: ");
	  print_type (vars);
	  printf ("\n");
	  printf ("Value:         ");
	  print_variable_list_value (vars);
	  printf ("\n");
	}
	else
	{
	  if (quiet)
	  {
	    print_variable_list_value (vars);
	    printf ("\n");
	  }
	  else
            print_variable_list (vars);
        }
      }				/* End of printing good variable */
    }				/* End of var loop */
  }				/* End of good response */

  if (response)
     snmp_free_pdu (response);
  snmp_close (ss);
  return (0);
}
