/**********************************************************************
 *
 *           Copyright 1998 by Carnegie Mellon University
 * 
 *                       All Rights Reserved
 * 
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose and without fee is hereby granted,
 * provided that the above copyright notice appear in all copies and that
 * both that copyright notice and this permission notice appear in
 * supporting documentation, and that the name of CMU not be
 * used in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.
 * 
 * CMU DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING
 * ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL
 * CMU BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR
 * ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
 * ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
 * SOFTWARE.
 * 
 **********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <errno.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <snmp.h>

#include "version.h"

void Version (void)
{
  fprintf (stderr,
           "snmpget, version %s\n"
           "using CMU SNMP Library, version %s\n",
           snmpapps_Version(), snmp_Version());
  exit (1);
}

int ascii_to_binary (u_char *cp, u_char *bufp)
{
  int	subidentifier;
  u_char *bp = bufp;

  for (; *cp != '\0'; cp++)
  {
    if (isspace(*cp))
       continue;
    if (!isdigit(*cp))
    {
      fprintf(stderr, "Input error\n");
      return (-1);
    }
    subidentifier = atoi((char *)cp);
    if (subidentifier > 255)
    {
      fprintf(stderr, "subidentifier %d is too large ( > 255)\n", subidentifier);
      return (-1);
    }
    *bp++ = (u_char)subidentifier;
    while(isdigit(*cp))
      cp++;
    cp--;
  }
  return (bp - bufp);
}

int hex_to_binary (u_char *cp, u_char *bufp)
{
  int	subidentifier;
  u_char *bp = bufp;

  for (; *cp != '\0'; cp++)
  {
    if (isspace(*cp))
       continue;
    if (!isxdigit(*cp))
    {
      fprintf(stderr, "Input error\n");
      return (-1);
    }
    sscanf((char *)cp, "%x", &subidentifier);
    if (subidentifier > 255)
    {
      fprintf(stderr, "subidentifier %d is too large ( > 255)\n", subidentifier);
      return (-1);
    }
    *bp++ = (u_char)subidentifier;
    while(isxdigit(*cp))
      cp++;
    cp--;
  }
  return (bp - bufp);
}

void Usage (void)
{
  fprintf (stderr,
           "usage: snmpset [-d] [-p port] gateway-name community-name "
           "oid type value\n");
  exit(1);
}

int main (int argc, char **argv)
{
  struct variable_list *vp, *vars;
  struct snmp_pdu      *pdu, *response;
  struct snmp_session   session, *ss;

  char *gateway   = NULL;
  char *community = NULL;
  char *name      = NULL;
  char *type      = NULL;
  char *newvalue  = NULL;
  oid   value[256];     /* Requested OID * */
  int   ch, status, count;
  int   port    = SNMP_DEFAULT_REMPORT;
  int   retries = SNMP_DEFAULT_RETRIES;
  int   timeout = SNMP_DEFAULT_TIMEOUT;

  /*
   * usage: snmpset [-d] gateway-name community-name oid type value
   */
  while ((ch = getopt(argc,argv, "?d")) != EOF)
    switch (ch)
    {
      case 'd':
           snmp_dump_packet(1);
           dbug_init();
           break;
      case 'p':
           port = atoi (optarg);
           break;
      case 'r':
           retries = atoi (optarg);
           break;
      case 't':
           timeout = atoi (optarg);
           break;
      case '?':
      default:
           Usage();
           break;
    }

  argc -= optind;
  argv += optind;
  if (argc < 4)
     Usage();

  gateway   = *argv++;
  community = *argv++;
  type      = *argv++;
  newvalue  = *argv++;

  /* Initialize the MIB
   */
  init_mib();
  sock_init();

  memset (&session,0,sizeof(struct snmp_session));
  session.peername      = gateway;
  session.community     = (u_char*)community;
  session.community_len = strlen (community);
  session.retries       = retries;
  session.timeout       = timeout;
  session.remote_port   = port;

  /* Initialize the variable list
   */
  vp = malloc (sizeof(*vp));
  vp->next_variable = NULL;
  vp->name = NULL;
  vp->val.string = NULL;

  /* Did the user request a usable OID?
   */
  vp->name_length = MAX_NAME_LEN;
  vp->name = malloc (sizeof(oid) * vp->name_length);
  if (!read_objid(name, vp->name, &vp->name_length))
     return (-1);

  /* Set the OID type
   */
  switch (type[0])
  {
    case 'i':
         vp->type           = INTEGER;
         vp->val.integer    = malloc (sizeof(int));
         *(vp->val.integer) = atoi (newvalue);
         vp->val_len        = sizeof (int);
         break;
    case 's':
         vp->type = STRING;
         strcpy ((char*)value, newvalue);
         vp->val_len = strlen (newvalue);
         vp->val.string = malloc (vp->val_len);
         memcpy (vp->val.string, &value, vp->val_len);
         break;
    case 'x':
         vp->type       = STRING;
         vp->val_len    = hex_to_binary ((u_char*)newvalue, (u_char*)value);
         vp->val.string = malloc (vp->val_len);
         memcpy (vp->val.string, value, vp->val_len);
         break;
    case 'd':
         vp->type       = STRING;
         vp->val_len    = ascii_to_binary ((u_char*)newvalue, (u_char*)value);
         vp->val.string = malloc (vp->val_len);
         memcpy (vp->val.string, value, vp->val_len);
         break;
    case 'n':
         vp->type       = NULLOBJ;
         vp->val_len    = 0;
         vp->val.string = NULL;
         break;
    case 'o':
         vp->type      = OBJID;
         vp->val_len   = MAX_NAME_LEN;;
         vp->val.objid = malloc (sizeof(oid) * vp->val_len);
         read_objid (newvalue, vp->val.objid, &(vp->val_len));
         vp->val_len *= sizeof (oid);
         break;
    case 't':
         vp->type           = SMI_TIMETICKS;
         vp->val.integer    = malloc (sizeof(int));
         *(vp->val.integer) = atoi (newvalue);
         vp->val_len        = sizeof (int);
         break;
    case 'a':
         vp->type           = SMI_IPADDRESS;
         vp->val.integer    = malloc (sizeof(int));
         *(vp->val.integer) = inet_addr (newvalue);
         vp->val_len        = sizeof (int);
         break;
    default:
         fprintf (stderr,
                  "bad type \"%c\", use \"i\", \"s\", \"x\", \"d\", "
                  "\"n\", \"o\", \"t\", or \"a\".\n", type[0]);
         return (-1);
  }

  /* Setup Session
   */
  snmp_synch_setup (&session);
  ss = snmp_open (&session);
  if (ss == NULL)
  {
    fprintf (stderr, "Unable to establish connection to %s\n", gateway);
    return (-1);
  }

  /* Create this PDU
   */
  pdu = snmp_pdu_create (SNMP_PDU_SET);
  pdu->variables = vp;

  /* Send it
   */
  status = snmp_synch_response (ss, pdu, &response);

  if (status == STAT_SUCCESS)
  {
    switch (response->command)
    {
      case SNMP_PDU_GET:
           printf("Received GET REQUEST ");
           break;
      case SNMP_PDU_GETNEXT:
           printf ("Received GETNEXT REQUEST ");
           break;
      case SNMP_PDU_RESPONSE:
           printf ("Received GET RESPONSE ");
           break;
      case SNMP_PDU_SET:
           printf ("Received SET REQUEST ");
           break;
      case TRP_REQ_MSG:
           printf ("Received TRAP REQUEST ");
           break;
    }
    printf ("from %s\n", inet_ntoa(response->address.sin_addr));
    printf ("requestid 0x%x errstat 0x%x errindex 0x%x\n",
            (unsigned int)response->reqid, (unsigned int)response->errstat,
            (unsigned int)response->errindex);
    if (response->errstat == SNMP_ERR_NOERROR)
    {
      for (vars = response->variables; vars; vars = vars->next_variable)
          print_variable_list (vars);
    }
    else
    {
      fprintf (stderr, "Error in packet.\nReason: %s\n",
	      snmp_errstring(response->errstat));

      if (response->errstat == SNMP_ERR_NOSUCHNAME)
      {
        for (count = 1, vars = response->variables;
             vars && count != response->errindex;
             vars = vars->next_variable, count++)
	    ;
        if (vars)
        {
          printf ("This name doesn't exist: ");
          print_oid_nums(vars->name, vars->name_length);
        }
        printf ("\n");
      }
    }
  }
  else if (status == STAT_TIMEOUT)
  {
    fprintf (stderr, "No Response from %s\n", gateway);
  }
  else     /* status == STAT_ERROR */
  {
    fprintf (stderr, "An error occurred, Quitting\n");
  }

  if (response)
      snmp_free_pdu (response);
  return (0);
}
