/*
 * Receive SNMP V1 and V2 traps
 */

/**********************************************************************
 *
 *           Copyright 1998 by Carnegie Mellon University
 * 
 *                       All Rights Reserved
 * 
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose and without fee is hereby granted,
 * provided that the above copyright notice appear in all copies and that
 * both that copyright notice and this permission notice appear in
 * supporting documentation, and that the name of CMU not be
 * used in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.
 * 
 * CMU DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING
 * ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL
 * CMU BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR
 * ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
 * ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
 * SOFTWARE.
 * 
 **********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <errno.h>
#include <snmp.h>
#include <netdb.h>
#include <sys/socket.h>
#include <sys/syslog.h>
#include <arpa/inet.h>

#include "version.h"

static int Print = 1;
static int Syslog = 0;

char *progname = "snmptrap";

#define PACKET_LENGTH 4500

/*
 * Return the description of this SNMPv1 trap
 */
char *trap_description (int trap)
{
  switch (trap)
  {
    case SNMP_TRAP_COLDSTART:
         return ("Cold Start");
    case SNMP_TRAP_WARMSTART:
         return ("Warm Start");
    case SNMP_TRAP_LINKDOWN:
         return ("Link Down");
    case SNMP_TRAP_LINKUP:
         return ("Link Up");
    case SNMP_TRAP_AUTHENTICATIONFAILURE:
         return ("Authentication Failure");
    case SNMP_TRAP_EGPNEIGHBORLOSS:
         return ("EGP Neighbor Loss");
    case SNMP_TRAP_ENTERPRISESPECIFIC:
         return ("Enterprise Specific");
    default:
         return ("Unknown Type");
  }
}

/*
 * Read a trap from the FD
 */
void read_trap (int fd)
{
  struct variable_list *vars;
  struct snmp_pdu      *pdu;
  struct sockaddr_in   from;
  struct snmp_session  session;
  char   buf   [PACKET_LENGTH];
  u_char packet[PACKET_LENGTH];
  int    fromlen = sizeof (from);
  int    len;
  u_char *community;

  len = recvfrom (fd,packet,PACKET_LENGTH,0,(struct sockaddr*)&from,&fromlen);
  if (len < 0)
  {
    perror ("recvfrom");
    return;
  }

  pdu = snmp_pdu_create (0);
  if (pdu == NULL)
  {
    perror ("malloc");
    return;
  }
  pdu->address = from;
  pdu->reqid   = 0;

  /* Silly hack to get snmp_parse to work.
   */
  community = snmp_parse (&session, pdu, packet, len);
  if (!community)
  {
    fprintf (stderr, "Mangled packet!\n");
    snmp_free_pdu (pdu);
    return;
  }
  free (community);

  /* Now that we have a completed PDU, see what it is
   */
  if (pdu->command == TRP_REQ_MSG)
  {
    if (Print)
    {
      printf ("***** BEGIN RECEIVED V1 TRAP *****\n");
      printf ("%s: %s Trap (%d) Uptime: %s\n",
	      inet_ntoa (pdu->agent_addr.sin_addr),
	      trap_description (pdu->trap_type),
	      pdu->specific_type,
	      uptime_string (pdu->time, buf));
      for (vars = pdu->variables; vars; vars = vars->next_variable)
          print_variable_list (vars);

      printf ("***** END OF RECEIVED V1 TRAP *****\n");
    }
  }
  else if (pdu->command == SNMP_PDU_V2TRAP)
  {
    if (Print)
    {
      printf ("***** BEGIN RECEIVED V2 TRAP *****\n");
      printf ("%s sent the following data:\n",
	      inet_ntoa (pdu->address.sin_addr));
      for (vars = pdu->variables; vars; vars = vars->next_variable)
          print_variable_list (vars);

      printf ("***** END RECEIVED V2 TRAP *****\n");
    }
  }
  else
    fprintf (stderr, "Weird!  Received PDU %d", pdu->command);

  snmp_free_pdu (pdu);
}


void init_syslog (void)
{
  /*
   * All messages will be logged to the local0 facility and will be sent to
   * the console if syslog doesn't work.
   */
  openlog ("snmptrapd", LOG_NDELAY, LOG_LOCAL0);
  syslog (LOG_INFO, "Starting snmptrapd");
}

void Version (void)
{
  fprintf (stderr,
           "snmptrapd, version %s\n"
           "\tUsing CMU SNMP Library, version %s\n",
           snmpapps_Version(), snmp_Version());
}

void Usage (void)
{
  fprintf (stderr,
           "Usage: snmptrapd [-P] [-V] [-S] [-d] [-p port]\n\n"
           "Args are:\n"
           "  -d     Dump packets being sent / received\n"
           "  -p P   Set port to P\n"
           "  -V     Print version information\n"
           "  -S     Write received traps to syslog\n");
  exit (1);
}

int main (int argc, char **argv)
{
  int    sock;
  fd_set fdset;
  int    ch, count, port = 0;
  int    quit = 0;
  struct sockaddr_in ME;

  /* usage: snmptrapd [-p] [-s] [-d]
   */
  while ((ch = getopt(argc,argv, "?dPSp:V")) != EOF)
      switch (ch)
      {
        case 'd':
	     snmp_dump_packet (1);
             dbug_init();
             break;
        case 'P':
	     Print++;
	     break;
        case 'S':
	     Syslog++;
	     break;
        case 'p':
             port = atoi (optarg);
	     break;
        case 'V':
             Version();
	     break;
        case '?':
        default:
             Usage();
	     break;
      }

  init_mib();
  sock_init();
  init_syslog();

  if (port == 0)
  {
    struct servent *sp = getservbyname ("snmp-trap", "udp");
    port = sp ? sp->s_port : htons (162);
  }

  /* Make stdout unbuffered so we can send it to a file and still be able to
   * see the messages in a timely fashion.
   */
  setbuf (stdout, NULL);

  if ((sock = socket(AF_INET, SOCK_DGRAM, 0)) < 0)
  {
    perror ("socket");
    return (-1);
  }

  ME.sin_family      = AF_INET;
  ME.sin_addr.s_addr = INADDR_ANY;
  ME.sin_port        = port;

  if (bind(sock,(struct sockaddr*)&ME,sizeof(ME)) < 0)
  {
    perror ("bind");
    close_s (sock);
    return (-1);
  }          

  while (!quit)
  {
    FD_ZERO (&fdset);
    FD_SET (sock, &fdset);
    if ((count = select_s (sock+1,&fdset,0,0,NULL)) > 0)
    {
      if (FD_ISSET (sock, &fdset))
         read_trap (sock);
      else
      {
        fprintf (stderr,
                 "Weird! select returned something other than what\n"
                 "we're waiting for! (%d)\n", count);
      }
    }
    else  /* select() returned <= 0 */
    {
      switch (count)
      {
        case 0:      /* Timeout! */
	     break;
        case -1:
	     if (errno == EINTR)
                  quit = 1;
             else perror ("select");
             quit = 1;
             break;
        default:
	     printf ("select returned %d\n", count);
             quit = 1;
             break;
      }
    }
  }
  close_s (sock);
  syslog (LOG_INFO, "Stopping snmptrapd");
  return (0);
}
