/***********************************************************
	Copyright 1988, 1989, 1990 by Carnegie Mellon University
	Copyright 1989	TGV, Incorporated

		      All Rights Reserved

Permission to use, copy, modify, and distribute this software and its
documentation for any purpose and without fee is hereby granted,
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in
supporting documentation, and that the name of CMU and TGV not be used
in advertising or publicity pertaining to distribution of the software
without specific, written prior permission.

CMU AND TGV DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
EVENT SHALL CMU OR TGV BE LIABLE FOR ANY SPECIAL, INDIRECT OR
CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF
USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
PERFORMANCE OF THIS SOFTWARE.
******************************************************************/

/*
 * snmp_vars.c - return a pointer to the named variable.
 *
 */

/*
 * additions, fixes and enhancements for Linux by 
 * Erik Schoenfelder <schoenfr@gaertner.de> and
 * Juergen Schoenwaelder <schoenw@cs.utwente.nl>
 * June 1996
 */

/*
 * Small Linux specific patch by Patrick Weemeeuw
 * (patrick.weemeeuw@kulnet.kuleuven.ac.be)
 * to keep track of HrProcessorLoad more accurately.
 * -- 8 July 1996
 */

#define USE_NAME_AS_DESCRIPTION	/*"se0" instead of text */
#define GATEWAY			/* MultiNet is always configured this way! */

#include <stdio.h>
#include <string.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <sys/syslog.h>
#include <sys/ioctl.h>
#include <net/if.h>
#include <net/route.h>
#include <netinet/in_pcb.h>
#include <netinet/if_ether.h>
#include <netinet/in_systm.h>
#include <snmp.h>


int snmp_enableauthentraps = 2;	/* default: 2 == disabled */

#ifndef  MIN
#define  MIN(a,b) (((a) < (b)) ? (a) : (b))
#endif

#define PROCESSSLOTINDEX  0
#define PROCESSID         4
#define PROCESSCOMMAND    8


extern int swap, mem;
extern char *Lookup_Device_Annotation();

static void ARP_Scan_Init();
static int  ARP_Scan_Next();

extern int  writeVersion(), writeSystem();
extern void Interface_Scan_Init();
extern int  Interface_Scan_Next();

static int  Interface_Scan_Get_Count();
static int  Interface_Scan_By_Index();
static int  Interface_Get_Ether_By_Index();
static int  TCP_Count_Connections();
static void TCP_Scan_Init();
static int  TCP_Scan_Next();
static int  compare_tree();

#define  KNLookup(nl_which, buf, s)  \
         (klookup((int) nl[nl_which].n_value, buf, s))


static struct nlist nl[] = {

#define N_IPSTAT	0
  {"_ipstat"},
#define N_IPFORWARDING	1
  {"_ipforwarding"},
#define N_TCP_TTL	2
  {"_tcp_ttl"},
#define N_UDPSTAT	3
  {"_udpstat"},
#define N_IN_INTERFACES 4
  {"_in_interfaces"},
#define N_ICMPSTAT	5
  {"_icmpstat"},
#define N_IFNET		6
  {"_ifnet"},
#define N_TCPSTAT	7
  {"_tcpstat"},
#define N_TCB		8
  {"_tcb"},
#define N_ARPTAB_SIZE	9
  {"_arptab_size"},
#define N_ARPTAB        10
  {"_arptab"},
#define N_IN_IFADDR     11
  {"_in_ifaddr"},
#define N_BOOTTIME	12
  {"_boottime"},
#define N_PROC		13
  {"_proc"},
#define N_NPROC		14
  {"_nproc"},
#define N_DMMIN		15
  {"_dmmin"},
#define N_DMMAX		16
  {"_dmmax"},
#define N_NSWAP		17
  {"_nswap"},
#define N_USRPTMAP	18
  {"_Usrptmap"},
#define N_USRPT		19
  {"_usrpt"},
  0,
};



/*
 *    Each variable name is placed in the variable table, without the
 * terminating substring that determines the instance of the variable.  When
 * a string is found that is lexicographicly preceded by the input string,
 * the function for that entry is called to find the method of access of the
 * instance of the named variable.  If that variable is not found, NULL is
 * returned, and the search through the table continues (it will probably
 * stop at the next entry).  If it is found, the function returns a character
 * pointer and a length or a function pointer.  The former is the address
 * of the operand, the latter is a write routine for the variable.
 *
 * u_char *
 * findVar(name, length, exact, var_len, write_method)
 * oid      *name;          IN/OUT - input name requested, output name found
 * int      length;         IN/OUT - number of sub-ids in the in and out oid's
 * int      exact;          IN - TRUE if an exact match was requested.
 * int      len;            OUT - length of variable or 0 if function returned.
 * int      write_method;   OUT - pointer to function to set variable,
 *                                otherwise 0
 *
 *     The writeVar function is returned to handle row addition or complex
 * writes that require boundary checking or executing an action.
 * This routine will be called three times for each varbind in the packet.
 * The first time for each varbind, action is set to RESERVE1.  The type
 * and value should be checked during this pass.  If any other variables
 * in the MIB depend on this variable, this variable will be stored away
 * (but *not* committed!) in a place where it can be found by a call to
 * writeVar for a dependent variable, even in the same PDU.  During
 * the second pass, action is set to RESERVE2.  If this variable is dependent
 * on any other variables, it will check them now.  It must check to see
 * if any non-committed values have been stored for variables in the same
 * PDU that it depends on.  Sometimes resources will need to be reserved
 * in the first two passes to guarantee that the operation can proceed
 * during the third pass.  During the third pass, if there were no errors
 * in the first two passes, writeVar is called for every varbind with action
 * set to COMMIT.  It is now that the values should be written.  If there
 * were errors during the first two passes, writeVar is called in the third
 * pass once for each varbind, with the action set to FREE.  An opportunity
 * is thus provided to free those resources reserved in the first two passes.
 * 
 * writeVar(action, var_val, var_val_type, var_val_len, statP, name, name_len)
 * int      action;         IN - RESERVE1, RESERVE2, COMMIT, or FREE
 * u_char   *var_val;       IN - input or output buffer space
 * u_char   var_val_type;   IN - type of input buffer
 * int      var_val_len;    IN - input and output buffer len
 * u_char   *statP;         IN - pointer to local statistic
 * oid      *name           IN - pointer to name requested
 * int      name_len        IN - number of sub-ids in the name
 */

long   long_return;
u_char return_buf [CLSIZE * NBPG];


char version_descr[256] = "DOS WatTCP";

void init_snmp (void)
{
  nlist ("/vmunix", nl);
  init_kmem ("/dev/kmem");
  init_routes();
  {               /* use real hostname for sysname, if not configured: */
    extern char sysName[256];
    char tmp[256];

    if (!*sysName)
      {
        if (!gethostname (tmp,sizeof(tmp)))
             strcpy (sysName, tmp);
        else strcpy (sysName, "Unknown");
      }
  }
  {   /* use authentrap from config-file, if found: */
    extern int conf_authentraps;
    if (conf_authentraps > 0)
      snmp_enableauthentraps = conf_authentraps;
  }
}

#define CMUMIB                  1, 3, 6, 1, 4, 1, 3
#define CMUUNIXMIB              CMUMIB, 2, 2

#define SNMPMODULES 		1, 3, 6, 1, 6, 3

#define SNMPSTATS		SNMPMODULES, 1, 1, 1
#define SNMPV1STATS		SNMPMODULES, 1, 1, 2
#define SNMPTRAP		SNMPMODULES, 1, 1, 4
#define SNMPSET			SNMPMODULES, 1, 1, 6
#define USECMIBOBJ		SNMPMODULES, 6, 1
#define USECAGENT		USECMIBOBJ, 1
#define USECSTATS		USECMIBOBJ, 2

#define HOST                    RMONMIB, 4
#define HOSTCONTROL             HOST, 1, 1	/* hostControlEntry */
#define HOSTTAB                 HOST, 2, 1	/* hostEntry */
#define HOSTTIMETAB             HOST, 3, 1	/* hostTimeEntry */
#define HOSTTOPN                RMONMIB, 5
#define HOSTTOPNCONTROL         HOSTTOPN, 1, 1  /* hostTopNControlEntry */
#define HOSTTOPNTAB             HOSTTOPN, 2, 1	/* hostTopNEntry */
#define HOSTTIMETABADDRESS        1
#define HOSTTIMETABCREATIONORDER  2
#define HOSTTIMETABINDEX          3
#define HOSTTIMETABINPKTS         4
#define HOSTTIMETABOUTPKTS        5
#define HOSTTIMETABINOCTETS       6
#define HOSTTIMETABOUTOCTETS      7
#define HOSTTIMETABOUTERRORS      8
#define HOSTTIMETABOUTBCASTPKTS   9
#define HOSTTIMETABOUTMCASTPKTS   10

/* various OIDs that are needed throughout the agent */

oid sysUpTimeOid[]  = { 1, 3, 6, 1, 2, 1, 1, 3, 0 };
int sysUpTimeOidLen = sizeof (sysUpTimeOid) / sizeof (oid);


/*
 * The subtree structure contains a subtree prefix which applies to
 * all variables in the associated variable list.
 * No subtree may be a subtree of another subtree in this list.  i.e.:
 * 1.2
 * 1.2.0
 */
struct subtree {
       oid    name[16];              /* objid prefix of subtree */
       u_char namelen;               /* number of subid's in name above */
       struct variable *variables;   /* pointer to variables array */
       int    variables_len;         /* number of entries in above array */
       int    variables_width;       /* sizeof each variable entry */
     };

/*
 * This is a new variable structure that doesn't have as much memory
 * tied up in the object identifier.  It's elements have also been re-arranged
 * so that the name field can be variable length.  Any number of these
 * structures can be created with lengths tailor made to a particular
 * application.  The first 5 elements of the structure must remain constant.
 */
struct variable2 {
    u_char  magic;              /* passed to function as a hint */
    char    type;               /* type of variable */
    u_short acl;		/* access control list for variable */
    u_char *(*findVar)();       /* function that finds variable */
    u_char  namelen;            /* length of name below */
    oid     name[2];            /* object identifier of variable */
  };

struct variable4 {
    u_char  magic;              /* passed to function as a hint */
    char    type;               /* type of variable */
    u_short acl;		/* access control list for variable */
    u_char *(*findVar)();       /* function that finds variable */
    u_char  namelen;            /* length of name below */
    oid     name[4];            /* object identifier of variable */
  };

struct variable7 {
    u_char  magic;              /* passed to function as a hint */
    char    type;               /* type of variable */
    u_short acl;		/* access control list for variable */
    u_char *(*findVar)();       /* function that finds variable */
    u_char  namelen;            /* length of name below */
    oid     name[7];            /* object identifier of variable */
  };

struct variable13 {
    u_char  magic;              /* passed to function as a hint */
    char    type;               /* type of variable */
    u_short acl;		/* access control list for variable */
    u_char *(*findVar)();       /* function that finds variable */
    u_char  namelen;            /* length of name below */
    oid     name[13];           /* object identifier of variable */
  };

/*
 * ##############################################################
 * IMPORTANT NOTE:
 * ##############################################################
 *
 * The format of the acl word in these entries has changed.  It is still
 * 2 bits per community, offset from the right by the index of the community.
 * The leftmost two bits denotes read access, and the rightmost denotes
 * write access.
 * The change is that the rightmost two bits are now reserved for the object's
 * max-access.  This is the minimum of what makes "protocol sense" for the
 * object and whether set support was implemented for that object.
 * These two bits will not map to any community.  The first community
 * entry will map to the 3rd and 4th bits.
 */

#define MTRBIGNUMBER	1
#define MTRNSAPADDRESS	2
#define MTRBITSTRING	3

struct variable2 demo_variables[] = {
  { MTRBIGNUMBER,   COUNTER64, RONLY, var_demo, 1, { 1 } },
  { MTRNSAPADDRESS, NSAP,      RONLY, var_demo, 1, { 2 } },
};

struct variable4 interface_variables[] = {
  { IFNUMBER,          INTEGER,   RONLY, var_system,  1, {1      }},
  { IFINDEX,           INTEGER,   RONLY, var_ifEntry, 3, {2, 1, 1}},
  { IFDESCR,           STRING,    RONLY, var_ifEntry, 3, {2, 1, 2}},
  { IFTYPE,            INTEGER,   RONLY, var_ifEntry, 3, {2, 1, 3}},
  { IFMTU,             INTEGER,   RONLY, var_ifEntry, 3, {2, 1, 4}},
  { IFSPEED,           GAUGE,     RONLY, var_ifEntry, 3, {2, 1, 5}},
  { IFPHYSADDRESS,     STRING,    RONLY, var_ifEntry, 3, {2, 1, 6}},
  { IFADMINSTATUS,     INTEGER,   RWRITE,var_ifEntry, 3, {2, 1, 7}},
  { IFOPERSTATUS,      INTEGER,   RONLY, var_ifEntry, 3, {2, 1, 8}},
  { IFLASTCHANGE,      TIMETICKS, RONLY, var_ifEntry, 3, {2, 1, 9}},
  { IFINOCTETS,        COUNTER,   RONLY, var_ifEntry, 3, {2, 1, 10}},
  { IFINUCASTPKTS,     COUNTER,   RONLY, var_ifEntry, 3, {2, 1, 11}},
  { IFINNUCASTPKTS,    COUNTER,   RONLY, var_ifEntry, 3, {2, 1, 12}},
  { IFINDISCARDS,      COUNTER,   RONLY, var_ifEntry, 3, {2, 1, 13}},
  { IFINERRORS,        COUNTER,   RONLY, var_ifEntry, 3, {2, 1, 14}},
  { IFINUNKNOWNPROTOS, COUNTER,   RONLY, var_ifEntry, 3, {2, 1, 15}},
  { IFOUTOCTETS,       COUNTER,   RONLY, var_ifEntry, 3, {2, 1, 16}},
  { IFOUTUCASTPKTS,    COUNTER,   RONLY, var_ifEntry, 3, {2, 1, 17}},
  { IFOUTNUCASTPKTS,   COUNTER,   RONLY, var_ifEntry, 3, {2, 1, 18}},
  { IFOUTDISCARDS,     COUNTER,   RONLY, var_ifEntry, 3, {2, 1, 19}},
  { IFOUTERRORS,       COUNTER,   RONLY, var_ifEntry, 3, {2, 1, 20}},
  { IFOUTQLEN,         GAUGE,     RONLY, var_ifEntry, 3, {2, 1, 21}}
};

struct variable2 system_variables[] = {
  { VERSION_DESCR,  STRING,   RONLY, var_system, 1, {1}},
  { VERSION_ID,     OBJID,    RONLY, var_system, 1, {2}},
  { UPTIME,         TIMETICKS,RONLY, var_system, 1, {3}},
  { SYSCONTACT,     STRING,   RWRITE,var_system, 1, {4}},
  { SYSNAME,        STRING,   RWRITE,var_system, 1, {5}},
  { SYSLOCATION,    STRING,   RWRITE,var_system, 1, {6}},
  { SYSSERVICES,    INTEGER,  RONLY, var_system, 1, {7}},
  { SYSORLASTCHANGE,TIMETICKS,RONLY, var_system, 1, {8}}
};

struct variable2 or_variables[] = {
  {2, OBJID,    RONLY, var_orEntry, 1, {2}},
  {3, STRING,   RONLY, var_orEntry, 1, {3}},
  {4, TIMETICKS,RONLY, var_orEntry, 1, {4}},
};

struct variable2 at_variables[] = {
  {ATIFINDEX,    INTEGER,  RONLY, var_atEntry, 1, {1}},
  {ATPHYSADDRESS,STRING,   RONLY, var_atEntry, 1, {2}},
  {ATNETADDRESS, IPADDRESS,RONLY, var_atEntry, 1, {3}}
};

struct variable4 ip_variables[] = {
  {IPFORWARDING,        INTEGER,  RONLY, var_ip, 1, {1}},
  {IPDEFAULTTTL,        INTEGER,  RONLY, var_ip, 1, {2}},
  {IPINRECEIVES,        COUNTER,  RONLY, var_ip, 1, {3}},
  {IPINHDRERRORS,       COUNTER,  RONLY, var_ip, 1, {4}},
  {IPINADDRERRORS,      COUNTER,  RONLY, var_ip, 1, {5}},
  {IPFORWDATAGRAMS,     COUNTER,  RONLY, var_ip, 1, {6}},
  {IPINUNKNOWNPROTOS,   COUNTER,  RONLY, var_ip, 1, {7}},
  {IPINDISCARDS,        COUNTER,  RONLY, var_ip, 1, {8}},
  {IPINDELIVERS,        COUNTER,  RONLY, var_ip, 1, {9}},
  {IPOUTREQUESTS,       COUNTER,  RONLY, var_ip, 1, {10}},
  {IPOUTDISCARDS,       COUNTER,  RONLY, var_ip, 1, {11}},
  {IPOUTNOROUTES,       COUNTER,  RONLY, var_ip, 1, {12}},
  {IPREASMTIMEOUT,      INTEGER,  RONLY, var_ip, 1, {13}},
  {IPREASMREQDS,        COUNTER,  RONLY, var_ip, 1, {14}},
  {IPREASMOKS,          COUNTER,  RONLY, var_ip, 1, {15}},
  {IPREASMFAILS,        COUNTER,  RONLY, var_ip, 1, {16}},
  {IPFRAGOKS,           COUNTER,  RONLY, var_ip, 1, {17}},
  {IPFRAGFAILS,         COUNTER,  RONLY, var_ip, 1, {18}},
  {IPFRAGCREATES,       COUNTER,  RONLY, var_ip, 1, {19}},
  {IPADADDR,            IPADDRESS,RONLY, var_ipAddrEntry,  3, {20, 1, 1}},
  {IPADIFINDEX,         INTEGER,  RONLY, var_ipAddrEntry,  3, {20, 1, 2}},
  {IPADNETMASK,         IPADDRESS,RONLY, var_ipAddrEntry,  3, {20, 1, 3}},
  {IPADBCASTADDR,       INTEGER,  RONLY, var_ipAddrEntry,  3, {20, 1, 4}},
  {IPADENTREASMMAXSIZE, INTEGER,  RONLY, var_ipAddrEntry,  3, {20, 1, 5}},
  {IPROUTEDEST,         IPADDRESS,RONLY, var_ipRouteEntry, 3, {21, 1, 1}},
  {IPROUTEIFINDEX,      INTEGER,  RONLY, var_ipRouteEntry, 3, {21, 1, 2}},
  {IPROUTEMETRIC1,      INTEGER,  RONLY, var_ipRouteEntry, 3, {21, 1, 3}},
  {IPROUTEMETRIC2,      INTEGER,  RONLY, var_ipRouteEntry, 3, {21, 1, 4}},
  {IPROUTEMETRIC3,      INTEGER,  RONLY, var_ipRouteEntry, 3, {21, 1, 5}},
  {IPROUTEMETRIC4,      INTEGER,  RONLY, var_ipRouteEntry, 3, {21, 1, 6}},
  {IPROUTENEXTHOP,      IPADDRESS,RONLY, var_ipRouteEntry, 3, {21, 1, 7}},
  {IPROUTETYPE,         INTEGER,  RONLY, var_ipRouteEntry, 3, {21, 1, 8}},
  {IPROUTEPROTO,        INTEGER,  RONLY, var_ipRouteEntry, 3, {21, 1, 9}},
  {IPROUTEAGE,          INTEGER,  RONLY, var_ipRouteEntry, 3, {21, 1, 10}}
};

struct variable2 icmp_variables[] = {
  {ICMPINMSGS, COUNTER, RONLY, var_icmp, 1,
   {1}},
  {ICMPINERRORS, COUNTER, RONLY, var_icmp, 1,
   {2}},
  {ICMPINDESTUNREACHS, COUNTER, RONLY, var_icmp, 1,
   {3}},
  {ICMPINTIMEEXCDS, COUNTER, RONLY, var_icmp, 1,
   {4}},
  {ICMPINPARMPROBS, COUNTER, RONLY, var_icmp, 1,
   {5}},
  {ICMPINSRCQUENCHS, COUNTER, RONLY, var_icmp, 1,
   {6}},
  {ICMPINREDIRECTS, COUNTER, RONLY, var_icmp, 1,
   {7}},
  {ICMPINECHOS, COUNTER, RONLY, var_icmp, 1,
   {8}},
  {ICMPINECHOREPS, COUNTER, RONLY, var_icmp, 1,
   {9}},
  {ICMPINTIMESTAMPS, COUNTER, RONLY, var_icmp, 1,
   {10}},
  {ICMPINTIMESTAMPREPS, COUNTER, RONLY, var_icmp, 1,
   {11}},
  {ICMPINADDRMASKS, COUNTER, RONLY, var_icmp, 1,
   {12}},
  {ICMPINADDRMASKREPS, COUNTER, RONLY, var_icmp, 1,
   {13}},
  {ICMPOUTMSGS, COUNTER, RONLY, var_icmp, 1,
   {14}},
  {ICMPOUTERRORS, COUNTER, RONLY, var_icmp, 1,
   {15}},
  {ICMPOUTDESTUNREACHS, COUNTER, RONLY, var_icmp, 1,
   {16}},
  {ICMPOUTTIMEEXCDS, COUNTER, RONLY, var_icmp, 1,
   {17}},
  {ICMPOUTPARMPROBS, COUNTER, RONLY, var_icmp, 1,
   {18}},
  {ICMPOUTSRCQUENCHS, COUNTER, RONLY, var_icmp, 1,
   {19}},
  {ICMPOUTREDIRECTS, COUNTER, RONLY, var_icmp, 1,
   {20}},
  {ICMPOUTECHOS, COUNTER, RONLY, var_icmp, 1,
   {21}},
  {ICMPOUTECHOREPS, COUNTER, RONLY, var_icmp, 1,
   {22}},
  {ICMPOUTTIMESTAMPS, COUNTER, RONLY, var_icmp, 1,
   {23}},
  {ICMPOUTTIMESTAMPREPS, COUNTER, RONLY, var_icmp, 1,
   {24}},
  {ICMPOUTADDRMASKS, COUNTER, RONLY, var_icmp, 1,
   {25}},
  {ICMPOUTADDRMASKREPS, COUNTER, RONLY, var_icmp, 1,
   {26}}
};

struct variable13 tcp_variables[] = {
  {TCPRTOALGORITHM, INTEGER, RONLY, var_tcp, 1,
   {1}},
  {TCPRTOMIN, INTEGER, RONLY, var_tcp, 1,
   {2}},
  {TCPRTOMAX, INTEGER, RONLY, var_tcp, 1,
   {3}},
  {TCPMAXCONN, INTEGER, RONLY, var_tcp, 1,
   {4}},
  {TCPACTIVEOPENS, COUNTER, RONLY, var_tcp, 1,
   {5}},
  {TCPPASSIVEOPENS, COUNTER, RONLY, var_tcp, 1,
   {6}},
  {TCPATTEMPTFAILS, COUNTER, RONLY, var_tcp, 1,
   {7}},
  {TCPESTABRESETS, COUNTER, RONLY, var_tcp, 1,
   {8}},
  {TCPCURRESTAB, GAUGE, RONLY, var_tcp, 1,
   {9}},
  {TCPINSEGS, COUNTER, RONLY, var_tcp, 1,
   {10}},
  {TCPOUTSEGS, COUNTER, RONLY, var_tcp, 1,
   {11}},
  {TCPRETRANSSEGS, COUNTER, RONLY, var_tcp, 1,
   {12}},
  {TCPCONNSTATE, INTEGER, RONLY, var_tcp, 3,
   {13, 1, 1}},
  {TCPCONNLOCALADDRESS, IPADDRESS, RONLY, var_tcp, 3,
   {13, 1, 2}},
  {TCPCONNLOCALPORT, INTEGER, RONLY, var_tcp, 3,
   {13, 1, 3}},
  {TCPCONNREMADDRESS, IPADDRESS, RONLY, var_tcp, 3,
   {13, 1, 4}},
  {TCPCONNREMPORT, INTEGER, RONLY, var_tcp, 3,
   {13, 1, 5}}
};

struct variable2 udp_variables[] = {
  {UDPINDATAGRAMS, COUNTER, RONLY, var_udp, 1,
   {1}},
  {UDPNOPORTS, COUNTER, RONLY, var_udp, 1,
   {2}},
  {UDPINERRORS, COUNTER, RONLY, var_udp, 1,
   {3}},
  {UDPOUTDATAGRAMS, COUNTER, RONLY, var_udp, 1,
   {4}}
};

struct variable2 process_variables[] = {
  {PROCESSSLOTINDEX, INTEGER, RONLY, var_process, 1,
   {1}},
  {PROCESSID, INTEGER, RONLY, var_proces, 1,
   {2}},
  {PROCESSCOMMAND, STRING, RONLY, var_process, 1,
   {3}}
};


struct variable2 snmpstats_variables[] = {
  {1, COUNTER, RONLY, var_snmpStats, 1,
   {1}},
  {3, COUNTER, RONLY, var_snmpStats, 1,
   {3}},
  {11, COUNTER, RONLY, var_snmpStats, 1,
   {11}},
  {12, COUNTER, RONLY, var_snmpStats, 1,
   {12}},
  {13, COUNTER, RONLY, var_snmpStats, 1,
   {13}},
};

struct variable2 snmpv1stats_variables[] = {
  {101, COUNTER, RONLY, var_snmpStats, 1,
   {1}},
  {102, COUNTER, RONLY, var_snmpStats, 1,
   {2}},
};

struct variable2 usecagent_variables[] = {
  {1, STRING, RONLY, var_usecAgent, 1,
   {1}},
  {2, GAUGE, RONLY, var_usecAgent, 1,
   {2}},
  {3, GAUGE, RONLY, var_usecAgent, 1,
   {3}},
  {4, INTEGER, RONLY, var_usecAgent, 1,
   {4}},
};

struct variable2 usecstats_variables[] = {
  {1, COUNTER, RONLY, var_usecStats, 1,
   {1}},
  {2, COUNTER, RONLY, var_usecStats, 1,
   {2}},
  {3, COUNTER, RONLY, var_usecStats, 1,
   {3}},
  {4, COUNTER, RONLY, var_usecStats, 1,
   {4}},
  {5, COUNTER, RONLY, var_usecStats, 1,
   {5}},
  {6, COUNTER, RONLY, var_usecStats, 1,
   {6}},
  {7, COUNTER, RONLY, var_usecStats, 1,
   {7}},
};

struct variable2 setserno_variables[] = {
  {1, INTEGER, RWRITE, var_rwstats, 1,
   {1}},
};

struct variable2 v2authtraps_variables[] = {
  {2, INTEGER, RWRITE, var_rwstats, 1,
   {4}},
};

/*
 * Note that the name field must be larger than any name that might
 * match that object.  For these variable length (objid) indexes
 * this might seem to be hard, but placing MAXINT in the first
 * subid of the index denotes an obcenely long objid, thereby ensuring that
 * none slip through.
 */

/* No access for community SNMP, RW possible for Secure SNMP */
#define PRIVRW   0x0003

/* No access for community SNMP, RO possible for Secure SNMP */
#define PRIVRO   0x0002

u_char *var_hosttimetab();

struct subtree subtrees[] = {
  {
    {MIB, 1}, 7, (struct variable *) system_variables,
    sizeof (system_variables) / sizeof (*system_variables),
    sizeof (*system_variables)},
  {
    {MIB, 1, 9, 1}, 9, (struct variable *) or_variables,
    sizeof (or_variables) / sizeof (*or_variables),
    sizeof (*or_variables)},
  {
    {MIB, 2}, 7, (struct variable *) interface_variables,
    sizeof (interface_variables) / sizeof (*interface_variables),
    sizeof (*interface_variables)},
  {
    {MIB, 3, 1, 1}, 9, (struct variable *) at_variables,
    sizeof (at_variables) / sizeof (*at_variables),
    sizeof (*at_variables)},
  {
    {MIB, 4}, 7, (struct variable *) ip_variables,
    sizeof (ip_variables) / sizeof (*ip_variables),
    sizeof (*ip_variables)},
  {
    {MIB, 5}, 7, (struct variable *) icmp_variables,
    sizeof (icmp_variables) / sizeof (*icmp_variables),
    sizeof (*icmp_variables)},
  {
    {MIB, 6}, 7, (struct variable *) tcp_variables,
    sizeof (tcp_variables) / sizeof (*tcp_variables),
    sizeof (*tcp_variables)},
  {
    {MIB, 7}, 7, (struct variable *) udp_variables,
    sizeof (udp_variables) / sizeof (*udp_variables),
    sizeof (*udp_variables)},

  /* Identification MIB (rfc 1414): */
  {
    {MIB, 24}, 7, (struct variable *) id_variables,
    sizeof (id_variables) / sizeof (*id_variables),
    sizeof (*id_variables)},

    /* Host Resources MIB ala rfc1514: */
  {
    {MIB, 25}, 7, (struct variable *) hr_variables,
    sizeof (hr_variables) / sizeof (*hr_variables),
    sizeof (*hr_variables)},
  {
    {SNMPSTATS}, 9, (struct variable *) snmpstats_variables,
    sizeof (snmpstats_variables) / sizeof (*snmpstats_variables),
    sizeof (*snmpstats_variables)},
  {
    {SNMPV1STATS}, 9, (struct variable *) snmpv1stats_variables,
    sizeof (snmpv1stats_variables) / sizeof (*snmpv1stats_variables),
    sizeof (*snmpv1stats_variables)},
  {
    {SNMPTRAP}, 9, (struct variable *) v2authtraps_variables,
    sizeof (v2authtraps_variables) / sizeof (*v2authtraps_variables),
    sizeof (*v2authtraps_variables)},
  {
    {SNMPSET}, 9, (struct variable *) setserno_variables,
    sizeof (setserno_variables) / sizeof (*setserno_variables),
    sizeof (*setserno_variables)},
  {
    {USECAGENT}, 9, (struct variable *) usecagent_variables,
    sizeof (usecagent_variables) / sizeof (*usecagent_variables),
    sizeof (*usecagent_variables)},
  {
    {USECSTATS}, 9, (struct variable *) usecstats_variables,
    sizeof (usecstats_variables) / sizeof (*usecstats_variables),
    sizeof (*usecstats_variables)},
};

static long setSerialNo       = 0;
static int  v2EnableAuthTraps = 1;

int in_view (oid *name, int namelen, int viewIndex)
{
  viewEntry *vwp, *savedvwp = NULL;
  extern viewEntry *views;

  for (vwp = views; vwp; vwp = vwp->next)
  {
    if (vwp->viewIndex != viewIndex)
       continue;
    if (vwp->viewSubtreeLen > namelen ||
        memcmp(vwp->viewSubtree, name, vwp->viewSubtreeLen * sizeof(oid)))
       continue;
    /* no wildcards here yet */
    if (!savedvwp)
         savedvwp = vwp;
    else if (vwp->viewSubtreeLen > savedvwp->viewSubtreeLen)
         savedvwp = vwp;
  }
  if (!savedvwp)
     return (FALSE);
  if (savedvwp->viewType == VIEWINCLUDED)
     return (TRUE);
  return (FALSE);
}

/*
 * getStatPtr - return a pointer to the named variable, as well as it's
 * type, length, and access control list.
 *
 * If an exact match for the variable name exists, it is returned.  If not,
 * and exact is false, the next variable lexicographically after the
 * requested one is returned.
 *
 * If no appropriate variable can be found, NULL is returned.
 */
u_char * getStatPtr (name, namelen, type, len, acl, exact, write_method,
                     snmpversion, noSuchObject, view)
     oid     *name;             /* IN - name of var, OUT - name matched */
     int     *namelen;          /* IN -number of sub-ids in name, OUT - subid-is in matched name */
     u_char  *type;             /* OUT - type of matched variable */
     int     *len;              /* OUT - length of matched variable */
     u_short *acl;		/* OUT - access control list */
     int      exact;            /* IN - TRUE if exact match wanted */
     int    (**write_method)(); /* OUT - pointer to function called to set variable, otherwise 0 */
     int     snmpversion;
     int    *noSuchObject;
     int     view;
{
  struct subtree *tp;
  struct variable *vp = 0;
  struct variable compat_var, *cvp = &compat_var;
  register int x;
  int y;
  register u_char *access = NULL;
  int result, treeresult;
  oid *suffix;
  int suffixlen;
  int found = FALSE;
  oid save[MAX_NAME_LEN];
  int savelen = 0;

  if (view == 0)
    return (NULL);

  if (!exact)
  {
    memcpy (save,name, *namelen *sizeof(oid));
    savelen = *namelen;
  }
  *write_method = NULL;
  for (y = 0, tp = subtrees; y < sizeof (subtrees) / sizeof (struct subtree);
       tp++, y++)
  {
    treeresult = compare_tree (name, *namelen, tp->name, (int) tp->namelen);

    /* if exact and treerresult == 0
     * if next  and treeresult <= 0
     */
    if (treeresult == 0 || (!exact && treeresult < 0))
    {
      result    = treeresult;
      suffixlen = *namelen - tp->namelen;
      suffix    = name + tp->namelen;

      /* the following is part of the setup for the compatability
       * structure below that has been moved out of the main loop.
       */
      memcpy (cvp->name,tp->name, tp->namelen*sizeof(oid));

      for (x = 0, vp = tp->variables;
           x < tp->variables_len;
           vp = (struct variable*)((char*) vp + tp->variables_width), x++)
      {
        /* if exact and ALWAYS
         * if next  and result >= 0
         */
        if (exact || result >= 0)
           result = compare_tree (suffix, suffixlen, vp->name,
                                  (int) vp->namelen);

        /* if exact and result == 0
         * if next  and result <= 0
         */
        if ((!exact && result <= 0) || (exact && result == 0))
        {
          /* builds an old (long) style variable structure to retain
           * compatability with var_* functions written previously.
           */
          memcpy ((cvp->name+tp->namelen, vp->name, vp->namelen*sizeof(oid));
          cvp->namelen = tp->namelen + vp->namelen;
          cvp->type    = vp->type;
          cvp->magic   = vp->magic;
          cvp->acl     = vp->acl;
          cvp->findVar = vp->findVar;
          access = (*(vp->findVar)) (cvp, name, namelen, exact,
                                     len, write_method);
          if (write_method)
             *acl = vp->acl;
          if (access && !in_view(name, *namelen, view))
          {
            access = NULL;
            *write_method = NULL;
          }
          else if (exact)
          {
            found = TRUE;
          }
          /* this code is incorrect if there is
           * a view configuration that exludes a particular
           * instance of a variable.  It would return noSuchObject,
           * which would be an error
           */
          if (access)
             break;
        }
        /* if exact and result <= 0 */
        if (exact && (result <= 0))
        {
          *type = vp->type;
          *acl = vp->acl;
          if (found)
               *noSuchObject = FALSE;
          else *noSuchObject = TRUE;
          return (NULL);
        }
      }
      if (access)
         break;
    }
  }
  if (y == sizeof(subtrees) / sizeof(struct subtree))
  {
    if (!access && !exact)
    {
      memcpy (name,save,savelen*sizeof(oid));
      *namelen = savelen;
    }
    if (found)
         *noSuchObject = FALSE;
    else *noSuchObject = TRUE;
    return (NULL);
  }
  /* vp now points to the approprate struct */
  *type = vp->type;
  *acl  = vp->acl;
  return (access);
}

int compare (oid *name1, int len1, oid *name2, int len2)
{
  int i, len;
#define cmpprintf  if(0) printf

  cmpprintf ("comparing ");
  for (i = 0; i < len1; i++)
      cmpprintf ("%ld%s", name1[i], i < len1 - 1 ? "." : "");
  cmpprintf (" with ");
  for (i = 0; i < len2; i++)
      cmpprintf ("%ld%s", name2[i], i < len2 - 1 ? "." : "");

  /* len = minimum of len1 and len2 */
  if (len1 < len2)
       len = len1;
  else len = len2;

  /* find first non-matching byte */
  while (len-- > 0)
  {
    if (*name1 < *name2)
    {
      cmpprintf (" giving -1\n");
      return (-1);
    }
    if (*name2++ < *name1++)
    {
      cmpprintf (" giving 1\n");
      return (1);
    }
  }
  /* bytes match up to length of shorter string */
  if (len1 < len2)
  {
    cmpprintf (" giving -1\n");
    return (-1);                /* name1 shorter, so it is "less" */
  }
  if (len2 < len1)
  {
    cmpprintf (" giving 1\n");
    return (1);
  }
  cmpprintf (" giving 0\n");
  return (0);                   /* both strings are equal */
}

static int compare_tree (oid *name1, int len1, oid *name2, int len2)
{
  int len;

  /* len = minimum of len1 and len2 */
  if (len1 < len2)
       len = len1;
  else len = len2;

  /* find first non-matching byte */
  while (len-- > 0)
  {
    if (*name1 < *name2)
       return (-1);
    if (*name2++ < *name1++)
       return (1);
  }
  /* bytes match up to length of shorter string */
  if (len1 < len2)
     return (-1);                  /* name1 shorter, so it is "less" */

  /* name1 matches name2 for length of name2, or they are equal */
  return (0);
}


char sysContact[256]  = "Unknown";
char sysLocation[256] = "Unknown";
char sysName[256]     = "Unknown";

oid version_id[] = { 1, 3, 6, 1, 4, 1, 3, 1, 1 };


u_long sysUpTime (void)
{
  struct timeval now, boottime;

  if (KNLookup (N_BOOTTIME, (char *) &boottime, sizeof (boottime)) == NULL)
     return (0);

  gettimeofday (&now,NULL);
  return (u_long) ((now.tv_sec - boottime.tv_sec) * 100 +
                   (now.tv_usec - boottime.tv_usec) / 10000);
}

u_char *var_hosttimetab (vp, name, length, exact, var_len, write_method)
  struct variable *vp;     /* IN - pointer to variable entry that points here */
  oid *name;               /* IN/OUT - input name requested,
                            *        - output name found */
  int *length;             /* IN/OUT - length of input and output oid's */
  int  exact;              /* IN - TRUE if an exact match was requested. */
  int *var_len;            /* OUT - length of variable or 0 if function returned. */
  int (**write_method)();  /* OUT - pointer to function to set variable,
                            * otherwise 0 */
{
  oid    newname[MAX_NAME_LEN];
  int    result;
  int    creationOrder;
  static int zero = 0;

  memcpy (newname,vp->name,vp->namelen*sizeof(oid));
  *write_method = 0;

  newname[vp->namelen] = (oid) 1;

  if (exact)
  {
    creationOrder = name[vp->namelen + 1];
    if (creationOrder > 2000)
       return (NULL);
    newname[vp->namelen + 1] = creationOrder;
  }
  else if (*length == vp->namelen + 2)
  {
    creationOrder = name[vp->namelen + 1] + 1;
    if (creationOrder > 2000)
      {
        if ((vp->name[vp->namelen - 1] != name[vp->namelen - 1]))
             creationOrder = 1;
        else return (NULL);
      }
    newname[vp->namelen + 1] = creationOrder;
  }
  else
  {
    printf ("Slow code\n");
    creationOrder = 1;
    while (creationOrder < 2000)
      {
        newname[vp->namelen + 1] = (oid) creationOrder++;
        result = compare (name, *length, newname, (int) vp->namelen + 2);
        if ((exact && (result == 0)) || (!exact && (result < 0)))
           break;
      }
    if (creationOrder == 2002)
       return (NULL);
  }
  bcopy ((char *) newname, (char *) name,
	 ((int) vp->namelen + 2) * sizeof (oid));
  *length = vp->namelen + 2;
  *var_len = sizeof (u_long);

  switch (vp->magic)
  {
    case HOSTTIMETABADDRESS:
         *var_len = sizeof (struct ether_addr);
         return (u_char *) "RMONRULES";
    case HOSTTIMETABCREATIONORDER:
         long_return = creationOrder;
         return (u_char*) &long_return;
    case HOSTTIMETABINDEX:
    case HOSTTIMETABINPKTS:
    case HOSTTIMETABOUTPKTS:
    case HOSTTIMETABINOCTETS:
    case HOSTTIMETABOUTOCTETS:
    case HOSTTIMETABOUTERRORS:
    case HOSTTIMETABOUTBCASTPKTS:
    case HOSTTIMETABOUTMCASTPKTS:
         return (u_char*) &zero;
    default:
         ERROR ("");
  }
}

u_char *var_system (vp, name, length, exact, var_len, write_method)
  struct variable *vp;     /* IN - pointer to variable entry that points here */
  oid   *name;             /* IN/OUT - input name requested, output name found */
  int   *length;           /* IN/OUT - length of input and output oid's */
  int    exact;            /* IN - TRUE if an exact match was requested. */
  int   *var_len;          /* OUT - length of variable or 0 if function returned. */
  int (**write_method)();  /* OUT - pointer to function to set variable, otherwise 0 */
{

  oid    newname[MAX_NAME_LEN];
  int    result;

  bcopy (vp->name,newname, vp->namelen*sizeof(oid));
  newname[8] = 0;
  result = compare (name, *length, newname, (int) vp->namelen + 1);
  if ((exact && result) || (!exact && result >= 0))
     return (NULL);

  bcopy (newname, name, (vp->namelen+1)*sizeof(oid));
  *length = vp->namelen + 1;
  *write_method = 0;
  *var_len = sizeof (long);	/* default length */

  switch (vp->magic)
  {
    case VERSION_DESCR:
         *var_len = strlen (version_descr);
         /* not writable:
          * *write_method = writeVersion;
          */
         return (u_char *) version_descr;
    case VERSION_ID:
         *var_len = sizeof (version_id);
         return (u_char *) version_id;
    case UPTIME:
    case SYSORLASTCHANGE:
         (u_long)long_return = sysUpTime();
         return (u_char*) &long_return;
    case IFNUMBER:
         long_return = Interface_Scan_Get_Count();
         return (u_char *) & long_return;
    case SYSCONTACT:
         *var_len = strlen (sysContact);
         *write_method = writeSystem;
         return (u_char *) sysContact;
    case SYSNAME:
         *var_len = strlen (sysName);
         *write_method = writeSystem;
         return (u_char *) sysName;
    case SYSLOCATION:
         *var_len = strlen (sysLocation);
         *write_method = writeSystem;
         return (u_char *) sysLocation;
    case SYSSERVICES:
         long_return = 72;
         return (u_char *) & long_return;
    default:
         ERROR ("");
  }
  return (NULL);
}

u_char *var_demo (vp, name, length, exact, var_len, write_method)
  struct variable *vp;    /* IN - pointer to variable entry that points here */
  oid   *name;            /* IN/OUT - input name requested, output name found */
  int   *length;          /* IN/OUT - length of input and output oid's */
  int    exact;           /* IN - TRUE if an exact match was requested. */
  int   *var_len;         /* OUT - length of variable or 0 if function returned. */
  int (**write_method)(); /* OUT - pointer to function to set variable, otherwise 0 */
{
  oid    newname[MAX_NAME_LEN];
  int    result;
  static u_char bitstring[64] = { 0, 0x83, 0 };
  int    bitstringlength  = 3;
  static u_char nsap[128] = { 0x14, 0x47, 0x00, 0x05, 0x80, 0xff, 0xff, 0x00,
                              0x00, 0x00, 0x01, 0x23, 0x01, 0x23, 0x01, 0x23,
                              0x45, 0x67, 0x89, 0xab, 0x01
                            };
  int    nsaplength = 21;
  static struct counter64 counter;

  counter.high = 0xFA202E75;
  counter.low  = 0x4FE92915;

  bcopy (vp->name,newname,vp->namelen*sizeof(oid));
  newname[7] = 0;
  result = compare (name, *length, newname, (int) vp->namelen + 1);
  if ((exact && result) || (!exact && result >= 0))
     return (NULL);

  bcopy (newname,name,(vp->namelen+1)*sizeof(oid));
  *length       = vp->namelen + 1;
  *write_method = 0;
  *var_len      = sizeof (long);     /* default length */

  switch (vp->magic)
  {
    case MTRBITSTRING:
         *var_len = bitstringlength;
         return (u_char *) bitstring;
    case MTRNSAPADDRESS:
         *var_len = nsaplength;
         return (u_char *) nsap;
    case MTRBIGNUMBER:
         *var_len = sizeof (counter);
         return (u_char *) & counter;
    default:
         ERROR ("");
  }
  return (NULL);
}


/*
 * XXX: version (sysDescr) is not writable (rfc 1213)
 */
int writeVersion (int action, u_char *var_val, u_char var_val_type,
                  int var_val_len, u_char *statP, oid *name, int name_len)
{
  int    bigsize = 1000;
  u_char buf[sizeof (version_descr)], *cp;
  int    count, size;

  if (var_val_type != STRING)
  {
    ERROR ("not string");
    return SNMP_ERR_WRONGTYPE;
  }
  if (var_val_len > sizeof (version_descr) - 1)
  {
    ERROR ("bad length");
    return SNMP_ERR_WRONGLENGTH;
  }
  size = sizeof (buf);
  asn_parse_string (var_val, &bigsize, &var_val_type, buf, &size);
  for (cp = buf, count = 0; count < size; count++, cp++)
  {
    if (!isprint (*cp))
    {
      printf ("not print %x\n", *cp);
      return SNMP_ERR_WRONGVALUE;
    }
  }
  buf[size] = 0;
  if (action == COMMIT)
     strcpy (version_descr, buf);

  return (SNMP_ERR_NOERROR);
}


static void save_into_conffile (char *key, char *val)
{
  extern char *snmp_configfile;
  char *p, *q, bak[1024], line[1024], newval[1000];
  int   saved = 0;
  FILE *in, *out;

  /* change newlines to spaces, cuz we cannot handle newlines in 
   * the configfile.
   */
  for (p = val, q = newval; *p && q < newval + 999; p++, q++)
      *q = *p == '\n' ? ' ' : *p;
  *q = 0;

  sprintf (bak, "%s.%d", snmp_configfile, getpid());

  if (!(in = fopen(snmp_configfile, "r")) ||
      !(out = fopen(bak, "w")))
  {
    fprintf (stderr, "snmpd: cannot change config file %s\n",
             snmp_configfile);
    return;
  }

  while (fgets (line, sizeof (line), in))
  {
    if (!strncmp (key, line, strlen (key)))
    {
      fprintf (out, "%s\t\t%s\n", key, newval);
      saved = 1;
    }
    else
      fputs (line, out);
  }
  fclose (in);
  if (!saved)
     fprintf (out, "\n## appended by set operation:\n%s\t\t%s\n",
              key, newval);

  fclose (out);
  if (rename(bak, snmp_configfile) < 0)
     perror ("snmpd: save: cannot rename config file");
}


u_char *var_rwstats (vp, name, length, exact, var_len, write_method)
  struct variable *vp;    /* IN - pointer to variable entry that points here */
  oid   *name;            /* IN/OUT - input name requested, output name found */
  int   *length;          /* IN/OUT - length of input and output oid's */
  int    exact;           /* IN - TRUE if an exact match was requested. */
  int   *var_len;         /* OUT - length of variable or 0 if function returned. */
  int (**write_method)(); /* OUT - pointer to function to set variable, otherwise 0 */
{
  oid newname[MAX_NAME_LEN];
  int result;

  bcopy (vp->name,newname,vp->namelen*sizeof(oid));
  newname[10] = 0;
  result = compare (name, *length, newname, (int) vp->namelen + 1);
  if ((exact && result) || (!exact && result >= 0))
     return (NULL);

  bcopy (newname,name,(vp->namelen+1)*sizeof(oid));
  *length       = vp->namelen + 1;
  *write_method = 0;
  *var_len      = sizeof(long);     /* default length */

  switch (vp->magic)
  {
    case 1:
         *write_method = writeSetSerialNumber;
         return (u_char *) & setSerialNo;
    case 2:
         return (u_char *) & v2EnableAuthTraps;
    default:
         ERROR ("");
  }
  return (NULL);
}

int writeSetSerialNumber (int action, u_char *var_val, u_char var_val_type,
                          int var_val_len, u_char *statP, oid *name, int name_len)
{
  int  size;
  long val;
  int  bigsize = 1000;

  if (var_val_type != INTEGER)
  {
    printf ("not string\n");
    return (SNMP_ERR_WRONGTYPE);
  }
  size = sizeof (long);
  asn_parse_int (var_val, &bigsize, &var_val_type, &val, size);
  if (val != setSerialNo)
     return (SNMP_ERR_INCONSISTENTVALUE);

  if (action == COMMIT)
     setSerialNo++;

  return (SNMP_ERR_NOERROR);
}


int writeSystem (int action, u_char *var_val, u_char var_val_type,
                 int var_val_len, u_char *statP, oid *name, int name_len)
{
  int    bigsize = 1000;
  u_char buf[sizeof(version_descr)], *cp;
  int    count, size;

  if (var_val_type != STRING)
  {
    printf ("not string\n");
    return (SNMP_ERR_WRONGTYPE);
  }
  if (var_val_len > sizeof(version_descr) - 1)
  {
    printf ("bad length\n");
    return (SNMP_ERR_WRONGLENGTH);
  }
  size = sizeof (buf);
  asn_parse_string (var_val, &bigsize, &var_val_type, buf, &size);
  for (cp = buf, count = 0; count < size; count++, cp++)
  {
    if (!isprint (*cp))
    {
      printf ("not print %x\n", *cp);
      return (SNMP_ERR_WRONGVALUE);
    }
  }
  buf[size] = 0;
  if (action == COMMIT)
  {
    switch ((char) name[7])
    {
      case 4:
           strcpy (sysContact, buf);
           save_into_conffile ("system contact:", sysContact);
           break;
      case 5:
           strcpy (sysName, buf);
           save_into_conffile ("system name:", sysName);
           break;
      case 6:
           strcpy (sysLocation, buf);
           save_into_conffile ("system location:", sysLocation);
           break;
    }
  }
  return (SNMP_ERR_NOERROR);
}

u_char *var_orEntry (vp, name, length, exact, var_len, write_method)
  struct variable *vp;    /* IN - pointer to variable entry that points here */
  oid   *name;            /* IN/OUT - input name requested, output name found */
  int   *length;          /* IN/OUT - length of input and output oid's */
  int    exact;           /* IN - TRUE if an exact match was requested. */
  int   *var_len;         /* OUT - length of variable or 0 if function returned. */
  int (**write_method)(); /* OUT - pointer to function to set variable, otherwise 0 */
{
  oid    newname[MAX_NAME_LEN];
  int    orindex;
  int    result, count;
  static struct orEntry {
         char *descr;
         int namelen;
         oid name[32];
       } orEntries[] = {
         { "CMU agent", 9, { 1, 3, 6, 1, 4, 1, 3, 1, 1 } }
       };

  bcopy (vp->name,newname, vp->namelen*sizeof(oid));

  count = sizeof (orEntries) / sizeof (struct orEntry);
  for (orindex = 1; orindex <= count; orindex++)
  {
    newname[10] = (oid) orindex;
    result = compare (name, *length, newname, (int) vp->namelen + 1);
    if ((exact && (result == 0)) || (!exact && (result < 0)))
      break;
  }
  if (orindex > count)
     return (NULL);

  bcopy (newname,name,(vp->namelen+1)*sizeof(oid));
  *length = vp->namelen + 1;
  *write_method = 0;

  orindex--;
  switch (vp->magic)
  {
    case 2:
         *var_len = orEntries[orindex].namelen * sizeof (oid);
         return (u_char *) orEntries[orindex].name;
    case 3:
         *var_len = strlen (orEntries[orindex].descr);
         return (u_char *) orEntries[orindex].descr;
    case 4:
         *var_len = sizeof (long);
         (u_long) long_return = sysUpTime();
         return (u_char *) & long_return;
    default:
         ERROR ("");
  }
  return (NULL);
}

u_char *var_ifEntry (vp, name, length, exact, var_len, write_method)
  struct variable *vp;    /* IN - pointer to variable entry that points here */
  oid   *name;            /* IN/OUT - input name requested, output name found */
  int   *length;          /* IN/OUT - length of input and output oid's */
  int    exact;           /* IN - TRUE if an exact match was requested. */
  int   *var_len;         /* OUT - length of variable or 0 if function returned. */
  int (**write_method)(); /* OUT - pointer to function to set variable, otherwise 0 */
{
  oid    newname[MAX_NAME_LEN];
  int    interface;
  int    result, count;
  static struct ifnet ifnet;
  static struct in_ifaddr in_ifaddr;
  static char Name[16];
  char *cp;

  bcopy (vp->name,newname,vp->namelen*sizeof(oid));
  /* find "next" interface */
  count = Interface_Scan_Get_Count();
  for (interface = 1; interface <= count; interface++)
  {
    newname[10] = (oid) interface;
    result = compare (name, *length, newname, (int) vp->namelen + 1);
    if ((exact && (result == 0)) || (!exact && (result < 0)))
      break;
  }
  if (interface > count)
    return (NULL);

  bcopy (newname,name,(vp->namelen+1)*sizeof(oid));
  *length = vp->namelen + 1;
  *write_method = 0;
  *var_len = sizeof (long);

  Interface_Scan_By_Index (interface, Name, &ifnet, &in_ifaddr);
  switch (vp->magic)
  {
    case IFINDEX:
      long_return = interface;
      return (u_char *) & long_return;
    case IFDESCR:
#define USE_NAME_AS_DESCRIPTION
#ifdef USE_NAME_AS_DESCRIPTION
      cp = Name;
#else	/* USE_NAME_AS_DESCRIPTION */
      cp = Lookup_Device_Annotation (Name, "snmp-descr");
      if (!cp)
	cp = Lookup_Device_Annotation (Name, 0);
      if (!cp)
	cp = Name;
#endif	/* USE_NAME_AS_DESCRIPTION */
      *var_len = strlen (cp);
      return (u_char *) cp;
    case IFTYPE:
        long_return = 1;        /* OTHER */
      return (u_char *) & long_return;
    case IFMTU:
      {
	long_return = (long) ifnet.if_mtu;
	return (u_char *) & long_return;
      }
    case IFSPEED:
	(u_long) long_return = 1;	/* OTHER */
      return (u_char *) & long_return;
    case IFPHYSADDRESS:
      Interface_Get_Ether_By_Index (interface, return_buf);
      *var_len = 6;
      return (u_char *) return_buf;
    case IFADMINSTATUS:
      long_return = ifnet.if_flags & IFF_RUNNING ? 1 : 2;
      return (u_char *) & long_return;
    case IFOPERSTATUS:
      long_return = ifnet.if_flags & IFF_UP ? 1 : 2;
      return (u_char *) & long_return;
    case IFLASTCHANGE:
      long_return = 0;		/* XXX */
      return (u_char *) & long_return;
    case IFINOCTETS:
      (u_long) long_return = ifnet.if_ipackets * 308;	/* XXX */
      return (u_char *) & long_return;
    case IFINUCASTPKTS:
      (u_long) long_return = ifnet.if_ipackets;
      return (u_char *) & long_return;
    case IFINNUCASTPKTS:
      (u_long) long_return = 0;	/* XXX */
      return (u_char *) & long_return;
    case IFINDISCARDS:
      (u_long) long_return = 0;	/* XXX */
      return (u_char *) & long_return;
    case IFINERRORS:
      return (u_char *) & ifnet.if_ierrors;
    case IFINUNKNOWNPROTOS:
      (u_long) long_return = 0;	/* XXX */
      return (u_char *) & long_return;
    case IFOUTOCTETS:
      (u_long) long_return = ifnet.if_opackets * 308;	/* XXX */
      return (u_char *) & long_return;
    case IFOUTUCASTPKTS:
      (u_long) long_return = ifnet.if_opackets;
      return (u_char *) & long_return;
    case IFOUTNUCASTPKTS:
      (u_long) long_return = 0;	/* XXX */
      return (u_char *) & long_return;
    case IFOUTDISCARDS:
      return (u_char *) & ifnet.if_snd.ifq_drops;
    case IFOUTERRORS:
      return (u_char *) & ifnet.if_oerrors;
    case IFOUTQLEN:
      return (u_char *) & ifnet.if_snd.ifq_len;
    default:
      ERROR ("");
    }
  return (NULL);
}

/*
 * Read the ARP table
 */

u_char *
var_atEntry (vp, name, length, exact, var_len, write_method)
     register struct variable *vp;	/* IN - pointer to variable entry that points here */
     register oid *name;	/* IN/OUT - input name requested, output name found */
     register int *length;	/* IN/OUT - length of input and output oid's */
     int exact;			/* IN - TRUE if an exact match was requested. */
     int *var_len;		/* OUT - length of variable or 0 if function returned. */
     int (**write_method)();   /* OUT - pointer to function to set variable, otherwise 0 */
{
  /*
   * object identifier is of form:
   * 1.3.6.1.2.1.3.1.1.1.interface.1.A.B.C.D,  where A.B.C.D is IP address.
   * Interface is at offset 10,
   * IPADDR starts at offset 12.
   */
  u_char *cp;
  oid *op;
  oid lowest[16];
  oid current[16];
  static char PhysAddr[6], LowPhysAddr[6];
  u_long Addr, LowAddr;

  /* fill in object part of name for current (less sizeof instance part) */

  bcopy ((char *) vp->name, (char *) current, (int) vp->namelen * sizeof (oid));

  LowAddr = -1;			/* Don't have one yet */
  ARP_Scan_Init();
  for (;;)
    {
      if (ARP_Scan_Next (&Addr, PhysAddr) == 0)
	break;
      current[10] = 1;		/* IfIndex == 1 (ethernet???) XXX */
      current[11] = 1;
      cp = (u_char *) & Addr;
      op = current + 12;
      *op++ = *cp++;
      *op++ = *cp++;
      *op++ = *cp++;
      *op++ = *cp++;

      if (exact)
	{
	  if (compare (current, 16, name, *length) == 0)
	    {
	      bcopy ((char *) current, (char *) lowest, 16 * sizeof (oid));
	      LowAddr = Addr;
	      bcopy (PhysAddr, LowPhysAddr, sizeof (PhysAddr));
	      break;		/* no need to search further */
	    }
	}
      else
	{
	  if ((compare (current, 16, name, *length) > 0) &&
	      ((LowAddr == -1) || (compare (current, 16, lowest, 16) < 0)))
	    {
	      /*
	       * if new one is greater than input and closer to input than
	       * previous lowest, save this one as the "next" one.
	       */
	      bcopy ((char *) current, (char *) lowest, 16 * sizeof (oid));
	      LowAddr = Addr;
	      bcopy (PhysAddr, LowPhysAddr, sizeof (PhysAddr));
	    }
	}
    }
  if (LowAddr == -1)
    return (NULL);

  bcopy ((char *) lowest, (char *) name, 16 * sizeof (oid));
  *length = 16;
  *write_method = 0;
  switch (vp->magic)
    {
    case ATIFINDEX:
      *var_len = sizeof long_return;
      long_return = 1;          /* XXX */
      return (u_char *) & long_return;
    case ATPHYSADDRESS:
      *var_len = sizeof (LowPhysAddr);
      return (u_char *) LowPhysAddr;
    case ATNETADDRESS:
      *var_len = sizeof long_return;
      long_return = LowAddr;
      return (u_char *) & long_return;
    default:
      ERROR ("");
    }
  return (NULL);
}


u_char *
var_ip (vp, name, length, exact, var_len, write_method)
     register struct variable *vp;	/* IN - pointer to variable entry that points here */
     oid *name;			/* IN/OUT - input name requested, output name found */
     int *length;		/* IN/OUT - length of input and output oid's */
     int exact;			/* IN - TRUE if an exact match was requested. */
     int *var_len;		/* OUT - length of variable or 0 if function returned. */
     int (**write_method)();   /* OUT - pointer to function to set variable, otherwise 0 */
{
  static struct ipstat ipstat;
  oid newname[MAX_NAME_LEN];
  int result;

  bcopy ((char *) vp->name, (char *) newname, (int) vp->namelen * sizeof (oid));
  newname[8] = 0;
  result = compare (name, *length, newname, (int) vp->namelen + 1);
  if ((exact && (result != 0)) || (!exact && (result >= 0)))
    return (NULL);
  bcopy ((char *) newname, (char *) name, ((int) vp->namelen + 1) * sizeof (oid));
  *length = vp->namelen + 1;

  *write_method = 0;
  *var_len = sizeof (long);	/* default length */
  /*
   *        Get the IP statistics from the kernel...
   */

  KNLookup (N_IPSTAT, (char *) &ipstat, sizeof (ipstat));

  switch (vp->magic)
    {
    case IPFORWARDING:
      KNLookup (N_IPFORWARDING, (char *) &i, sizeof (i));
      fflush (stderr);
      if (i == 1)
	{
	  long_return = 1;	/* GATEWAY */
	}
      else
	{
	  long_return = 2;	/* HOST    */
	}
      return (u_char *) & long_return;
    case IPDEFAULTTTL:
      /*
       *        Allow for a kernel w/o TCP.
       */
      if (nl[N_TCP_TTL].n_value)
	{
	  KNLookup (N_TCP_TTL, (char *) &long_return, sizeof (long_return));
	}
      else
	long_return = 60;	/* XXX */
      return (u_char *) & long_return;
    case IPINRECEIVES:
      return (u_char *) & ipstat.ips_total;
    case IPINHDRERRORS:
      long_return = ipstat.ips_badsum + ipstat.ips_tooshort +
	ipstat.ips_toosmall + ipstat.ips_badhlen +
	ipstat.ips_badlen;
      return (u_char *) & long_return;
    case IPINADDRERRORS:
      return (u_char *) & ipstat.ips_cantforward;

    case IPFORWDATAGRAMS:
      return (u_char *) & ipstat.ips_forward;

    case IPINUNKNOWNPROTOS:
      long_return = 0;
      return (u_char *) & long_return;
    case IPINDISCARDS:
      long_return = 0;
      return (u_char *) & long_return;
    case IPINDELIVERS:

      long_return = ipstat.ips_total -
	(ipstat.ips_badsum + ipstat.ips_tooshort +
	 ipstat.ips_toosmall + ipstat.ips_badhlen +
	 ipstat.ips_badlen);
      return (u_char *) & long_return;

    case IPOUTREQUESTS:
      long_return = 0;
      return (u_char *) & long_return;
    case IPOUTDISCARDS:
      long_return = 0;
      return (u_char *) & long_return;
    case IPOUTNOROUTES:
      return (u_char *) & ipstat.ips_cantforward;

    case IPREASMTIMEOUT:
      long_return = IPFRAGTTL;
      return (u_char *) & long_return;
    case IPREASMREQDS:
      return (u_char *) & ipstat.ips_fragments;

    case IPREASMOKS:
      return (u_char *) & ipstat.ips_fragments;

    case IPREASMFAILS:
      long_return = ipstat.ips_fragdropped + ipstat.ips_fragtimeout;
      return (u_char *) & long_return;

    case IPFRAGOKS:
      long_return = 0;
      return (u_char *) & long_return;
    case IPFRAGFAILS:
      long_return = 0;
      return (u_char *) & long_return;
    case IPFRAGCREATES:
      long_return = 0;
      return (u_char *) & long_return;
    default:
      ERROR ("");
    }
  return (NULL);
}



u_char *
var_ipAddrEntry (vp, name, length, exact, var_len, write_method)
     register struct variable *vp;	/* IN - pointer to variable entry that points here */
     register oid *name;	/* IN/OUT - input name requested, output name found */
     register int *length;	/* IN/OUT - length of input and output oid's */
     int exact;			/* IN - TRUE if an exact match was requested. */
     int *var_len;		/* OUT - length of variable or 0 if function returned. */
     int (**write_method)();   /* OUT - pointer to function to set variable, otherwise 0 */
{
  /*
   * object identifier is of form:
   * 1.3.6.1.2.1.4.20.1.?.A.B.C.D,  where A.B.C.D is IP address.
   * IPADDR starts at offset 10.
   */
  oid lowest[14];
  oid current[14], *op;
  u_char *cp;
  int interface, lowinterface = 0;
  static struct ifnet ifnet, lowin_ifnet;
  static struct in_ifaddr in_ifaddr, lowin_ifaddr;

  /* fill in object part of name for current (less sizeof instance part) */

  bcopy ((char *) vp->name, (char *) current, (int) vp->namelen * sizeof (oid));

  Interface_Scan_Init();
  for (;;)
    {
      if (Interface_Scan_Next (&interface, (char *) 0, &ifnet, &in_ifaddr) == 0)
	break;
      cp = (u_char *) & (((struct sockaddr_in *) &(in_ifaddr.ia_addr))->sin_addr.s_addr);

      op = current + 10;
      *op++ = *cp++;
      *op++ = *cp++;
      *op++ = *cp++;
      *op++ = *cp++;
      if (exact)
	{
	  if (compare (current, 14, name, *length) == 0)
	    {
	      bcopy ((char *) current, (char *) lowest, 14 * sizeof (oid));
	      lowinterface = interface;
	      lowin_ifaddr = in_ifaddr;
	      break;		/* no need to search further */
	    }
	}
      else
	{
	  if ((compare (current, 14, name, *length) > 0) &&
	      (!lowinterface || (compare (current, 14, lowest, 14) < 0)))
	    {
	      /*
	       * if new one is greater than input and closer to input than
	       * previous lowest, save this one as the "next" one.
	       */
	      lowinterface = interface;
	      lowin_ifaddr = in_ifaddr;
	      bcopy ((char *) current, (char *) lowest, 14 * sizeof (oid));
	    }
	}
    }

  if (!lowinterface)
    return (NULL);
  bcopy ((char *) lowest, (char *) name, 14 * sizeof (oid));
  *length = 14;
  *write_method = 0;
  *var_len = sizeof (long_return);
  switch (vp->magic)
    {
    case IPADADDR:
      return (u_char *) & ((struct sockaddr_in *) &lowin_ifaddr.ia_addr)->sin_addr.s_addr;
    case IPADIFINDEX:
      long_return = lowinterface;
      return (u_char *) & long_return;
    case IPADNETMASK:
      long_return = ntohl (lowin_ifaddr.ia_subnetmask);
      return (u_char *) & long_return;
    case IPADBCASTADDR:
      long_return = ntohl (((struct sockaddr_in *) &lowin_ifaddr.ia_addr)->sin_addr.s_addr) & 1;
      return (u_char *) & long_return;
    default:
      ERROR ("");
    }
  return (NULL);
}


u_char *
var_icmp (vp, name, length, exact, var_len, write_method)
     register struct variable *vp;	/* IN - pointer to variable entry that points here */
     oid *name;			/* IN/OUT - input name requested, output name found */
     int *length;		/* IN/OUT - length of input and output oid's */
     int exact;			/* IN - TRUE if an exact match was requested. */
     int *var_len;		/* OUT - length of variable or 0 if function returned. */
     int (**write_method)();   /* OUT - pointer to function to set variable, otherwise 0 */
{
  static struct icmpstat icmpstat;
  oid newname[MAX_NAME_LEN];
  int result;
  int i;

  bcopy ((char *) vp->name, (char *) newname, (int) vp->namelen * sizeof (oid));
  newname[8] = 0;
  result = compare (name, *length, newname, (int) vp->namelen + 1);
  if ((exact && (result != 0)) || (!exact && (result >= 0)))
    return (NULL);
  bcopy ((char *) newname, (char *) name, ((int) vp->namelen + 1) * sizeof (oid));
  *length = vp->namelen + 1;

  *write_method = 0;
  *var_len = sizeof (long);	/* all following variables are sizeof long */

  /*
   *        Get the ICMP statistics from the kernel...
   */

  KNLookup (N_ICMPSTAT, (char *) &icmpstat, sizeof (icmpstat));

  switch (vp->magic)
    {
    case ICMPINMSGS:
      long_return = icmpstat.icps_badcode + icmpstat.icps_tooshort +
	icmpstat.icps_checksum + icmpstat.icps_badlen;
      for (i = 0; i <= ICMP_MAXTYPE; i++)
	long_return += icmpstat.icps_inhist[i];
      return (u_char *) & long_return;
    case ICMPINERRORS:
      long_return = icmpstat.icps_badcode + icmpstat.icps_tooshort +
	icmpstat.icps_checksum + icmpstat.icps_badlen;
      return (u_char *) & long_return;
    case ICMPINDESTUNREACHS:
      return (u_char *) & icmpstat.icps_inhist[ICMP_UNREACH];
    case ICMPINTIMEEXCDS:
      return (u_char *) & icmpstat.icps_inhist[ICMP_TIMXCEED];
    case ICMPINPARMPROBS:
      return (u_char *) & icmpstat.icps_inhist[ICMP_PARAMPROB];
    case ICMPINSRCQUENCHS:
      return (u_char *) & icmpstat.icps_inhist[ICMP_SOURCEQUENCH];
    case ICMPINREDIRECTS:
      return (u_char *) & icmpstat.icps_inhist[ICMP_REDIRECT];
    case ICMPINECHOS:
      return (u_char *) & icmpstat.icps_inhist[ICMP_ECHO];
    case ICMPINECHOREPS:
      return (u_char *) & icmpstat.icps_inhist[ICMP_ECHOREPLY];
    case ICMPINTIMESTAMPS:
      return (u_char *) & icmpstat.icps_inhist[ICMP_TSTAMP];
    case ICMPINTIMESTAMPREPS:
      return (u_char *) & icmpstat.icps_inhist[ICMP_TSTAMPREPLY];
    case ICMPINADDRMASKS:
      return (u_char *) & icmpstat.icps_inhist[ICMP_MASKREQ];
    case ICMPINADDRMASKREPS:
      return (u_char *) & icmpstat.icps_inhist[ICMP_MASKREPLY];
    case ICMPOUTMSGS:
      long_return = icmpstat.icps_oldshort + icmpstat.icps_oldicmp;
      for (i = 0; i <= ICMP_MAXTYPE; i++)
	long_return += icmpstat.icps_outhist[i];
      return (u_char *) & long_return;
    case ICMPOUTERRORS:
      long_return = icmpstat.icps_oldshort + icmpstat.icps_oldicmp;
      return (u_char *) & long_return;
    case ICMPOUTDESTUNREACHS:
      return (u_char *) & icmpstat.icps_outhist[ICMP_UNREACH];
    case ICMPOUTTIMEEXCDS:
      return (u_char *) & icmpstat.icps_outhist[ICMP_TIMXCEED];
    case ICMPOUTPARMPROBS:
      return (u_char *) & icmpstat.icps_outhist[ICMP_PARAMPROB];
    case ICMPOUTSRCQUENCHS:
      return (u_char *) & icmpstat.icps_outhist[ICMP_SOURCEQUENCH];
    case ICMPOUTREDIRECTS:
      return (u_char *) & icmpstat.icps_outhist[ICMP_REDIRECT];
    case ICMPOUTECHOS:
      return (u_char *) & icmpstat.icps_outhist[ICMP_ECHO];
    case ICMPOUTECHOREPS:
      return (u_char *) & icmpstat.icps_outhist[ICMP_ECHOREPLY];
    case ICMPOUTTIMESTAMPS:
      return (u_char *) & icmpstat.icps_outhist[ICMP_TSTAMP];
    case ICMPOUTTIMESTAMPREPS:
      return (u_char *) & icmpstat.icps_outhist[ICMP_TSTAMPREPLY];
    case ICMPOUTADDRMASKS:
      return (u_char *) & icmpstat.icps_outhist[ICMP_MASKREQ];
    case ICMPOUTADDRMASKREPS:
      return (u_char *) & icmpstat.icps_outhist[ICMP_MASKREPLY];
    default:
      ERROR ("");
    }
  return (NULL);
}


u_char *
var_udp (vp, name, length, exact, var_len, write_method)
     register struct variable *vp;	/* IN - pointer to variable entry that points here */
     oid *name;			/* IN/OUT - input name requested, output name found */
     int *length;		/* IN/OUT - length of input and output oid's */
     int exact;			/* IN - TRUE if an exact match was requested. */
     int *var_len;		/* OUT - length of variable or 0 if function returned. */
     int (**write_method)();   /* OUT - pointer to function to set variable, otherwise 0 */
{
  static struct udpstat udpstat;
  oid newname[MAX_NAME_LEN];
  int result;

  bcopy ((char *) vp->name, (char *) newname, (int) vp->namelen * sizeof (oid));
  newname[8] = 0;
  result = compare (name, *length, newname, (int) vp->namelen + 1);
  if ((exact && (result != 0)) || (!exact && (result >= 0)))
    return (NULL);
  bcopy ((char *) newname, (char *) name, ((int) vp->namelen + 1) * sizeof (oid));
  *length = vp->namelen + 1;

  *write_method = 0;
  *var_len = sizeof (long); /* default length */
  /*
   *        Get the IP statistics from the kernel...
   */

  KNLookup (N_UDPSTAT, (char *) &udpstat, sizeof (udpstat));

  switch (vp->magic)
  {
    case UDPINDATAGRAMS:
    case UDPNOPORTS:
    case UDPOUTDATAGRAMS:
         long_return = 0;
         return (u_char *) & long_return;
    case UDPINERRORS:
         long_return = udpstat.udps_hdrops + udpstat.udps_badsum +
                       udpstat.udps_badlen;
         return (u_char *) & long_return;
    default:
         ERROR ("");
  }
  return (NULL);
}


u_char *
var_tcp (vp, name, length, exact, var_len, write_method)
     register struct variable *vp;	/* IN - pointer to variable entry that points here */
     oid *name;			/* IN/OUT - input name requested, output name found */
     int *length;		/* IN/OUT - length of input and output oid's */
     int exact;			/* IN - TRUE if an exact match was requested. */
     int *var_len;		/* OUT - length of variable or 0 if function returned. */
     int (**write_method)();   /* OUT - pointer to function to set variable, otherwise 0 */
{
  int i, result;
  static struct tcpstat tcpstat;
  oid newname[MAX_NAME_LEN], lowest[MAX_NAME_LEN], *op;
  u_char *cp;
  int State, LowState;
  static struct inpcb inpcb, Lowinpcb;

  /*
   *        Allow for a kernel w/o TCP
   */

  if (nl[N_TCPSTAT].n_value == 0)
    return (NULL);

  if (vp->magic < TCPCONNSTATE)
    {
      bcopy ((char *) vp->name, (char *) newname,
	     (int) vp->namelen * sizeof (oid));
      newname[8] = 0;
      result = compare (name, *length, newname, (int) vp->namelen + 1);
      if ((exact && (result != 0)) || (!exact && (result >= 0)))
        return (NULL);
      bcopy ((char *) newname, (char *) name,
	     ((int) vp->namelen + 1) * sizeof (oid));
      *length = vp->namelen + 1;

      *write_method = 0;
      *var_len = sizeof (long);	/* default length */
      /*
       *  Get the TCP statistics from the kernel...
       */

      KNLookup (N_TCPSTAT, (char *) &tcpstat, sizeof (tcpstat));

      switch (vp->magic)
	{
	case TCPRTOALGORITHM:
          long_return = 4;      /* Van Jacobsen's algorithm *//* XXX */
	  return (u_char *) & long_return;
          if (!tcpstat.TcpRtoAlgorithm)
	    {
	      /* 0 is illegal: assume `other' algorithm: */
	      long_return = 1;
	      return (u_char *) & long_return;
	    }
	  return (u_char *) & tcpstat.TcpRtoAlgorithm;
        case TCPRTOMIN:
          long_return = TCPTV_MIN / PR_SLOWHZ * 1000;
	  return (u_char *) & long_return;
	case TCPRTOMAX:
          long_return = TCPTV_REXMTMAX / PR_SLOWHZ * 1000;
	  return (u_char *) & long_return;
	case TCPMAXCONN:
          long_return = -1;
	  return (u_char *) & long_return;
	case TCPACTIVEOPENS:
          return (u_char *) & tcpstat.tcps_connattempt;

	case TCPPASSIVEOPENS:
          return (u_char *) & tcpstat.tcps_accepts;

	case TCPATTEMPTFAILS:
	  return (u_char *) & tcpstat.tcps_conndrops;

	case TCPESTABRESETS:
	  return (u_char *) & tcpstat.tcps_drops;

	case TCPCURRESTAB:
          long_return = TCP_Count_Connections();
	  return (u_char *) & long_return;

        case TCPINSEGS:
	  return (u_char *) & tcpstat.tcps_rcvtotal;

	case TCPOUTSEGS:
	  return (u_char *) & tcpstat.tcps_sndtotal;
	case TCPRETRANSSEGS:
	  return (u_char *) & tcpstat.tcps_sndrexmitpack;
	default:
	  ERROR ("");
	}
    }
  else
    {				/* Info about a particular connection */
      bcopy ((char *) vp->name, (char *) newname, (int) vp->namelen * sizeof (oid));
      /* find "next" connection */
    Again:
      LowState = -1;		/* Don't have one yet */
      TCP_Scan_Init();
      for (;;)
	{
	  if ((i = TCP_Scan_Next (&State, &inpcb)) < 0)
	    goto Again;
	  if (i == 0)
	    break;		/* Done */
	  cp = (u_char *) & inpcb.inp_laddr.s_addr;
	  op = newname + 10;
	  *op++ = *cp++;
	  *op++ = *cp++;
	  *op++ = *cp++;
	  *op++ = *cp++;

	  newname[14] = ntohs (inpcb.inp_lport);

	  cp = (u_char *) & inpcb.inp_faddr.s_addr;
	  op = newname + 15;
	  *op++ = *cp++;
	  *op++ = *cp++;
	  *op++ = *cp++;
	  *op++ = *cp++;

	  newname[19] = ntohs (inpcb.inp_fport);

	  if (exact)
	    {
	      if (compare (newname, 20, name, *length) == 0)
		{
		  bcopy ((char *) newname, (char *) lowest, 20 * sizeof (oid));
		  LowState = State;
		  Lowinpcb = inpcb;
		  break;	/* no need to search further */
		}
	    }
	  else
	    {
	      if ((compare (newname, 20, name, *length) > 0) &&
		((LowState < 0) || (compare (newname, 20, lowest, 20) < 0)))
		{
		  /*
		   * if new one is greater than input and closer to input than
		   * previous lowest, save this one as the "next" one.
		   */
		  bcopy ((char *) newname, (char *) lowest, 20 * sizeof (oid));
		  LowState = State;
		  Lowinpcb = inpcb;
		}
	    }
	}
      if (LowState < 0)
	return (NULL);
      bcopy ((char *) lowest, (char *) name, ((int) vp->namelen + 10) * sizeof (oid));
      *length = vp->namelen + 10;
      *write_method = 0;
      *var_len = sizeof (long);
      switch (vp->magic)
	{
	case TCPCONNSTATE:
	  {
	    static int StateMap[] =
	    {1, 2, 3, 4, 5, 8, 6, 10, 9, 7, 11};
	    return (u_char *) & StateMap[LowState];
	  }
	case TCPCONNLOCALADDRESS:
	  return (u_char *) & Lowinpcb.inp_laddr.s_addr;
	case TCPCONNLOCALPORT:
	  long_return = ntohs (Lowinpcb.inp_lport);
	  return (u_char *) & long_return;
	case TCPCONNREMADDRESS:
	  return (u_char *) & Lowinpcb.inp_faddr.s_addr;
	case TCPCONNREMPORT:
	  long_return = ntohs (Lowinpcb.inp_fport);
	  return (u_char *) & long_return;
	}
    }
  return (NULL);
}


/*
 *    Print INTERNET connections
 */

static int 
TCP_Count_Connections()
{
  int Established;
  struct inpcb cb;
  register struct inpcb *prev, *next;
  struct inpcb inpcb;
  struct tcpcb tcpcb;

Again:				/*
				 *    Prepare to scan the control blocks
				 */
  Established = 0;

  KNLookup (N_TCB, (char *) &cb, sizeof (struct inpcb));
  inpcb = cb;
  prev = (struct inpcb *) nl[N_TCB].n_value;
  /*
   *    Scan the control blocks
   */
  while (inpcb.inp_next != (struct inpcb *) nl[N_TCB].n_value)
    {
      next = inpcb.inp_next;

      klookup ((int) next, (char *) &inpcb, sizeof (inpcb));
      if (inpcb.inp_prev != prev)
	{			/* ??? */
          sleep (1);
          goto Again;
	}
      if (inet_lnaof (inpcb.inp_laddr) == INADDR_ANY)
	{
	  prev = next;
	  continue;
	}
      klookup ((int) inpcb.inp_ppcb, (char *) &tcpcb, sizeof (tcpcb));

      if ((tcpcb.t_state == TCPS_ESTABLISHED) ||
	  (tcpcb.t_state == TCPS_CLOSE_WAIT))
	Established++;
      prev = next;
    }
  return (Established);
}


static struct inpcb inpcb, *prev;

static void TCP_Scan_Init()
{
  KNLookup (N_TCB, (char *) &inpcb, sizeof (inpcb));
  prev = (struct inpcb *) nl[N_TCB].n_value;
}

static int
TCP_Scan_Next (State, RetInPcb)
     int *State;
     struct inpcb *RetInPcb;
{
  register struct inpcb *next;
  struct tcpcb tcpcb;

  if (inpcb.inp_next == (struct inpcb *) nl[N_TCB].n_value)
    {
      return (0);		/* "EOF" */
    }

  next = inpcb.inp_next;

  klookup ((int) next, (char *) &inpcb, sizeof (inpcb));
  if (inpcb.inp_prev != prev)	/* ??? */
    return (-1);		/* "FAILURE" */

  klookup ((int) inpcb.inp_ppcb, (char *) &tcpcb, sizeof (tcpcb));
  *State = tcpcb.t_state;
  *RetInPcb = inpcb;
  prev = next;
  return (1);			/* "OK" */
}

static int arptab_size, arptab_current;
static struct arptab *at = 0;

static void
ARP_Scan_Init()
{
  if (!at)
    {
      KNLookup (N_ARPTAB_SIZE, (char *) &arptab_size, sizeof arptab_size);
      at = (struct arptab *) malloc (arptab_size * sizeof (struct arptab));
    }

  KNLookup (N_ARPTAB, (char *) at, arptab_size * sizeof (struct arptab));
  arptab_current = 0;
}

static int ARP_Scan_Next (u_long *IPAddr, char *PhysAddr)
{
  register struct arptab *atab;

  while (arptab_current < arptab_size)
    {
      atab = &at[arptab_current++];
      if (!(atab->at_flags & ATF_COM))
	continue;
      *IPAddr = atab->at_iaddr.s_addr;
      bcopy ((char *) &atab->at_enaddr, PhysAddr, sizeof (atab->at_enaddr));
      return (1);
    }
  return (0);			/* "EOF" */
}


static struct in_ifaddr savein_ifaddr;
static struct ifnet *ifnetaddr, saveifnet, *saveifnetaddr;
static int saveIndex = 0;
static char saveName[16];


void Interface_Scan_Init()
{
  KNLookup (N_IFNET, (char *) &ifnetaddr, sizeof (ifnetaddr));
  saveIndex = 0;
}



int Interface_Scan_Next (Index, Name, Retifnet, Retin_ifaddr)
     int *Index;
     char *Name;
     struct ifnet *Retifnet;
     struct in_ifaddr *Retin_ifaddr;
{
  struct ifnet ifnet;
  struct in_ifaddr *ia, in_ifaddr;
  char *cp;

  while (ifnetaddr)
    {
      /*
       *            Get the "ifnet" structure and extract the device name
       */
      klookup ((int) ifnetaddr, (char *) &ifnet, sizeof ifnet);
      klookup ((int) ifnet.if_name, (char *) saveName, 16);

      saveName[15] = '\0';
      cp = index (saveName, '\0');
      *cp++ = ifnet.if_unit + '0';
      *cp = '\0';
      if (1 || strcmp (saveName, "lo0") != 0)
	{			/* XXX */
	  /*
	   *  Try to find an address for this interface
	   */

	  KNLookup (N_IN_IFADDR, (char *) &ia, sizeof (ia));
	  while (ia)
	    {
	      klookup ((int) ia, (char *) &in_ifaddr, sizeof (in_ifaddr));
	      if (in_ifaddr.ia_ifp == ifnetaddr)
		break;
	      ia = in_ifaddr.ia_next;
	    }

	  ifnet.if_addrlist = (struct ifaddr *) ia;	/* WRONG DATA TYPE; ONLY A FLAG */
	  /*              ifnet.if_addrlist = (struct ifaddr *)&ia->ia_ifa;   *//* WRONG DATA TYPE; ONLY A FLAG */

	  if (Index)
	    *Index = ++saveIndex;
	  if (Retifnet)
	    *Retifnet = ifnet;
	  if (Retin_ifaddr)
	    *Retin_ifaddr = in_ifaddr;
	  if (Name)
	    strcpy (Name, saveName);
	  saveifnet = ifnet;
	  saveifnetaddr = ifnetaddr;
	  savein_ifaddr = in_ifaddr;
	  ifnetaddr = ifnet.if_next;

	  return (1);		/* DONE */
	}
      ifnetaddr = ifnet.if_next;
    }
  return (0);			/* EOF */
}


static int 
Interface_Scan_By_Index (Index, Name, Retifnet, Retin_ifaddr)
     int Index;
     char *Name;
     struct ifnet *Retifnet;
     struct in_ifaddr *Retin_ifaddr;
{
  int i;
  static time_t last = 0;
  time_t now = time ((time_t *) 0);

  /* allow ``optimisation'' only for a period of one second: */
  if (last + 1 < now)
    {
      last = now;
      saveIndex = -1;
    }

  if (saveIndex != Index)
    {				/* Optimization! */
      Interface_Scan_Init();
      while (Interface_Scan_Next (&i, Name, Retifnet, Retin_ifaddr))
	{
	  if (i == Index)
	    break;
	}
      if (i != Index)
	return (-1);		/* Error, doesn't exist */
    }
  else
    {
      if (Retifnet)
	*Retifnet = saveifnet;
      if (Retin_ifaddr)
	*Retin_ifaddr = savein_ifaddr;
      if (Name)
	strcpy (Name, saveName);
    }
  return (0);			/* DONE */
}


static int Interface_Count = 0;

static int 
Interface_Scan_Get_Count()
{

  if (!Interface_Count)
    {
      Interface_Scan_Init();
      while (Interface_Scan_Next ((int *) 0, (char *) 0, (struct ifnet *) 0) != 0)
          Interface_Count++;
    }
  return (Interface_Count);
}


static int 
Interface_Get_Ether_By_Index (Index, EtherAddr)
     int Index;
     u_char *EtherAddr;
{
  int i;
  struct arpcom arpcom;

  if (saveIndex != Index)
    {				/* Optimization! */

      Interface_Scan_Init();
      while (Interface_Scan_Next ((int *) &i, (char *) 0, (struct ifnet *) 0, (struct in_ifaddr *) 0) != 0)
            if (i == Index)
               break;

      if (i != Index)
	return (-1);		/* Error, doesn't exist */
    }

  /*
   *  the arpcom structure is an extended ifnet structure which
   *  contains the ethernet address.
   */
  klookup ((int) saveifnetaddr, (char *) &arpcom, sizeof (struct arpcom));
  if (strncmp ("lo", saveName, 2) == 0)
    {
      /*
       *  Loopback doesn't have a HW addr, so return (0)0:00:00:00:00:00
       */
      bzero (EtherAddr, sizeof (arpcom.ac_enaddr));
    }
  else
    bcopy (&arpcom.ac_enaddr, EtherAddr, sizeof(arpcom.ac_enaddr));

  return (0);
}




u_char *
var_snmpStats (vp, name, length, exact, var_len, write_method)
     register struct variable *vp;	/* IN - pointer to variable entry that points here */
     oid *name;			/* IN/OUT - input name requested, output name found */
     int *length;		/* IN/OUT - length of input and output oid's */
     int exact;			/* IN - TRUE if an exact match was requested. */
     int *var_len;		/* OUT - length of variable or 0 if function returned. */
     int (**write_method)();   /* OUT - pointer to function to set variable, otherwise 0 */
{
  oid newname[MAX_NAME_LEN];
  int result;
  extern u_long snmpStats[];

  bcopy ((char *) vp->name, (char *) newname, (int) vp->namelen * sizeof (oid));
  newname[10] = 0;
  result = compare (name, *length, newname, (int) vp->namelen + 1);
  if ((exact && (result != 0)) || (!exact && (result >= 0)))
    return (NULL);
  bcopy ((char *) newname, (char *) name, ((int) vp->namelen + 1) * sizeof (oid));
  *length = vp->namelen + 1;

  *write_method = 0;
  *var_len = sizeof (long);	/* all following variables are sizeof long */

  switch (vp->magic)
    {
    case 1:
      return (u_char *) & snmpStats[SNMP_STAT_PACKETS];
    case 3:
      return (u_char *) & snmpStats[SNMP_STAT_ENCODING_ERRORS];
    case 11:
      return (u_char *) & snmpStats[SNMP_STAT_BAD_OPERATIONS];
    case 12:
      return (u_char *) & snmpStats[SNMP_STAT_SILENT_DROPS];
    case 13:
      return (u_char *) & snmpStats[SNMP_STAT_PROXY_DROPS];
    case 101:
      return (u_char *) & snmpStats[SNMP_STAT_V1_BAD_COMMUNITY_NAMES];
    case 102:
      return (u_char *) & snmpStats[SNMP_STAT_V1_BAD_COMMUNITY_USES];
    default:
      ERROR ("");
    }
  return (NULL);
}

u_char *
var_usecAgent (vp, name, length, exact, var_len, write_method)
     register struct variable *vp;	/* IN - pointer to variable entry that points here */
     oid *name;			/* IN/OUT - input name requested, output name found */
     int *length;		/* IN/OUT - length of input and output oid's */
     int exact;			/* IN - TRUE if an exact match was requested. */
     int *var_len;		/* OUT - length of variable or 0 if function returned. */
     int (**write_method)();   /* OUT - pointer to function to set variable, otherwise 0 */
{
  oid newname[MAX_NAME_LEN];
  int result;
  static u_long now;
  extern u_char _agentID[];
  extern u_long _agentBoots;
  extern u_long _agentStartTime;
  extern u_long _agentSize;

  bcopy ((char *) vp->name, (char *) newname, (int) vp->namelen * sizeof (oid));
  newname[10] = 0;
  result = compare (name, *length, newname, (int) vp->namelen + 1);
  if ((exact && (result != 0)) || (!exact && (result >= 0)))
    return (NULL);
  bcopy ((char *) newname, (char *) name, ((int) vp->namelen + 1) * sizeof (oid));
  *length = vp->namelen + 1;

  *write_method = 0;
  *var_len = sizeof (long);

  switch (vp->magic)
    {
    case 1:
      *var_len = 12;
      return (u_char *) _agentID;
    case 2:
      return (u_char *) & _agentBoots;
    case 3:
      now = _agentStartTime + time (NULL);
      return (u_char *) & now;
    case 4:
      return (u_char *) & _agentSize;
    default:
      ERROR ("");
    }
  return (NULL);
}

u_char *
var_usecStats (vp, name, length, exact, var_len, write_method)
     register struct variable *vp;	/* IN - pointer to variable entry that points here */
     oid *name;			/* IN/OUT - input name requested, output name found */
     int *length;		/* IN/OUT - length of input and output oid's */
     int exact;			/* IN - TRUE if an exact match was requested. */
     int *var_len;		/* OUT - length of variable or 0 if function returned. */
     int (**write_method)();   /* OUT - pointer to function to set variable, otherwise 0 */
{
  oid newname[MAX_NAME_LEN];
  int result;
  extern u_long snmpStats[];

  bcopy ((char *) vp->name, (char *) newname, (int) vp->namelen * sizeof (oid));
  newname[10] = 0;
  result = compare (name, *length, newname, (int) vp->namelen + 1);
  if ((exact && (result != 0)) || (!exact && (result >= 0)))
    return (NULL);
  bcopy ((char *) newname, (char *) name, ((int) vp->namelen + 1) * sizeof (oid));
  *length = vp->namelen + 1;

  *write_method = 0;
  *var_len = sizeof (long);	/* all following variables are sizeof long */

  switch (vp->magic)
    {
    case 1:
      return (u_char *) & snmpStats[USEC_STAT_UNSUPPORTED_QOS];
    case 2:
      return (u_char *) & snmpStats[USEC_STAT_NOT_IN_WINDOWS];
    case 3:
      return (u_char *) & snmpStats[USEC_STAT_UNKNOWN_USERNAMES];
    case 4:
      return (u_char *) & snmpStats[USEC_STAT_WRONG_DIGEST_VALUES];
    case 5:
      return (u_char *) & snmpStats[USEC_STAT_UNKNOWN_CONTEXT_SELECTORS];
    case 6:
      return (u_char *) & snmpStats[USEC_STAT_BAD_PARAMETERS];
    case 7:
      return (u_char *) & snmpStats[USEC_STAT_UNAUTHORIZED_OPERATIONS];
    default:
      ERROR ("");
    }
  return (NULL);
}
