/* 
 * Abstract Syntax Notation One, ASN.1
 * As defined in ISO/IS 8824 and ISO/IS 8825
 * This implements a subset of the above International Standards that
 * is sufficient to implement SNMP.
 *
 * Encodes abstract data types into a machine independent stream of bytes.
 *
 */
/***************************************************************************
 *
 *           Copyright 1998 by Carnegie Mellon University
 * 
 *                       All Rights Reserved
 * 
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose and without fee is hereby granted,
 * provided that the above copyright notice appear in all copies and that
 * both that copyright notice and this permission notice appear in
 * supporting documentation, and that the name of CMU not be
 * used in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.
 * 
 * CMU DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING
 * ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL
 * CMU BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR
 * ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
 * ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
 * SOFTWARE.
 *
 * Changes by G. Vanem 1998
 * Adapted to Waterloo tcp/ip
 ***************************************************************************/

#include "snmp_lib.h"

BYTE *asn_build_header (BYTE *data,         /* IN - ptr to start of object */
                        int  *datalength,   /* IN/OUT - # of valid bytes */
                                            /*          left in buffer */
                        BYTE type,          /* IN - ASN type of object */
                        int  length)        /* IN - length of object */
{
  /* Truth is 0 'cause we don't know yet */
  return (asn_build_header_with_truth (data, datalength, type, length, 0));
}


/* 
 * asn_parse_int - pulls an int out of an ASN int type.
 *  On entry, datalength is input as the number of valid bytes following
 *   "data".  On exit, it is returned as the number of valid bytes
 *   following the end of this object.
 *
 *  Returns a pointer to the first byte past the end
 *   of this object (i.e. the start of the next object).
 *  Returns NULL on any error.
 */
BYTE *asn_parse_int (
      BYTE *data,       /* IN     - pointer to start of object */
      int  *datalength, /* IN/OUT - # of valid bytes left in buffer */
      BYTE *type,       /* OUT    - asn type of object */
      int  *intp,       /* IN/OUT - pointer to start of output buffer */
      int   intsize)    /* IN     - size of output buffer */
{
  /*
   * ASN.1 integer ::= 0x02 asnlength byte {byte}*
   */
  BYTE  *bufp  = data;
  int    value = 0;
  DWORD  asn_length;

  /* Room to store int?
   */
  if (intsize != sizeof(int))
  {
    snmp_set_api_error (SNMPERR_ASN_DECODE);
    return (NULL);
  }

  /* Type
   */
  *type = *bufp++;

  /* Extract length
   */
  bufp = asn_parse_length (bufp, &asn_length);
  if (!bufp)
     return (NULL);

  /* Make sure the entire int is in the buffer
   */
  if (asn_length + (bufp - data) > (DWORD)*datalength)
  {
    snmp_set_api_error (SNMPERR_ASN_DECODE);
    return (NULL);
  }

  /* Can we store this int?
   */
  if (asn_length > (DWORD)intsize)
  {
    snmp_set_api_error (SNMPERR_ASN_DECODE);
    return (NULL);
  }

  /* Remaining data
   */
  *datalength -= (int) asn_length + (bufp - data);

  /* Is the int negative?
   */
  if (*bufp & 0x80)
    value = -1;                        /* integer is negative */

  /* Extract the bytes
   */
  while (asn_length--)
     value = (value << 8) | *bufp++;

  *intp = value;
  return (bufp);
}

/* 
 * asn_parse_unsigned_int - pulls an unsigned int out of an ASN int type.
 *  On entry, datalength is input as the number of valid bytes following
 *   "data".  On exit, it is returned as the number of valid bytes
 *   following the end of this object.
 *
 *  Returns a pointer to the first byte past the end
 *   of this object (i.e. the start of the next object).
 *  Returns NULL on any error.
 */
BYTE *asn_parse_unsigned_int (
      BYTE  *data,        /* IN     - pointer to start of object */
      int   *datalength,  /* IN/OUT - # of valid bytes left in buffer */
      BYTE  *type,        /* OUT    - asn type of object */
      DWORD *intp,        /* IN/OUT - pointer to start of output buffer */
      int    intsize)     /* IN     - size of output buffer */
{
  /*
   * ASN.1 integer ::= 0x02 asnlength byte {byte}*
   */
  BYTE *bufp  = data;
  int   value = 0;
  DWORD asn_length;

  /* Room to store int? */
  if (intsize != sizeof(int))
  {
    snmp_set_api_error (SNMPERR_ASN_DECODE);
    return (NULL);
  }

  /* Type */
  *type = *bufp++;

  /* Extract length */
  bufp = asn_parse_length (bufp, &asn_length);
  if (!bufp)
    return (NULL);

  /* Make sure the entire int is in the buffer */
  if (asn_length + (bufp - data) > (DWORD)*datalength)
  {
    snmp_set_api_error (SNMPERR_ASN_DECODE);
    return (NULL);
  }

  /* Can we store this int? */
  if ((asn_length > ((DWORD)intsize + 1)) ||
      ((asn_length == (DWORD)intsize + 1) && *bufp != 0))
  {
    snmp_set_api_error (SNMPERR_ASN_DECODE);
    return (NULL);
  }

  /* Remaining data */
  *datalength -= (int) asn_length + (bufp - data);

  /* Is the int negative? */
  if (*bufp & 0x80)
    value = -1;                        /* integer is negative */

  /* Extract the bytes */
  while (asn_length--)
    value = (value << 8) | *bufp++;

  /* That's it! */
  *intp = value;
  return (bufp);
}

/* 
 * asn_build_int - builds an ASN object containing an integer.
 *  On entry, datalength is input as the number of valid bytes following
 *   "data".  On exit, it is returned as the number of valid bytes
 *   following the end of this object.
 *
 *  Returns a pointer to the first byte past the end
 *   of this object (i.e. the start of the next object).
 *  Returns NULL on any error.
 */
BYTE *asn_build_int (
      BYTE *data,       /* IN - pointer to start of output buffer */
      int  *datalength, /* IN/OUT - # of valid bytes left in buffer */
      BYTE  type,       /* IN - asn type of object */
      int  *intp,       /* IN - pointer to start of integer */
      int   intsize)    /* IN - size of *intp */
{
  /*
   * ASN.1 integer ::= 0x02 asnlength byte {byte}*
   */
  int   integer;
  DWORD mask;

  if (intsize != sizeof (int))
  {
    snmp_set_api_error (SNMPERR_ASN_ENCODE);
    return (NULL);
  }

  integer = *intp;

  /*
   * Truncate "unnecessary" bytes off of the most significant end of this
   * 2's complement integer.  There should be no sequence of 9
   * consecutive 1's or 0's at the most significant end of the
   * integer.
   */
  mask = 0x1FFUL << ((8 * (sizeof(int) - 1)) - 1);

  /* mask is 0xFF800000 on a big-endian machine */

  while ((((integer & mask) == 0) || ((integer & mask) == mask)) &&
         intsize > 1)
  {
    intsize--;
    integer <<= 8;
  }

  data = asn_build_header_with_truth (data, datalength, type, intsize, 1);
  if (!data)
    return (NULL);

  /* Enough room for what we just encoded? */
  if (*datalength < intsize)
  {
    snmp_set_api_error (SNMPERR_ASN_ENCODE);
    return (NULL);
  }

  /* Insert it */
  *datalength -= intsize;
  mask = 0xFFUL << (8 * (sizeof(int) - 1));

  /* mask is 0xFF000000 on a big-endian machine */
  while (intsize--)
  {
    *data++ = (BYTE) ((integer & mask) >> (8 * (sizeof(int) - 1)));

    integer <<= 8;
  }
  return (data);
}

/* 
 * asn_build_unsigned_int - builds an ASN object containing an integer.
 *  On entry, datalength is input as the number of valid bytes following
 *   "data".  On exit, it is returned as the number of valid bytes
 *   following the end of this object.
 *
 *  Returns a pointer to the first byte past the end
 *   of this object (i.e. the start of the next object).
 *  Returns NULL on any error.
 */
BYTE *asn_build_unsigned_int (
      BYTE  *data,        /* IN     - pointer to start of output buffer */
      int   *datalength,  /* IN/OUT - # of valid bytes left in buffer */
      BYTE   type,        /* IN     - asn type of object */
      DWORD *intp,        /* IN     - pointer to start of integer */
      int    intsize)     /* IN     - size of *intp */
{
  /*
   * ASN.1 integer ::= 0x02 asnlength byte {byte}*
   */
  DWORD integer;
  DWORD mask;
  int   add_null_byte = 0;

  if (intsize != sizeof (int))
  {
    snmp_set_api_error (SNMPERR_ASN_ENCODE);
    return (NULL);
  }

  integer = *intp;
  mask = 0xFFUL << (8 * (sizeof(int) - 1));

  /* mask is 0xFF000000 on a big-endian machine */
  if ((BYTE) ((integer & mask) >> (8 * (sizeof(int) - 1))) & 0x80)
  {
    /* if MSB is set */
    add_null_byte = 1;
    intsize++;
  }

  /*
   * Truncate "unnecessary" bytes off of the most significant end of
   * this 2's complement integer. 
   * There should be no sequence of 9 consecutive 1's or 0's at the
   * most significant end of the integer.
   */
  mask = 0x1FFUL << ((8 * (sizeof(int) - 1)) - 1);

  /* mask is 0xFF800000 on a big-endian machine */
  while ((((integer & mask) == 0) ||
          ((integer & mask) == mask)) && intsize > 1)
  {
    intsize--;
    integer <<= 8;
  }

  data = asn_build_header_with_truth (data, datalength, type, intsize, 1);
  if (!data)
    return (NULL);

  if (*datalength < intsize)
  {
    snmp_set_api_error (SNMPERR_ASN_ENCODE);
    return (NULL);
  }

  *datalength -= intsize;
  if (add_null_byte == 1)
  {
    *data++ = '\0';
    intsize--;
  }

  mask = 0xFFUL << (8 * (sizeof(int) - 1));

  /* mask is 0xFF000000 on a big-endian machine */
  while (intsize--)
  {
    *data++ = (BYTE) ((integer & mask) >> (8 * (sizeof (int) - 1)));

    integer <<= 8;
  }
  return (data);
}

/* 
 * asn_parse_string - pulls an octet string out of an ASN octet string type.
 *  On entry, datalength is input as the number of valid bytes following
 *   "data".  On exit, it is returned as the number of valid bytes
 *   following the beginning of the next object.
 *
 *  "string" is filled with the octet string.
 *
 *  Returns a pointer to the first byte past the end
 *   of this object (i.e. the start of the next object).
 *  Returns NULL on any error.
 */
BYTE *asn_parse_string (
      BYTE *data,       /* IN - pointer to start of object */
      int  *datalength, /* IN/OUT - # of valid bytes left in buffer */
      BYTE *type,       /* OUT - asn type of object */
      BYTE *string,     /* IN/OUT - pointer to start of output buffer */
      int  *strlength)  /* IN/OUT - size of output buffer */
{
  /*
   * ASN.1 octet string ::= primstring | cmpdstring
   * primstring ::= 0x04 asnlength byte {byte}*
   * cmpdstring ::= 0x24 asnlength string {string}*
   */
  BYTE  *bufp = data;
  DWORD  asn_length;

  *type = *bufp++;
  bufp  = asn_parse_length (bufp, &asn_length);
  if (!bufp)
    return (NULL);

  if (asn_length + (bufp - data) > (DWORD)*datalength)
  {
    snmp_set_api_error (SNMPERR_ASN_DECODE);
    return (NULL);
  }

  if (asn_length > (DWORD)*strlength)
  {
    snmp_set_api_error (SNMPERR_ASN_DECODE);
    return (NULL);
  }

  memcpy (string, bufp, asn_length);
  *strlength   = (int) asn_length;
  *datalength -= (int) asn_length + (bufp - data);
  return (bufp + asn_length);
}

/* 
 * asn_build_string - Builds an ASN octet string object containing the input
 *   string.  On entry, datalength is input as the number of valid bytes 
 *   following "data".  On exit, it is returned as the number of valid bytes
 *   following the beginning of the next object.
 *
 *  Returns a pointer to the first byte past the end
 *   of this object (i.e. the start of the next object).
 *  Returns NULL on any error.
 */
BYTE *asn_build_string (
      BYTE *data,       /* IN - pointer to start of object */
      int  *datalength, /* IN/OUT - # of valid bytes left in buf */
      BYTE  type,       /* IN - ASN type of string */
      BYTE *string,     /* IN - pointer to start of input buffer */
      int   strlength)  /* IN - size of input buffer */
{
  /*
   * ASN.1 octet string ::= primstring | cmpdstring
   * primstring ::= 0x04 asnlength byte {byte}*
   * cmpdstring ::= 0x24 asnlength string {string}*
   * This code will never send a compound string.
   */
  data = asn_build_header_with_truth (data, datalength, type, strlength, 1);
  if (!data)
     return (NULL);

  if (*datalength < strlength)
  {
    snmp_set_api_error (SNMPERR_ASN_DECODE);
    return (NULL);
  }

  memcpy (data, string, strlength);
  *datalength -= strlength;
  return (data + strlength);
}


/* 
 * asn_parse_header - interprets the ID and length of the current object.
 *  On entry, datalength is input as the number of valid bytes following
 *   "data".  On exit, it is returned as the number of valid bytes
 *   in this object following the id and length.
 *
 *  Returns a pointer to the first byte of the contents of this object.
 *  Returns NULL on any error.
 */
BYTE *asn_parse_header (
      BYTE *data,       /* IN - pointer to start of object */
      int  *datalength, /* IN/OUT - # of valid bytes left in buffer */
      BYTE *type)       /* OUT - ASN type of object */
{
  BYTE *bufp = data;
  int   header_len;
  DWORD asn_length;

  /* this only works on data types < 30, i.e. no extension octets */
  if (IS_EXTENSION_ID (*bufp))
  {
    snmp_set_api_error (SNMPERR_ASN_DECODE);
    return (NULL);
  }

  *type = *bufp;
  bufp = asn_parse_length (bufp + 1, &asn_length);
  if (!bufp)
     return (NULL);

  header_len = bufp - data;
  if (header_len + asn_length > (DWORD)*datalength)
  {
    snmp_set_api_error (SNMPERR_ASN_DECODE);
    return (NULL);
  }

  *datalength = (int) asn_length;
  return (bufp);
}

/* 
 * asn_build_header - builds an ASN header for an object with the ID and
 * length specified.
 *  On entry, datalength is input as the number of valid bytes following
 *   "data".  On exit, it is returned as the number of valid bytes
 *   in this object following the id and length.
 *
 *  This only works on data types < 30, i.e. no extension octets.
 *  The maximum length is 0xFFFF;
 *
 *  Returns a pointer to the first byte of the contents of this object.
 *  Returns NULL on any error.
 */

BYTE *asn_build_header_with_truth (
      BYTE *data,        /* IN - pointer to start of object */
      int  *datalength,  /* IN/OUT - # of valid bytes left in buffer */
      BYTE  type,        /* IN - ASN type of object */
      int   length,      /* IN - length of object */
      int   truth)       /* IN - Whether length is truth */
{
  if (*datalength < 1)
  {
    snmp_set_api_error (SNMPERR_ASN_ENCODE);
    return (NULL);
  }

  *data++ = type;
  (*datalength)--;
  return (asn_build_length (data, datalength, length, truth));
}

#if 0
/* 
 * asn_build_sequence - builds an ASN header for a sequence with the ID and
 * length specified.
 *  On entry, datalength is input as the number of valid bytes following
 *   "data".  On exit, it is returned as the number of valid bytes
 *   in this object following the id and length.
 *
 *  This only works on data types < 30, i.e. no extension octets.
 *  The maximum length is 0xFFFF;
 *
 *  Returns a pointer to the first byte of the contents of this object.
 *  Returns NULL on any error.
 */
static BYTE *asn_build_sequence (
       BYTE *data,        /* IN - pointer to start of object */
       int  *datalength,  /* IN/OUT - # of valid bytes left in buffer */
       BYTE  type,        /* IN - ASN type of object */
       int   length)      /* IN - length of object */
{
  *datalength -= 4;
  if (*datalength < 0)
  {
    *datalength += 4;                /* fix up before punting */
    snmp_set_api_error (SNMPERR_ASN_ENCODE);
    return (NULL);
  }

  *data++ = type;
  *data++ = (BYTE) (0x02 | ASN_LONG_LEN);
  *data++ = (BYTE) ((length >> 8) & 0xFF);
  *data++ = (BYTE) (length & 0xFF);
  return (data);
}
#endif

/*
 * asn_parse_length - interprets the length of the current object.
 *  On exit, length contains the value of this length field.
 *
 *  Returns a pointer to the first byte after this length
 *  field (aka: the start of the data field).
 *  Returns NULL on any error.
 */
BYTE *asn_parse_length (
      BYTE  *data,    /* IN  - pointer to start of length field */
      DWORD *length)  /* OUT - value of length field */
{
  BYTE lengthbyte = *data;

  if (lengthbyte & ASN_LONG_LEN)
  {
    lengthbyte &= ~ASN_LONG_LEN;        /* turn MSb off */

    if (lengthbyte == 0)
    {
      snmp_set_api_error (SNMPERR_ASN_DECODE);
      return (NULL);
    }

    if (lengthbyte > sizeof (int))
    {
      snmp_set_api_error (SNMPERR_ASN_DECODE);
      return (NULL);
    }

    *length = 0UL;
    memcpy (length, data + 1, lengthbyte);
    *length = ntohl (*length);
    *length >>= (8 * ((sizeof *length) - lengthbyte));
    return (data + lengthbyte + 1);
  }

  /* short ASN-length
   */
  *length = (int) lengthbyte;
  return (data + 1);
}


BYTE *asn_build_length (
      BYTE *data,       /* IN - pointer to start of object */
      int  *datalength, /* IN/OUT - # of valid bytes left in buf */
      int   length,     /* IN - length of object */
      int   truth)      /* IN - If 1, this is the true len. */
{
  BYTE *start_data = data;

  if (truth)
  {
    /* no indefinite lengths sent */
    if (length < 0x80)
    {
      if (*datalength < 1)
      {
        snmp_set_api_error (SNMPERR_ASN_ENCODE);
        return (NULL);
      }
      *data++ = (BYTE) length;

    }
    else if (length <= 0xFF)
    {
      if (*datalength < 2)
      {
        snmp_set_api_error (SNMPERR_ASN_ENCODE);
        return (NULL);
      }
      *data++ = (BYTE) (0x01 | ASN_LONG_LEN);
      *data++ = (BYTE) length;
    }
    else
    {                                /* 0xFF < length <= 0xFFFF */
      if (*datalength < 3)
      {
        snmp_set_api_error (SNMPERR_ASN_ENCODE);
        return (NULL);
      }
      *data++ = (BYTE) (0x02 | ASN_LONG_LEN);
      *data++ = (BYTE) ((length >> 8) & 0xFF);
      *data++ = (BYTE) (length & 0xFF);
    }

  }
  else
  {
    /* Don't know if this is the true length.  Make sure it's large
     * enough for later.
     */
    if (*datalength < 3)
    {
      snmp_set_api_error (SNMPERR_ASN_ENCODE);
      return (NULL);
    }
    *data++ = (BYTE) (0x02 | ASN_LONG_LEN);
    *data++ = (BYTE) ((length >> 8) & 0xFF);
    *data++ = (BYTE) (length & 0xFF);
  }

  *datalength -= (data - start_data);
  return (data);
}

/* 
 * asn_parse_objid - pulls an object indentifier out of an ASN object
 * identifier type.
 *  On entry, datalength is input as the number of valid bytes following
 *   "data".  On exit, it is returned as the number of valid bytes
 *   following the beginning of the next object.
 *
 *  "objid" is filled with the object identifier.
 *
 *  Returns a pointer to the first byte past the end
 *   of this object (i.e. the start of the next object).
 *  Returns NULL on any error.
 */
BYTE *asn_parse_objid (
      BYTE *data,        /* IN     - pointer to start of object */
      int  *datalength,  /* IN/OUT - # of valid bytes left in buf */
      BYTE *type,        /* OUT    - ASN type of object */
      oid  *objid,       /* IN/OUT - pointer to start of output buffer */
      int  *objidlength) /* IN/OUT - number of sub-id's in objid */
{
  /*
   * ASN.1 objid ::= 0x06 asnlength subidentifier {subidentifier}*
   * subidentifier ::= {leadingbyte}* lastbyte
   * leadingbyte ::= 1 7bitvalue
   * lastbyte ::= 0 7bitvalue
   */
  BYTE  *bufp = data;
  oid   *oidp = objid + 1;
  DWORD  subidentifier;
  int    length;
  DWORD  asn_length;

  *type = *bufp++;
  bufp  = asn_parse_length (bufp, &asn_length);
  if (!bufp)
     return (NULL);

  if (asn_length + (bufp - data) > (DWORD)*datalength)
  {
    snmp_set_api_error (SNMPERR_ASN_DECODE);
    return (NULL);
  }
  *datalength -= (int) asn_length + (bufp - data);

  /* Handle invalid object identifier encodings of the form 06 00 robustly
   */
  if (asn_length == 0)
     objid[0] = objid[1] = 0;

  length = asn_length;
  (*objidlength)--;                /* account for expansion of first byte */

  while (length > 0 && (*objidlength)-- > 0)
  {
    subidentifier = 0;

    do
    {                                /* shift and add in low order 7 bits */
      subidentifier = (subidentifier << 7)
                    + (*(BYTE*)bufp & ~ASN_BIT8);
      length--;
    }
    while (*(BYTE*)bufp++ & ASN_BIT8);

    /* while last byte has high bit clear
     */
    if (subidentifier > (DWORD)MAX_SUBID)
    {
      snmp_set_api_error (SNMPERR_ASN_DECODE);
      return (NULL);
    }
    *oidp++ = (oid) subidentifier;
  }

  /*
   * The first two subidentifiers are encoded into the first component
   * with the value (X * 40) + Y, where:
   *  X is the value of the first subidentifier.
   *  Y is the value of the second subidentifier.
   */
  subidentifier = (DWORD)objid[1];
  if (subidentifier == 0x2B)
  {
    objid[0] = 1;
    objid[1] = 3;
  }
  else
  {
    objid[1] = (BYTE) (subidentifier % 40);
    objid[0] = (BYTE) ((subidentifier - objid[1]) / 40);
  }

  *objidlength = (int) (oidp - objid);
  return (bufp);
}

/* 
 * asn_build_objid - Builds an ASN object identifier object containing the
 * input string.
 *  On entry, datalength is input as the number of valid bytes following
 *   "data".  On exit, it is returned as the number of valid bytes
 *   following the beginning of the next object.
 *
 *  Returns a pointer to the first byte past the end
 *   of this object (i.e. the start of the next object).
 *  Returns NULL on any error.
 */
BYTE *asn_build_objid (
      BYTE *data,         /* IN - pointer to start of object */
      int  *datalength,   /* IN/OUT - # of valid bytes left in buf */
      BYTE  type,         /* IN - ASN type of object */
      oid  *objid,        /* IN - pointer to start of input buffer */
      int   objidlength)  /* IN - number of sub-id's in objid */
{
  /*
   * ASN.1 objid ::= 0x06 asnlength subidentifier {subidentifier}*
   * subidentifier ::= {leadingbyte}* lastbyte
   * leadingbyte ::= 1 7bitvalue
   * lastbyte ::= 0 7bitvalue
   */
  BYTE  buf[MAX_OID_LEN];
  BYTE *bp = buf;
  oid  *op = objid;
  int   asnlength;
  DWORD subid, mask, testmask;
  int   bits, testbits;

  if (objidlength < 2)
  {
    *bp++ = 0;
    objidlength = 0;
  }
  else
  {
    *bp++ = op[1] + (op[0] * 40);
    objidlength -= 2;
    op += 2;
  }

  while (objidlength-- > 0)
  {
    subid = *op++;
    if (subid < 127)
    {                                /* off by one? */
      *bp++ = subid;
    }
    else
    {
      mask = 0x7F;                /* handle subid == 0 case */
      bits = 0;
      /* testmask *MUST* !!!! be of an unsigned type */
      for (testmask = 0x7F, testbits = 0; testmask != 0;
           testmask <<= 7, testbits += 7)
      {
        if (subid & testmask)
        {                        /* if any bits set */
          mask = testmask;
          bits = testbits;
        }
      }
      /* mask can't be zero here */
      for (; mask != 0x7F; mask >>= 7, bits -= 7)
      {
        /* fix a mask that got truncated above */
        if (mask == 0x1E00000)
          mask = 0xFE00000;
        *bp++ = (BYTE) (((subid & mask) >> bits) | ASN_BIT8);
      }
      *bp++ = (BYTE) (subid & mask);
    }
  }

  asnlength = bp - buf;
  data = asn_build_header_with_truth (data, datalength, type, asnlength, 1);
  if (!data)
     return (NULL);

  if (*datalength < asnlength)
  {
    snmp_set_api_error (SNMPERR_ASN_DECODE);
    return (NULL);
  }
  memcpy (data, buf, asnlength);
  *datalength -= asnlength;
  return (data + asnlength);
}

/* 
 * asn_parse_null - Interprets an ASN null type.
 *  On entry, datalength is input as the number of valid bytes following
 *   "data".  On exit, it is returned as the number of valid bytes
 *   following the beginning of the next object.
 *
 *  Returns a pointer to the first byte past the end
 *   of this object (i.e. the start of the next object).
 *  Returns NULL on any error.
 */
BYTE *asn_parse_null (
      BYTE *data,       /* IN - pointer to start of object */
      int  *datalength, /* IN/OUT - # of valid bytes left in buf */
      BYTE *type)       /* OUT - ASN type of object */
{
  /*
   * ASN.1 null ::= 0x05 0x00
   */
  BYTE  *bufp = data;
  DWORD  asn_length;

  *type = *bufp++;
  bufp = asn_parse_length (bufp, &asn_length);
  if (!bufp)
    return (NULL);

  if (asn_length)
  {
    snmp_set_api_error (SNMPERR_ASN_DECODE);
    return (NULL);
  }

  *datalength -= (bufp - data);
  return (bufp + asn_length);
}

/* 
 * asn_build_null - Builds an ASN null object.
 *  On entry, datalength is input as the number of valid bytes following
 *   "data".  On exit, it is returned as the number of valid bytes
 *   following the beginning of the next object.
 *
 *  Returns a pointer to the first byte past the end
 *   of this object (i.e. the start of the next object).
 *  Returns NULL on any error.
 */
BYTE *asn_build_null (
      BYTE *data,       /* IN - pointer to start of object */
      int  *datalength, /* IN/OUT - # of valid bytes left in buf */
      BYTE  type)       /* IN - ASN type of object */
{
  /*
   * ASN.1 null ::= 0x05 0x00
   */
  return (asn_build_header_with_truth (data, datalength, type, 0, 1));
}

/* 
 * asn_parse_bitstring - pulls a bitstring out of an ASN bitstring type.
 *  On entry, datalength is input as the number of valid bytes following
 *   "data".  On exit, it is returned as the number of valid bytes
 *   following the beginning of the next object.
 *
 *  "string" is filled with the bit string.
 *
 *  Returns a pointer to the first byte past the end
 *   of this object (i.e. the start of the next object).
 *  Returns NULL on any error.
 */
BYTE *asn_parse_bitstring (
      BYTE *data,         /* IN - pointer to start of object */
      int  *datalength,   /* IN/OUT - # of valid bytes left in buf */
      BYTE *type,         /* OUT - asn type of object */
      BYTE *string,       /* IN/OUT - pointer to start of output buf */
      int  *strlength)    /* IN/OUT - size of output buffer */
{
  /*
   * bitstring ::= 0x03 asnlength unused {byte}*
   */
  BYTE  *bufp = data;
  DWORD  asn_length;

  *type = *bufp++;
  bufp = asn_parse_length (bufp, &asn_length);
  if (!bufp)
    return (NULL);

  if (asn_length + (bufp - data) > (DWORD)*datalength)
  {
    snmp_set_api_error (SNMPERR_ASN_DECODE);
    return (NULL);
  }

  if (asn_length > (DWORD)*strlength)
  {
    snmp_set_api_error (SNMPERR_ASN_DECODE);
    return (NULL);
  }

  if (asn_length < 1)
  {
    snmp_set_api_error (SNMPERR_ASN_DECODE);
    return (NULL);
  }

  if (*bufp > 7)
  {
    snmp_set_api_error (SNMPERR_ASN_DECODE);
    return (NULL);
  }

  memcpy (string, bufp, asn_length);
  *strlength   = (int) asn_length;
  *datalength -= (int) asn_length + (bufp - data);
  return (bufp + asn_length);
}

/* 
 * asn_build_bitstring - Builds an ASN bit string object containing the
 * input string.
 *  On entry, datalength is input as the number of valid bytes following
 *   "data".  On exit, it is returned as the number of valid bytes
 *   following the beginning of the next object.
 *
 *  Returns a pointer to the first byte past the end
 *   of this object (i.e. the start of the next object).
 *  Returns NULL on any error.
 */
BYTE *asn_build_bitstring (
      BYTE *data,        /* IN - pointer to start of object */
      int  *datalength,  /* IN/OUT - # of valid bytes left in buf */
      BYTE  type,        /* IN - ASN type of string */
      BYTE *string,      /* IN - pointer to start of input buffer */
      int   strlength)   /* IN - size of input buffer */
{
  /*
   * ASN.1 bit string ::= 0x03 asnlength unused {byte}*
   */
  if (strlength < 1 || *string > 7)
  {
    snmp_set_api_error (SNMPERR_ASN_ENCODE);
    return (NULL);
  }

  data = asn_build_header_with_truth (data, datalength, type, strlength, 1);
  if (!data)
    return (NULL);

  if (*datalength < strlength)
  {
    snmp_set_api_error (SNMPERR_ASN_ENCODE);
    return (NULL);
  }

  memcpy (data, string, strlength);
  *datalength -= strlength;
  return (data + strlength);
}


/* 
 * To do: Write an asn_parse_exception function to go with the new 
 * asn_build_exception function below so that the exceptional values can
 * be handled in input packets aswell as output ones.
 */

/* 
 * asn_build_exception - Builds an ASN exception object.
 *  On entry, datalength is input as the number of valid bytes following
 *   "data".  On exit, it is returned as the number of valid bytes
 *   following the beginning of the next object.
 *
 *  Returns a pointer to the first byte past the end
 *   of this object (i.e. the start of the next object).
 *  Returns NULL on any error.
 *
 * ASN.1 variable exception ::= 0x8i 0x00, where 'i' is one of these 
 *                                         exception identifiers:
 *                                           0 -- noSuchObject
 *                                           1 -- noSuchInstance
 *                                           2 -- endOfMibView
 */
BYTE *asn_build_exception (
      BYTE *data,        /* IN     - pointer to start of object */
      int  *datalength,  /* IN/OUT - # of valid bytes left in buf */
      BYTE  type)        /* IN     - ASN type of object */
{
  return (asn_build_header_with_truth (data, datalength, type, 0, 1));
}
