/**********************************************************************
 *
 *           Copyright 1998 by Carnegie Mellon University
 * 
 *                       All Rights Reserved
 * 
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose and without fee is hereby granted,
 * provided that the above copyright notice appear in all copies and that
 * both that copyright notice and this permission notice appear in
 * supporting documentation, and that the name of CMU not be
 * used in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.
 * 
 * CMU DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING
 * ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL
 * CMU BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR
 * ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
 * ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
 * SOFTWARE.
 * 
 * $Id: snmp_client.c,v 1.12 1998/05/06 04:07:25 ryan Exp $
 *
 * Changes by G. Vanem 1998
 * Adapted to Waterloo tcp/ip
 **********************************************************************/

#include "snmp_lib.h"

int snmp_synch_input (int Operation, struct snmp_session *Session,
                      int RequestID, struct snmp_pdu *pdu, void *magic)
{
  struct variable_list *var;
  struct synch_state   *state = (struct synch_state *) magic;
  struct snmp_pdu      *newpdu;
  struct variable_list *varPtr;

#if DEBUG_CLIENT
  printf ("CLIENT %p: Synchronizing input.\n", pdu);
#endif

  /* Make sure this is the proper request
   */
  if (RequestID != state->reqid)
     return (0);

  state->waiting = 0;

  /* Did we receive a Get Response
   */
  if (Operation == RECEIVED_MESSAGE && pdu->command == SNMP_PDU_RESPONSE)
  {  
    /* clone the pdu */
    state->pdu = newpdu = snmp_pdu_clone (pdu);
    newpdu->variables = 0;

    /* Clone all variables */
    var = pdu->variables;
    if (var)
    {
      newpdu->variables = snmp_var_clone (var);
      varPtr = newpdu->variables;

      /* While there are more variables */
      while (var->next_variable)
      {
        /* Clone the next one */
        varPtr->next_variable = snmp_var_clone (var->next_variable);

        /* And move on */
        var = var->next_variable;
        varPtr = varPtr->next_variable;

      }
      varPtr->next_variable = NULL;
    }
    state->status = STAT_SUCCESS;
    snmp_errno = 0;             /* XX all OK when msg received ? */
    Session->s_snmp_errno = 0;
  }
  else if (Operation == TIMED_OUT)
  {
    state->pdu = NULL;          /* XX from other lib ? */
    state->status = STAT_TIMEOUT;
    snmp_errno = SNMPERR_NO_RESPONSE;
    Session->s_snmp_errno = SNMPERR_NO_RESPONSE;
  }
  return (1);
}


int snmp_synch_response (struct snmp_session *Session,
                         struct snmp_pdu     *PDU,
                         struct snmp_pdu    **ResponsePDUP)
{
  struct synch_state  *state = Session->snmp_synch_state;
  struct session_list *slp;
  struct timeval       timeout, *tvp;
  fd_set fdset;
  int    block, numfds, count;

  state->reqid = snmp_send (Session, PDU);
  if (state->reqid == 0)
  {
    *ResponsePDUP = NULL;
    snmp_free_pdu (PDU);
    return (STAT_ERROR);
  }
  state->waiting = 1;

  while (state->waiting)
  {
    FD_ZERO (&fdset);
    numfds = 0;
    block  = 1;
    tvp    = &timeout;
    timerclear (tvp);
    tcp_tick (NULL);
    printf ("!\n");      // !!

    snmp_select_info (&numfds, &fdset, tvp, &block);
    if (block == 1)
       tvp = NULL;          /* block without timeout */

    count = 0;
    for (slp = Sessions; slp; slp = slp->next)
    {
      tcp_tick (slp->session->socket);
      if (sock_dataready(slp->session->socket))
      {
        FD_SET (slp->internal->sd, &fdset);
        count++;
      }
    }

    if (count > 0)
       snmp_read (&fdset);
    else
      switch (count)
      {
        case 0:
             snmp_timeout();
             break;
        case -1:
             snmp_errno = SNMPERR_GENERR;
             /* fall through */
        default:
             snmp_free_pdu (PDU);
             *ResponsePDUP = NULL;
             return (STAT_ERROR);
      }
  }
  *ResponsePDUP = state->pdu;
  return (state->status);
}

int snmp_sess_synch_response (void *sessp,
                              struct snmp_pdu *PDU,
                              struct snmp_pdu **ResponsePDUP)
{
  struct snmp_session *Session = snmp_sess_session (sessp);
  struct synch_state  *state   = Session->snmp_synch_state;
  struct session_list *slp;
  struct timeval       timeout, *tvp;
  int    numfds, count, block;
  fd_set fdset;

  state->reqid = snmp_sess_send (sessp, PDU);
  if (state->reqid == 0)
  {
    *ResponsePDUP = NULL;
    snmp_free_pdu (PDU);
    return (STAT_ERROR);
  }
  state->waiting = 1;

  while (state->waiting)
  {
    numfds = 0;
    FD_ZERO (&fdset);
    block = 1;
    tvp   = &timeout;
    timerclear (tvp);
    tcp_tick (NULL);
    printf ("!\n");      // !! test

    snmp_sess_select_info (sessp, &numfds, &fdset, tvp, &block);
    if (block == 1)
      tvp = NULL;                /* block without timeout */

    count = 0;
    for (slp = Sessions; slp; slp = slp->next)
    {
      tcp_tick (slp->session->socket);
      if (sock_dataready(slp->session->socket))
      {
        FD_SET (slp->internal->sd,&fdset);
        count++;
      }
    }
    if (count > 0)
       snmp_sess_read (sessp,&fdset);
    else
      switch (count)
      {
        case 0:
             snmp_sess_timeout (sessp);
             break;
        case -1:
             snmp_errno = SNMPERR_GENERR;
             /* fall through */
        default:
             snmp_free_pdu (PDU);
             *ResponsePDUP = NULL;
             return (STAT_ERROR);
      }
  }
  *ResponsePDUP = state->pdu;
  return (state->status);
}

void snmp_synch_reset (struct snmp_session *Session)
{
  if (Session && Session->snmp_synch_state)
     free (Session->snmp_synch_state);
}

void snmp_synch_setup (struct snmp_session *Session)
{
  struct synch_state *rp = calloc (sizeof(*rp), 1);

  Session->snmp_synch_state = rp;
  Session->callback         = snmp_synch_input;
  Session->callback_magic   = (void*)rp;
}

