
/***********************************************************
        Copyright 1998 by Carnegie Mellon University

                      All Rights Reserved

Permission to use, copy, modify, and distribute this software and its 
documentation for any purpose and without fee is hereby granted, 
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in 
supporting documentation, and that the name of CMU not be
used in advertising or publicity pertaining to distribution of the
software without specific, written prior permission.  

CMU DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING
ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL
CMU BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR
ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
SOFTWARE.

Changes by G. Vanem 1998
Adapted to Waterloo tcp/ip

******************************************************************/

#include "snmp_lib.h"

#undef STATUS      /* in <arpa/nameser.h> */

/*
 * This is one element of an object identifier with either an integer
 * subidentifier, or a textual string label, or both.  The subid is
 * -1 if not present, and label is NULL if not present.
 */
struct subid {
       int   subid;
       char *label;
     };

/*
 * A linked list of nodes.
 */
struct node {
       struct node *next;
       char   label [TOKENSIZE];  /* This node's (unique) textual name */
       DWORD  subid;              /* This node's integer subidentifier */
       char   parent[TOKENSIZE];  /* The parent's textual name */
       int    type;               /* The type of object this represents */
       struct enum_list *enums;   /* (optional) list of enumerated integers (otherwise NULL) */
     };

static int Line = 1;

/*
 * types of tokens
 */
#undef OBJID
#undef INTEGER
#define CONTINUE     -1
#define ENDOFFILE     0
#define LABEL         1
#define SUBTREE       2
#define SYNTAX        3
#define OBJID         4
#define OCTETSTR      5
#define INTEGER       6
#define NETADDR       7
#define IPADDR        8
#define COUNTER       9
#define GAUGE         10
#define TIMETICKS     11
#define OPAQUE        12
#define NUL           13
#define SEQUENCE      14
#define OF            15        /* SEQUENCE OF */
#define OBJTYPE       16
#define ACCESS        17
#define READONLY      18
#define READWRITE     19
#define WRITEONLY     20
#define NOACCESS      21
#define STATUS        22
#define MANDATORY     23
#define OPTIONAL      24
#define OBSOLETE      25
#define RECOMMENDED   26
#define PUNCT         27
#define EQUALS        28
#define NUMBER        29
#define LEFTBRACKET   30
#define RIGHTBRACKET  31
#define LEFTPAREN     32
#define RIGHTPAREN    33
#define COMMA         34

/* For SNMPv2 SMI pseudo-compliance
 */
#define DESCRIPTION   35
#define INDEX         36
#define QUOTE         37

struct tok {
       char  *name;        /* token name */
       int    len;         /* length not counting nul */
       int    token;       /* value */
       int    hash;        /* hash of name */
       struct tok *next;   /* pointer to next in hash table */
     };


static struct tok tokens[] = {
  { "obsolete",        sizeof("obsolete")-1,        OBSOLETE     },
  { "Opaque",          sizeof("Opaque")-1,          OPAQUE       },
  { "recommended",     sizeof("recommended")-1,     RECOMMENDED  },
  { "optional",        sizeof("optional")-1,        OPTIONAL     },
  { "mandatory",       sizeof("mandatory")-1,       MANDATORY    },
  { "current",         sizeof("current")-1,         MANDATORY    },
  { "not-accessible",  sizeof("not-accessible")-1,  NOACCESS     },
  { "write-only",      sizeof("write-only")-1,      WRITEONLY    },
  { "read-write",      sizeof("read-write")-1,      READWRITE    },
  { "TimeTicks",       sizeof("TimeTicks")-1,       TIMETICKS    },
  { "OBJECTIDENTIFIER",sizeof("OBJECTIDENTIFIER")-1,OBJID        },
  /*
   * This CONTINUE appends the next word onto OBJECT,
   * hopefully matching OBJID above.
   */
  { "OBJECT",          sizeof("OBJECT")-1,          CONTINUE     },
  { "NetworkAddress",  sizeof("NetworkAddress")-1,  NETADDR      },
  { "Gauge",           sizeof("Gauge")-1,           GAUGE        },
  { "OCTETSTRING",     sizeof("OCTETSTRING")-1,     OCTETSTR     },
  { "OCTET",           sizeof("OCTET")-1,           -1           },
  { "OF",              sizeof("OF")-1,              OF           },
  { "SEQUENCE",        sizeof("SEQUENCE")-1,        SEQUENCE     },
  { "NULL",            sizeof("NULL")-1,            NUL          },
  { "IpAddress",       sizeof("IpAddress")-1,       IPADDR       },
  { "INTEGER",         sizeof("INTEGER")-1,         INTEGER      },
  { "Counter",         sizeof("Counter")-1,         COUNTER      },
  { "read-only",       sizeof("read-only")-1,       READONLY     },
  { "ACCESS",          sizeof("ACCESS")-1,          ACCESS       },
  { "MAX-ACCESS",      sizeof("MAX-ACCESS")-1,      ACCESS       },
  { "STATUS",          sizeof("STATUS")-1,          STATUS       },
  { "SYNTAX",          sizeof("SYNTAX")-1,          SYNTAX       },
  { "OBJECT-TYPE",     sizeof("OBJECT-TYPE")-1,     OBJTYPE      },
  { "{",               sizeof("{")-1,               LEFTBRACKET  },
  { "}",               sizeof("}")-1,               RIGHTBRACKET },
  { "::=",             sizeof("::=")-1,             EQUALS       },
  { "(",               sizeof("(")-1,               LEFTPAREN    },
  { ")",               sizeof(")")-1,               RIGHTPAREN   },
  { ",",               sizeof(",")-1,               COMMA        },
  { "DESCRIPTION",     sizeof("DESCRIPTION")-1,     DESCRIPTION  },
  { "INDEX",           sizeof("INDEX")-1,           INDEX        },
  { "\"",              sizeof("\"")-1,              QUOTE        },

  /* Hacks for easier MIBFILE coercing
   */
  { "read-create",    sizeof("read-create")-1,      READWRITE    },
  { NULL }
};


#define HASHSIZE    32
#define BUCKET(x)   (x & (HASHSIZE-1))
#define NHASHSIZE   128
#define NBUCKET(x)  (x & (NHASHSIZE-1))

static struct tok  *buckets [HASHSIZE];
static struct node *nbuckets[NHASHSIZE];

static void hash_init (void)
{
  struct tok *tp;
  char  *cp;
  int    h, b;

  memset (buckets,0,sizeof(buckets));
  for (tp = tokens; tp->name; tp++)
  {
    for (h = 0, cp = tp->name; *cp; cp++)
      h += *cp;
    tp->hash = h;
    b = BUCKET (h);
    if (buckets[b])
      tp->next = buckets[b];        /* BUG ??? */
    buckets[b] = tp;
  }
}

static void init_node_hash (struct node *nodes)
{
  struct node *np, *nextp;
  char  *cp;
  int    h, b;

  memset (nbuckets,0,sizeof(nbuckets));
  for (np = nodes; np;)
  {
    nextp = np->next;
    h = 0;
    for (cp = np->parent; *cp; cp++)
      h += *cp;
    b = NBUCKET (h);
    np->next = nbuckets [b];
    nbuckets [b] = np;
    np = nextp;
  }
}


static void print_error (char *string, char *token, int type)
{
  if (type == ENDOFFILE)
       fprintf (stderr, "%s(EOF): On or around line %d\n", string, Line);
  else if (token)
       fprintf (stderr, "%s(%s): On or around line %d. (Type: %d)\n",
                string, token, Line, type);
  else fprintf (stderr, "%s: On or around line %d.  Type: %d\n",
                string, Line, type);
}


static int translation_table[40];

static void build_translation_table (void)
{
  int count;

  for (count = 0; count < DIM(translation_table); count++)
  {
    switch (count)
    {
      case OBJID:
           translation_table[count] = TYPE_OBJID;
           break;
      case OCTETSTR:
           translation_table[count] = TYPE_OCTETSTR;
           break;
      case INTEGER:
           translation_table[count] = TYPE_INTEGER;
           break;
      case NETADDR:
           translation_table[count] = TYPE_IPADDR;
           break;
      case IPADDR:
           translation_table[count] = TYPE_IPADDR;
           break;
      case COUNTER:
           translation_table[count] = TYPE_COUNTER;
           break;
      case GAUGE:
           translation_table[count] = TYPE_GAUGE;
           break;
      case TIMETICKS:
           translation_table[count] = TYPE_TIMETICKS;
           break;
      case OPAQUE:
           translation_table[count] = TYPE_OPAQUE;
           break;
      case NUL:
           translation_table[count] = TYPE_NULL;
           break;
      default:
           translation_table[count] = TYPE_OTHER;
           break;
    }
  }
}

/* 
 * Find all the children of root in the list of nodes.  Link them into the
 * tree and out of the nodes list.
 */
static void do_subtree (struct snmp_mib_tree *root, struct node **nodes)
{
  struct snmp_mib_tree *tp   = root;
  struct snmp_mib_tree *peer = NULL;
  struct node          *np, **headp;
  struct node          *oldnp      = NULL;
  struct node          *child_list = NULL;
  struct node          *childp     = NULL;
  char  *cp;
  int    hash = 0;

  for (cp = tp->label; *cp; cp++)
     hash += *cp;
  headp = &nbuckets [NBUCKET(hash)];

  /*
   * Search each of the nodes for one whose parent is root, and
   * move each into a separate list.
   */
  for (np = *headp; np; np = np->next)
  {
    if ((*tp->label != *np->parent) || strcmp (tp->label, np->parent))
    {
      if ((*tp->label == *np->label) && !strcmp (tp->label, np->label))
      {
        /* if there is another node with the same label, assume that
         * any children after this point in the list belong to the other node.
         * This adds some scoping to the table and allows vendors to
         * reuse names such as "ip".
         */
        break;
      }
      oldnp = np;
    }
    else
    {
      if (!child_list)
      {
        child_list = childp = np;        /* first entry in child list */
      }
      else
      {
        childp->next = np;
        childp = np;
      }
      /* take this node out of the node list
       */
      if (oldnp)
           oldnp->next = np->next;  /* link around this node */
      else *headp      = np->next;  /* fix root of node list */
    }
  }
  if (childp)
     childp->next = NULL;           /* re-terminate list */

  /*
   * Take each element in the child list and place it into the tree.
   */
  for (np = child_list; np; np = np->next)
  {
    tp = calloc (sizeof(*tp),1);

    tp->parent = root;
    strcpy (tp->label, np->label);
    tp->subid = np->subid;
    tp->type  = translation_table[np->type];
    tp->enums = np->enums;
    np->enums = NULL;                /* so we don't free them later */
    if (root->child_list)
         peer->next_peer  = tp;
    else root->child_list = tp;

    peer = tp;
    do_subtree (tp, nodes); /* recurse on this child if it isn't an end node */
  }

  /* free all nodes that were copied into tree
   */
  oldnp = NULL;
  for (np = child_list; np; np = np->next)
  {
    if (oldnp)
      free (oldnp);
    oldnp = np;
  }
  if (oldnp)
    free (oldnp);
}

static struct snmp_mib_tree *build_tree (struct node *nodes)
{
  struct node          *np;
  struct snmp_mib_tree *tp;
  int    bucket, nodes_left = 0;

  /* build root node
   */
  tp = calloc (sizeof(*tp),1);
  if (!tp)
     return (NULL);

  strcpy (tp->label, "iso");
  tp->subid = 1;
  build_translation_table();

  /* grow tree from this root node
   */
  init_node_hash (nodes);

  /* XXX nodes isn't needed in do_subtree() ???
   */
  do_subtree (tp, &nodes);

#ifdef TEST
  snmp_print_subtree (tp, 0);
#endif

  /* If any nodes are left, the tree is probably inconsistent
   */
  for (bucket = 0; bucket < NHASHSIZE; bucket++)
  {
    if (nbuckets[bucket])
    {
      nodes_left = 1;
      break;
    }
  }
  if (nodes_left)
  {
    fprintf (stderr, "The mib description doesn't seem to be consistent.\n");
    fprintf (stderr, "Some nodes couldn't be linked under the \"iso\" tree.\n");
    fprintf (stderr, "these nodes are left:\n");
    for (bucket = 0; bucket < NHASHSIZE; bucket++)
    {
      for (np = nbuckets[bucket]; np; np = np->next)
        fprintf (stderr, "%s ::= { %s %ld } (%d)\n",
                 np->label, np->parent, np->subid, np->type);
    }
  }
  return (tp);
}


/* 
 * Parses a token from the file.  The type of the token parsed is returned,
 * and the text is placed in the string pointed to by token.
 */
static char last = ' ';

static int get_token (FILE *fp, char *token)
{
  struct tok *tp;
  char  *cp = token;
  int    ch, hash = 0;

  *cp = 0;
  ch  = last;

  /* skip all white space
   */
  while (isspace(ch) && ch != EOF)
  {
    ch = getc (fp);
    if (ch == '\n')
      Line++;
  }
  if (ch == EOF)
     return (ENDOFFILE);

  /*
   * Accumulate characters until end of token is found.  Then attempt to
   * match this token as a reserved word.  If a match is found, return
   * the type.  Else it is a label.
   */
  do
  {
    if (ch == '\n')
       Line++;
    if (isspace(ch) || ch == '(' || ch == ')' ||
        ch == '{' || ch == '}' || ch == ',' ||
        ch == '"')
    {
      if (!isspace(ch) && *token == 0)
      {
        hash += ch;
        *cp++ = ch;
        last  = ' ';
      }
      else
        last = ch;

      *cp = '\0';

      for (tp = buckets[BUCKET(hash)]; tp; tp = tp->next)
         if (tp->hash == hash && !strcmp(tp->name,token))
            break;

      if (tp)
      {
        if (tp->token == CONTINUE)
           continue;

        RYAN_DEBUG1 ("`%s'\n", token);
        return (tp->token);
      }

      if (token[0] == '-' && token[1] == '-')
      {
        /* strip comment */
        while ((ch = getc (fp)) != EOF)
          if (ch == '\n')
          {
            Line++;
            break;
          }
        if (ch == EOF)
           return (ENDOFFILE);

        last = ch;
        return get_token (fp, token);
      }
      for (cp = token; *cp; cp++)
        if (!isdigit (*cp))
        {
          RYAN_DEBUG1 ("`%s' (LABEL)\n", token);
          return (LABEL);
        }
      RYAN_DEBUG1 ("`%s' (NUMBER)\n", token);
      return (NUMBER);
    }
    hash += ch;
    *cp++ = ch;
    if (ch == '\n')
       Line++;
  }
  while ((ch = getc(fp)) != EOF);
  return (ENDOFFILE);
}

/* 
 * Takes a list of the form:
 * { iso org(3) dod(6) 1 }
 * and creates several nodes, one for each parent-child pair.
 * Returns NULL on error.
 */
static int getoid (FILE *fp,
                   struct subid *SubOid, /* an array of subids */
                   int length)           /* the length of the array */
{
  char token[TOKENSIZE];
  int  count;
  int  type = get_token (fp, token);

  if (type != LEFTBRACKET)
  {
    print_error ("Expected \"{\"", token, type);
    return (0);
  }
  type = get_token (fp, token);

  for (count = 0; count < length; count++, SubOid++)
  {
    SubOid->label = 0;
    SubOid->subid = -1;
    if (type == RIGHTBRACKET)
       return (count);

    if (type != LABEL && type != NUMBER)
    {
      print_error ("Not valid for object identifier", token, type);
      return (0);
    }
    if (type == LABEL)
    {
      /* this entry has a label */
      SubOid->label = strdup (token);
      type = get_token (fp, token);
      if (type == LEFTPAREN)
      {
        type = get_token (fp, token);
        if (type == NUMBER)
        {
          SubOid->subid = atoi (token);
          type = get_token (fp, token);
          if (type != RIGHTPAREN)
          {
            print_error ("Unexpected a closing parenthesis", token, type);
            return (0);
          }
        }
        else
        {
          print_error ("Expected a number", token, type);
          return (0);
        }
      }
      else
        continue;
    }
    else
    {
      /* this entry  has just an integer sub-identifier
       */
      SubOid->subid = atoi (token);
    }
    type = get_token (fp, token);
  }
  return (count);
}

static void free_node (struct node *np)
{
  struct enum_list *ep = np->enums;

  while (ep)
  {
    struct enum_list *tep = ep;
    ep = ep->next;
    free ((void*)tep);
  }
  free ((void*)np);
}

/* 
 * Parse an entry of the form:
 * label OBJECT IDENTIFIER ::= { parent 2 }
 * The "label OBJECT IDENTIFIER" portion has already been parsed.
 * Returns 0 on error.
 */
static struct node * parse_objectid (FILE *fp, char *name)
{
  struct subid *op, *nop;
  struct node  *np, *root;
  struct node  *oldnp = NULL;
  struct subid  SubOid[32];
  char   token [TOKENSIZE];
  int    type, count, length;

  type = get_token (fp, token);
  if (type != EQUALS)
  {
    print_error ("Bad format", token, type);
    return (NULL);
  }
  length = getoid (fp, SubOid, DIM(SubOid));
  if (length)
  {
    np = root = calloc (sizeof(*np),1);

    /*
     * For each parent-child subid pair in the subid array,
     * create a node and link it into the node list.
     */
    for (count = 0, op = SubOid, nop = SubOid + 1;
         count < (length - 2);
         count++, op++, nop++)
    {
      /* every node must have parent's name and child's name or number
       */
      if (op->label && (nop->label || nop->subid != -1))
      {
        strcpy (np->parent, op->label);
        if (nop->label)
           strcpy (np->label, nop->label);
        if (nop->subid != -1)
           np->subid = nop->subid;
        np->type  = 0;
        np->enums = 0;
        /* set up next entry
         */
        np->next = calloc (sizeof(struct node),1);
        oldnp = np;
        np    = np->next;
      }
    }
    np->next = NULL;
    /*
     * The above loop took care of all but the last pair.  This pair is taken
     * care of here.  The name for this node is taken from the label for this
     * entry.
     * np still points to an unused entry.
     */
    if (count == (length - 2))
    {
      if (op->label)
      {
        strcpy (np->parent, op->label);
        strcpy (np->label, name);
        if (nop->subid != -1)
             np->subid = nop->subid;
        else print_error ("Warning: This entry is pretty silly",
                          np->label, type);
      }
      else
      {
        free_node (np);
        if (oldnp)
             oldnp->next = NULL;
        else return (NULL);
      }
    }
    else
    {
      print_error ("Missing end of oid", NULL, type);
      free_node (np);                /* the last node allocated wasn't used */
      if (oldnp)
        oldnp->next = NULL;
      return (NULL);
    }

    /* free the oid array
     */
    for (count = 0, op = SubOid; count < length; count++, op++)
    {
      if (op->label)
        free (op->label);
      op->label = 0;
    }
    return (root);
  }

  print_error ("Bad object identifier", NULL, type);
  return (NULL);
}

/* 
 * Parses an asn type.  This structure is ignored by this parser.
 * Returns NULL on error.
 */
static int parse_asntype (FILE *fp)
{
  int  type;
  char token[TOKENSIZE];

  type = get_token (fp, token);
  if (type != SEQUENCE)
  {
    print_error ("Not a sequence", token, type);  /* should we handle this */
    return (ENDOFFILE);
  }
  while ((type = get_token (fp, token)) != ENDOFFILE)
  {
    if (type == RIGHTBRACKET)
       return (type);
  }
  print_error ("Expected \"}\"", token, type);
  return (ENDOFFILE);
}


/*
 * Parses an OBJECT TYPE macro.
 * Returns 0 on error.
 */
static struct node *parse_objecttype (FILE *fp, char *name)
{
  char   token    [TOKENSIZE];
  char   syntax   [TOKENSIZE];
  char   nexttoken[TOKENSIZE];
  int    type, nexttype, count, length;

  struct subid      SubOid[32];
  struct node      *np;
  struct enum_list *ep = NULL;

  type = get_token (fp, token);
  if (type != SYNTAX)
  {
    print_error ("Bad format for OBJECT TYPE", token, type);
    return (NULL);
  }
  np   = calloc (sizeof(*np),1);
  type = get_token (fp, token);
  nexttype = get_token (fp, nexttoken);
  np->type = type;

  switch (type)
  {
    case SEQUENCE:
         strcpy (syntax, token);
         if (nexttype == OF)
         {
           strcat (syntax, " ");
           strcat (syntax, nexttoken);
           nexttype = get_token (fp, nexttoken);
           strcat (syntax, " ");
           strcat (syntax, nexttoken);
           nexttype = get_token (fp, nexttoken);
         }
         break;

    case INTEGER:
         strcpy (syntax, token);
         if (nexttype == LEFTBRACKET)
         {
           /* if there is an enumeration list, parse it
            */
           while ((type = get_token (fp, token)) != ENDOFFILE)
           {
             if (type == RIGHTBRACKET)
                break;

             if (type == LABEL)
             {
               /* this is an enumerated label
                */
               if (!np->enums)
                 ep = np->enums = calloc (sizeof(*ep),1);
               else
               {
                 ep->next = calloc (sizeof(*ep),1);
                 ep = ep->next;
               }
               ep->next = NULL;

               /* a reasonable approximation for the length
                */
               ep->label = strdup (token);
               type = get_token (fp, token);
               if (type != LEFTPAREN)
               {
                 print_error ("Expected \"(\"", token, type);
                 free_node (np);
                 return (NULL);
               }
               type = get_token (fp, token);
               if (type != NUMBER)
               {
                 print_error ("Expected integer", token, type);
                 free_node (np);
                 return (NULL);
               }
               ep->value = atoi (token);
               type = get_token (fp, token);
               if (type != RIGHTPAREN)
               {
                 print_error ("Expected \")\"", token, type);
                 free_node (np);
                 return (NULL);
               }
             }
           }
           if (type == ENDOFFILE)
           {
             print_error ("Expected \"}\"", token, type);
             free_node (np);
             return (NULL);
           }
           nexttype = get_token (fp, nexttoken);
         }
         else if (nexttype == LEFTPAREN)
         {
           /* ignore the "constrained integer" for now
            */
           nexttype = get_token (fp, nexttoken);
           nexttype = get_token (fp, nexttoken);
           nexttype = get_token (fp, nexttoken);
         }
         break;

    case OBJID:
    case OCTETSTR:
    case NETADDR:
    case IPADDR:
    case COUNTER:
    case GAUGE:
    case TIMETICKS:
    case OPAQUE:
    case NUL:
    case LABEL:
         strcpy (syntax, token);
         break;

    default:
         print_error ("Bad syntax", token, type);
         free_node (np);
         return (NULL);
  }

  if (nexttype != ACCESS)
  {
    print_error ("Should be ACCESS", nexttoken, nexttype);
    free_node (np);
    return (NULL);
  }
  type = get_token (fp, token);
  if (type != READONLY  && type != READWRITE &&
      type != WRITEONLY && type != NOACCESS)
  {
    print_error ("Bad access type", nexttoken, nexttype);
    free_node (np);
    return (NULL);
  }
  type = get_token (fp, token);
  if (type != STATUS)
  {
    print_error ("Should be STATUS", token, nexttype);
    free_node (np);
    return (NULL);
  }
  type = get_token (fp, token);
  if (type != MANDATORY && type != OPTIONAL && type != OBSOLETE && type != RECOMMENDED)
  {
    print_error ("Bad status", token, type);
    free_node (np);
    return (NULL);
  }

  /* Fetch next token.  Either:
   *
   * -> EQUALS (Old MIB format)
   * -> DESCRIPTION, INDEX (New MIB format)
   */
  type = get_token (fp, token);
  if (type != DESCRIPTION && type != INDEX && type != EQUALS)
  {
    print_error ("Should be DESCRIPTION, INDEX, or EQUALS", token, nexttype);
    free_node (np);
    return (NULL);
  }

  if (type == DESCRIPTION)
  {
    type = get_token (fp, token);
    if (type != QUOTE)
    {
      print_error ("Should be Description open quote", token, nexttype);
      free_node (np);
      return (NULL);
    }

    /* Fetch description string
     */
    {
      int readChar = last;

      while (readChar != '"' && readChar != EOF)
      {
        readChar = getc (fp);
        if (readChar == '\n')
           Line++;
      }
      last = ' ';
    }
    /* ASSERT:  Done with description
     */
    type = get_token (fp, token);
  }

  if (type != INDEX && type != EQUALS)
  {
    print_error ("Should be INDEX, or EQUALS", token, nexttype);
    free_node (np);
    return (NULL);
  }

  if (type == INDEX)
  {
    /* Scarf INDEX */

    type = get_token (fp, token);
    if (type != LEFTBRACKET)
    {
      print_error ("Should be INDEX left brace", token, type);
      free_node (np);
      return (NULL);
    }

    /* Fetch description string */
    {
      int readChar = last;
      while (readChar != '}' && readChar != EOF)
      {
        readChar = getc (fp);
        if (readChar == '\n')
           Line++;
      }
      last = ' ';
    }
    /* ASSERT:  Done with INDEX.
     */
    type = get_token (fp, token);
  }

  if (type != EQUALS)
  {
    print_error ("Bad format", token, type);
    free_node (np);
    return (NULL);
  }

  length = getoid (fp, SubOid, DIM(SubOid));
  if (length > 1 && length <= DIM(SubOid))
  {
    /* just take the last pair in the oid list
     */
    if (SubOid[length-2].label)
       strncpy (np->parent, SubOid[length-2].label, TOKENSIZE);
    strcpy (np->label, name);
    if (SubOid[length-1].subid != -1)
         np->subid = SubOid[length-1].subid;
    else print_error ("Warning: This entry is pretty silly", np->label, type);
  }
  else
  {
    print_error ("No end to oid", NULL, type);
    free_node (np);
    np = NULL;
  }

  /* free oid array
   */
  for (count = 0; count < length; count++)
  {
    if (SubOid[count].label)
       free (SubOid[count].label);
    SubOid[count].label = NULL;
  }
  return (np);
}


#if 0

#define EAT_TOKEN(x, where) {          \
          type = get_token(fp, token); \
          if (type != x) {             \
             fprintf(stderr, "Error eating token %d, read %d (%s) - %s\n", \
                     x, type, token, where); exit(0);  \
          } \
        }

/* 
 * Load an OBJECT-IDENTITY entry. ** SMIv2 **
 *
 * Create the node, and return the next token.  Much like
 * object identifier, but with more data.
 */
static struct node *parse_object_identity (FILE * fp, char *name)
{
  int    type;                    /* For EAT_TOKEN macro */
  char   token[TOKENSIZE];        /* For EAT_TOKEN macro */
  int    count, length;

  struct subid *op, *nop;
  struct node  *np, *root, *oldnp = NULL;
  struct subid  SubOid[32];

  EAT_TOKEN (STATUS,      "Status in parse_object_identity");
  EAT_TOKEN (MANDATORY,   "Status level");
  EAT_TOKEN (DESCRIPTION, "Object Identity Description");
  EAT_TOKEN (QUOTE,       "Object Identity Description Open Quote");

  /* Fetch description string */
  {
    int ReadChar = last;
    /* skip everything until closing quote */
    while (ReadChar != '"' && ReadChar != EOF)
    {
      ReadChar = getc (fp);
      if (ReadChar == '\n')
         Line++;
    }
    last = ' ';
  }
  EAT_TOKEN (EQUALS, "Object identity equals");

  /* Now read the oid, and do the oid thing */

  if ((length = getoid (fp, SubOid,DIM(SubOid))) != 0)
  {
    np = root = calloc (sizeof(*np),1);

    /*
     * For each parent-child subid pair in the subid array,
     * create a node and link it into the node list.
     */
    for (count = 0, op = SubOid, nop = SubOid + 1;
         count < (length - 2);
         count++, op++, nop++)
    {
      /* every node must have parent's name and child's name or number
       */
      if (op->label && (nop->label || (nop->subid != -1)))
      {
        strcpy (np->parent, op->label);
        if (nop->label)
          strcpy (np->label, nop->label);
        if (nop->subid != -1)
          np->subid = nop->subid;
        np->type  = 0;
        np->enums = 0;
        /* set up next entry */
        np->next = calloc (sizeof(*np->next),1);
        oldnp = np;
        np    = np->next;
      }
    }
    np->next = NULL;

    /*
     * The above loop took care of all but the last pair.  This pair is taken
     * care of here.  The name for this node is taken from the label for this
     * entry.
     * np still points to an unused entry.
     */
    if (count == (length - 2))
    {
      if (op->label)
      {
        strcpy (np->parent, op->label);
        strcpy (np->label, name);
        if (nop->subid != -1)
             np->subid = nop->subid;
        else print_error ("Warning: This entry is pretty silly",
                          np->label, type);
      }
      else
      {
        free_node (np);
        if (oldnp)
             oldnp->next = NULL;
        else return (NULL);
      }
    }
    else
    {
      print_error ("Missing end of oid", NULL, type);
      free_node (np);                /* the last node allocated wasn't used */
      if (oldnp)
        oldnp->next = NULL;
      return (NULL);
    }
    /* free the oid array
     */
    for (count = 0, op = SubOid; count < length; count++, op++)
    {
      if (op->label)
        free (op->label);
      op->label = 0;
    }
    return (root);
  }
  else
  {
    print_error ("Bad object identifier", NULL, type);
    return (NULL);
  }
}
#endif


/* 
 * Parses a mib file and returns a linked list of nodes found in the file.
 * Returns NULL on error.
 */
static struct node *parse (FILE *fp)
{
  struct node *np   = NULL;
  struct node *root = NULL;
  int    type       = ENDOFFILE-1;
  char   token[TOKENSIZE];
  char   name [TOKENSIZE];

  hash_init();

  while (type != ENDOFFILE)
  {
    PARSER_DEBUG0 ("Top of the loop\n");

    type = get_token (fp, token);
    if (type != LABEL)
    {
      if (type == ENDOFFILE)
         return (root);

      print_error (token, "is a reserved word", type);
      return (NULL);
    }
    strncpy (name, token, TOKENSIZE-1);
    type = get_token (fp, token);
    if (type == OBJTYPE)
    {
      if (!root)
      {
        /* first link in chain
         */
        np = root = parse_objecttype (fp, name);
        if (!np)
        {
          print_error ("Bad parse of object type", NULL, type);
          return (NULL);
        }
      }
      else
      {
        np->next = parse_objecttype (fp, name);
        if (!np->next)
        {
          print_error ("Bad parse of objecttype", NULL, type);
          return (NULL);
        }
      }
      /* now find end of chain
       */
      while (np->next)
        np = np->next;

    }
    else if (type == OBJID)
    {
      if (!root)
      {
        /* first link in chain
         */
        np = root = parse_objectid (fp, name);
        if (!np)
        {
          print_error ("Bad parse of object id", NULL, type);
          return (NULL);
        }
      }
      else
      {
        np->next = parse_objectid (fp, name);
        if (!np->next)
        {
          print_error ("Bad parse of object type", NULL, type);
          return (NULL);
        }
      }
      while (np->next)
        np = np->next;

    }
    else if (type == EQUALS)
    {
      type = parse_asntype (fp);
    }
    else if (type == ENDOFFILE)
    {
      break;
    }
    else
    {
      print_error ("Bad operator", NULL, type);
      return (NULL);
    }
  }
#ifdef TEST
  {
    struct enum_list *ep;

    for (np = root; np; np = np->next)
    {
      printf ("%s ::= { %s %ld } (%d)\n",
              np->label, np->parent, np->subid, np->type);
      if (np->enums)
      {
	printf ("Enums: \n");
	for (ep = np->enums; ep; ep = ep->next)
           printf ("%s(%d)\n", ep->label, ep->value);
      }
    }
  }
#endif
  return (root);
}

struct snmp_mib_tree *read_mib (const char *filename)
{
  struct node          *nodes;
  struct snmp_mib_tree *tree;
  FILE  *fp;

  if (!filename)
     return (NULL);

  fp = fopen (filename, "rt");
  if (!fp)
     return (NULL);

  nodes = parse (fp);
  if (!nodes)
  {
    fprintf (stderr, "Mib table is bad.  Exiting\n");
    return (NULL);
  }
  tree = build_tree (nodes);
  fclose (fp);
  return (tree);
}
