
/* 
 * SNMP Variable Binding.  Complies with:
 *
 * RFC 1905: Protocol Operations for SNMPv2
 *
 */

/**********************************************************************
 *
 *           Copyright 1998 by Carnegie Mellon University
 * 
 *                       All Rights Reserved
 * 
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose and without fee is hereby granted,
 * provided that the above copyright notice appear in all copies and that
 * both that copyright notice and this permission notice appear in
 * supporting documentation, and that the name of CMU not be
 * used in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.
 * 
 * CMU DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING
 * ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL
 * CMU BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR
 * ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
 * ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
 * SOFTWARE.
 * 
 * Author: Ryan Troll <ryan+@andrew.cmu.edu>
 *
 * Changes by G. Vanem 1998
 * Adapted to Waterloo tcp/ip
 **********************************************************************/

#include "snmp_lib.h"

#define ASN_PARSE_ERROR() do {                          \
                            snmpInASNParseErrs_Add (1); \
                            return (NULL);              \
                          } while (0)

/* Create a new variable_list structure representing oid Name of length Len.
 *
 * Returns NULL upon error.
 */
struct variable_list *snmp_var_new (oid * Name, int Len)
{
  struct variable_list *New;

  VARS_DEBUG0 ("VARS: Creating.\n");

  New = calloc (sizeof(*New),1);
  if (!New)
  {
    snmp_set_api_error (SNMPERR_OS_ERR);
    return (NULL);
  }
  New->type        = ASN_NULL;
  New->name_length = Len;
  if (New->name_length == 0)
     return (New);

  New->name = malloc (Len*sizeof(oid));
  if (!New->name)
  {
    free (New);
    snmp_set_api_error (SNMPERR_OS_ERR);
    return (NULL);
  }

  VARS_DEBUG1 ("Copying name, size (%d)\n", Len);

  /* Only copy a name if it was specified.
   */
  if (Name)
    memcpy (New->name,Name, Len*sizeof(oid));

  VARS_DEBUG1 ("Created %x.\n",  (unsigned int)New);
  VARS_DEBUG1 ("Created (%x)\n", (unsigned int)New);
  VARS_DEBUG1 ("Name is (%x)\n", (unsigned int)New->name);
  return (New);
}

/*
 * Clone a variable list.
 *
 * Returns NULL upon error.
 */

struct variable_list *snmp_var_clone (struct variable_list *src)
{
  struct variable_list *dest;

  VARS_DEBUG0 ("Cloning.\n");

  dest = malloc (sizeof(*dest));
  if (!dest)
  {
    snmp_set_api_error (SNMPERR_OS_ERR);
    return (NULL);
  }

  VARS_DEBUG1 ("Copying entire variable list.  (Size %d)\n",
               (int)sizeof(*dest));

  memcpy (dest,src,sizeof(*dest));

  if (src->name)
  {
    dest->name = malloc (src->name_length * sizeof(oid));
    if (!dest->name)
    {
      snmp_set_api_error (SNMPERR_OS_ERR);
      free (dest);
      return (NULL);
    }
    VARS_DEBUG1 ("Copying name OID. (Size %d)\n", src->name_length);

    memcpy (dest->name, src->name, src->name_length * sizeof(oid));
  }

  /* CISCO Catalyst 2900 returns NULL strings as data of length 0.
   */
  if (src->val.string && src->val_len)
  {
    dest->val.string = malloc (src->val_len);
    if (!dest->val.string)
    {
      snmp_set_api_error (SNMPERR_OS_ERR);
      free (dest->name);
      free (dest);
      return (NULL);
    }

    VARS_DEBUG1 ("Copying value (Size %d)\n", src->val_len);
    memcpy (dest->val.string, src->val.string, src->val_len);
  }

  VARS_DEBUG1 ("Cloned %x.\n",   (unsigned int)dest);
  VARS_DEBUG1 ("Cloned  (%x)\n", (unsigned int)dest);
  VARS_DEBUG1 ("Name is (%x)\n", (unsigned int)dest->name);

  return (dest);
}

/*
 * Free a variable_list.
 */
void snmp_var_free (struct variable_list *Ptr)
{
  VARS_DEBUG1 ("Free'd  (%x)\n", (unsigned int)Ptr);
  VARS_DEBUG1 ("Name was(%x)\n", (unsigned int)Ptr->name);

  if (Ptr->name && Ptr->name_length > 0)
     free (Ptr->name);

  if (Ptr->val.string && Ptr->val_len > 0)
     free (Ptr->val.string);

  free (Ptr);
}

/**********************************************************************/

/* Build a variable binding.
 *
 * RFC 1905: Protocol Operations for SNMPv2
 *
 * VarBind ::=
 *   SEQUENCE {
 *     name
 * ObjectName
 *     CHOICE {
 *       value ObjectSyntax
 *       unSpecified NULL
 *       noSuchObject[0] NULL
 *
 * noSuchInstance[1] NULL
 *       endOfMibView[2] NULL
 *     }
 *   }
 */
BYTE *snmp_var_EncodeVarBind (BYTE *Buffer, int *BufLenP,
                              struct variable_list *VarList,
                              int Version)
{
  struct variable_list *Vars;
  BYTE  *bufp;
  BYTE  *HeaderStart;
  BYTE  *HeaderEnd;
  int    FakeArg  = *BufLenP;
  int    Counter  = 1;

#if DEBUG_VARS
  int  StartLen = *BufLenP;
#endif

  bufp = Buffer;

  VARS_DEBUG1 ("Encoding Variable list into buffer at 0x%x.\n",
               (unsigned)Buffer);

  for (Vars = VarList; Vars; Vars = Vars->next_variable)
  {  
    VARS_DEBUG2 ("%d: Encoding Variable 0x%x.\n", Counter, (unsigned)Vars);
    VARS_DEBUG3 ("%d: Starting at 0x%x (%d bytes left)\n",
                 Counter, (unsigned)bufp, *BufLenP);

    /* Build the header for this variable
     *
     * Use Maximum size.
     */
    HeaderStart = bufp;
    HeaderEnd = asn_build_header (HeaderStart, BufLenP,
                                  (BYTE) (ASN_SEQUENCE | ASN_CONSTRUCTOR),
                                  FakeArg);
    if (!HeaderEnd)
       return (NULL);

#ifdef DEBUG_VARS
    VARS_DEBUG5 ("%d: Encoding Object Identifier 0x%x (%d bytes) at "
                 "0x%x (%d bytes left)\n",
                 Counter, (unsigned)Vars, Vars->name_length,
                 (unsigned)HeaderEnd, *BufLenP);

    print_oid_nums (Vars->name, Vars->name_length),
#endif

    /* Now, let's put the Object Identifier into the buffer
     */
    bufp = asn_build_objid (HeaderEnd, BufLenP,
                            (BYTE) (ASN_UNIVERSAL |
                                    ASN_PRIMITIVE |
                                    ASN_OBJECT_ID),
                            Vars->name, Vars->name_length);
    if (!bufp)
       return (NULL);

    /* Now put the data in
     */
    switch (Vars->type)
    {
      case ASN_INTEGER:
           VARS_DEBUG3 ("%d: Encoding Integer %d at 0x%x\n",
                        Counter, *Vars->val.integer, (unsigned)bufp);
           bufp = asn_build_int (bufp, BufLenP, Vars->type,
                                 (int*) Vars->val.integer, Vars->val_len);
           break;

      case SMI_COUNTER32:
      case SMI_GAUGE32:
   /* case SMI_UNSIGNED32: */
      case SMI_TIMETICKS:
           VARS_DEBUG3 ("%d: Encoding Timeticks %d at 0x%x\n",
                        Counter, *(Vars->val.integer), (unsigned)bufp);
           bufp = asn_build_unsigned_int (bufp, BufLenP, Vars->type,
                                          (DWORD *) Vars->val.integer,
                                          Vars->val_len);
           break;

      case ASN_OCTET_STR:
      case SMI_IPADDRESS:
      case SMI_OPAQUE:
           VARS_DEBUG4 ("%d: Encoding String %s (%d bytes) at 0x%x\n",
                        Counter, Vars->val.string, Vars->val_len, (unsigned)bufp);
           bufp = asn_build_string (bufp, BufLenP, Vars->type,
                                    Vars->val.string, Vars->val_len);
           break;

      case ASN_OBJECT_ID:
           VARS_DEBUG3 ("%d: Encoding Object Identifier (%d bytes) at 0x%x\n",
                        Counter, Vars->val_len, (unsigned)bufp);
           bufp = asn_build_objid (bufp, BufLenP, Vars->type,
                                   (oid*) Vars->val.objid,
                                   Vars->val_len/sizeof(oid));
           break;

      case SMI_NOSUCHINSTANCE:
      case SMI_NOSUCHOBJECT:
      case SMI_ENDOFMIBVIEW:
           VARS_DEBUG2 ("%d: Encoding NULL at 0x%x\n",
                        Counter, (unsigned)bufp);

           if (Version == SNMP_VERSION_1)
           {
             /* SNMP Version 1 does not support these error codes. */
             bufp = asn_build_null (bufp, BufLenP, SMI_NOSUCHOBJECT);
           }
           else
           {
             bufp = asn_build_exception (bufp, BufLenP, Vars->type);
           }
           break;

      case ASN_NULL:
           VARS_DEBUG2 ("VARS %d: Encoding NULL at 0x%x\n",
                        Counter, (unsigned)bufp);
           bufp = asn_build_null (bufp, BufLenP, Vars->type);
           break;

      case SMI_COUNTER64:
           fprintf (stderr, "Unable to encode type SMI_COUNTER64!\n");
           /* Fall through */

      default:
           snmp_set_api_error (SNMPERR_UNSUPPORTED_TYPE);
           return (NULL);
    }

    /* ASSERT:  bufp should now point to the next valid byte
     */
    if (!bufp)
      return (NULL);

    /* Rebuild the header with the appropriate length
     */
    VARS_DEBUG4 ("%d: Resetting length to %ld at 0x%x (%d bytes left)\n",
                 Counter, (bufp - HeaderEnd), (unsigned)HeaderStart, *BufLenP);

    HeaderEnd = asn_build_header (HeaderStart, &FakeArg,
                                  (BYTE) (ASN_SEQUENCE | ASN_CONSTRUCTOR),
                                  (bufp - HeaderEnd));

    /* Returns NULL */
    if (!HeaderEnd)
      return (NULL);

    Counter++;
  }


  VARS_DEBUG2 ("Variable list of %d vars takes up %d bytes.\n",
               --Counter, StartLen - *BufLenP);

  /* or the end of the entire thing */
  return (bufp);
}


/*
 * Parse all Vars from the buffer
 */
BYTE *snmp_var_DecodeVarBind (BYTE *Buffer, int *BufLen,
                              struct variable_list **VarP,
                              int version)
{
  struct variable_list *Var, **VarLastP = VarP;
  BYTE  *bufp, *tmp;
  BYTE   VarBindType;
  BYTE  *DataPtr;
  int    DataLen;
  oid    TmpBuf[MAX_NAME_LEN];
  int    AllVarLen  = *BufLen;
  int    ThisVarLen = 0;

  VARS_DEBUG1 ("Decoding buffer of length %d\n", *BufLen);

  /* Now parse the variables */
  bufp = asn_parse_header (Buffer, &AllVarLen, &VarBindType);
  if (!bufp)
     ASN_PARSE_ERROR();

  if (VarBindType != (BYTE)(ASN_SEQUENCE | ASN_CONSTRUCTOR))
  {
    snmp_set_api_error (SNMPERR_PDU_PARSE);
    ASN_PARSE_ERROR();
  }

  VARS_DEBUG1 ("All Variable length %d\n", AllVarLen);

  /* We know how long the variable list is.  Parse it.
   */
  while ((int) AllVarLen > 0)
  {
    /* Create a new variable
     */
    Var = snmp_var_new (NULL, MAX_NAME_LEN);
    if (!Var)
       return (NULL);

    /* Parse the header to find out the length of this variable
     */
    ThisVarLen = AllVarLen;
    tmp = asn_parse_header (bufp, &ThisVarLen, &VarBindType);
    if (!tmp)
       ASN_PARSE_ERROR();

    /* Now that we know the length , figure out how it relates to 
     * the entire variable list
     */
    AllVarLen = AllVarLen - (ThisVarLen + (tmp - bufp));
    bufp = tmp;

    /* Is it valid?
     */
    if (VarBindType != (BYTE)(ASN_SEQUENCE | ASN_CONSTRUCTOR))
    {
      snmp_set_api_error (SNMPERR_PDU_PARSE);
      ASN_PARSE_ERROR();
    }

    VARS_DEBUG2 ("Header type 0x%x (%d bytes left)\n",
                 VarBindType, ThisVarLen);

    /* Parse the OBJID
     */
    bufp = asn_parse_objid (bufp, &ThisVarLen, &VarBindType,
                            Var->name, &(Var->name_length));
    if (!bufp)
       ASN_PARSE_ERROR();

    if (VarBindType != (BYTE) (ASN_UNIVERSAL |
                               ASN_PRIMITIVE |
                               ASN_OBJECT_ID))
    {
      snmp_set_api_error (SNMPERR_PDU_PARSE);
      ASN_PARSE_ERROR();
    }

    VARS_DEBUG2 ("Decoded OBJID (%d bytes). (%d bytes left)\n",
                 Var->name_length, ThisVarLen);

    /* Keep a pointer to this object */
    DataPtr = bufp;
    DataLen = ThisVarLen;

    /* find out type of object */
    bufp = asn_parse_header (bufp, &ThisVarLen, &(Var->type));
    if (!bufp)
      ASN_PARSE_ERROR();
    ThisVarLen = DataLen;

    VARS_DEBUG1 ("Data type %d\n", Var->type);

    /* Parse the type */

    switch ((short) Var->type)
    {
      case ASN_INTEGER:
           Var->val.integer = malloc (sizeof(int));
           if (!Var->val.integer)
           {
             snmp_set_api_error (SNMPERR_OS_ERR);
             return (NULL);
           }
           Var->val_len = sizeof (int);

           bufp = asn_parse_int (DataPtr, &ThisVarLen, &Var->type,
                                 (int*)Var->val.integer, Var->val_len);
           VARS_DEBUG2 ("Decoded integer '%d' (%d bytes left)\n",
                        *(Var->val.integer), ThisVarLen);
           break;

      case SMI_COUNTER32:
      case SMI_GAUGE32:
   /* case SMI_UNSIGNED32: */
      case SMI_TIMETICKS:
           Var->val.integer = malloc (sizeof(DWORD));
           if (!Var->val.integer)
           {
             snmp_set_api_error (SNMPERR_OS_ERR);
             return (NULL);
           }
           Var->val_len = sizeof (DWORD);
           bufp = asn_parse_unsigned_int (DataPtr, &ThisVarLen, &Var->type,
                                          (DWORD*)Var->val.integer, Var->val_len);
           VARS_DEBUG2 ("Decoded timeticks '%d' (%d bytes left)\n",
                        *(Var->val.integer), ThisVarLen);
           break;

      case ASN_OCTET_STR:
      case SMI_IPADDRESS:
      case SMI_OPAQUE:
           Var->val_len    = *&ThisVarLen;     /* String is this at most */
           Var->val.string = malloc (Var->val_len);
           if (!Var->val.string)
           {
             snmp_set_api_error (SNMPERR_OS_ERR);
             return (NULL);
           }
           bufp = asn_parse_string (DataPtr, &ThisVarLen, &Var->type,
                                    Var->val.string, &Var->val_len);
           VARS_DEBUG3 ("Decoded string '%s' (length %d) (%d bytes left)\n",
                        (Var->val.string), Var->val_len, ThisVarLen);
           break;

      case ASN_OBJECT_ID:
           Var->val_len = MAX_NAME_LEN;
           bufp = asn_parse_objid (DataPtr, &ThisVarLen,
                                   &Var->type, TmpBuf, &Var->val_len);
           Var->val_len  *= sizeof (oid);
           Var->val.objid = malloc ((unsigned)Var->val_len);
           if (!Var->val.integer)
           {
             snmp_set_api_error (SNMPERR_OS_ERR);
             return (NULL);
           }
           /* Only copy if we successfully decoded something
            */
           if (bufp)
              memcpy (Var->val.objid, TmpBuf, Var->val_len);

           VARS_DEBUG2 ("Decoded OBJID (length %d) (%d bytes left)\n",
                        Var->val_len, ThisVarLen);
           break;

      case ASN_NULL:
      case SMI_NOSUCHINSTANCE:
      case SMI_NOSUCHOBJECT:
      case SMI_ENDOFMIBVIEW:
           Var->val_len = 0;
           Var->val.objid = NULL;
           bufp = asn_parse_null (DataPtr, &ThisVarLen, &Var->type);
           VARS_DEBUG2 ("Decoded ASN_NULL (length %d) (%d bytes left)\n",
                        Var->val_len, ThisVarLen);
           break;

      case SMI_COUNTER64:
           fprintf (stderr, "Unable to parse type SMI_COUNTER64!\n");
           snmp_set_api_error (SNMPERR_UNSUPPORTED_TYPE);
           return (NULL);

      default:
           fprintf (stderr, "bad type returned (%x)\n", Var->type);
           snmp_set_api_error (SNMPERR_PDU_PARSE);
           return (NULL);
    }

    /* Why is this here? XXXXX */
    if (!bufp)
       return (NULL);


    VARS_DEBUG1 ("Adding to list of decoded variables. (%d bytes remain.)\n",
                 AllVarLen);

    /* Add variable to the list */
    *VarLastP = Var;
    VarLastP  = &(Var->next_variable);
  }
  ARGSUSED (version);
  return (bufp);
}
